(*
 * Copyright © 1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

(* $Id: abs-focimc.ml,v 1.3 2005/05/10 04:54:22 rupak Exp $
 *
 * This file is part of the BLAST Project.
 *)


(**
 * This file contains all the code for the Abstraction module that pertains
   to focimodelchecker. The first half is for "concrete" focimodelchecker,
   the second half is for "abstract" focimodelchecker.
 *)

(**************************** CONCRETE FOCIMODELCHECKER   ******************************************)




(***************************************************************************************************)
(**********focimodelchecker: foci_post. updates the lvalue map, and the location info [CONCRETE] ***)
(***************************************************************************************************)

  let pred_to_data_region pred = Region.Concrete {Region.pred = pred}
(*	
  type lvmap_t = ((int * Expression.lval), Expression.expression) Hashtbl.t
*)
(* we need to know ALL interesting lvals of the program, for aliasing reasons *)
  let foci_load_lv_table () = () (* load all interesting lvals into the trace_lv_table *)

(* return new location and new stack *)
  let location_post atomic_reg op = 
    (* let (new_location, new_stack) = *)
    let callStack = Region.list_of_stack atomic_reg.Region.stack in  
    if (C_SD.isFunCall op) then
        begin
          match (C_SD.get_command op).C_SD.Command.code with
              C_SD.Command.FunctionCall fcexp -> 
                let (_fname,_,target) = C_SD.deconstructFunCall fcexp in
	
                  if (_fname = C_SD.__NotImplementedFunctionName) 
                  then 
		    ((C_SD.get_target op), atomic_reg.Region.stack)
                  else   
		    begin
                      if ( C_SD.is_defined _fname) then
			begin
			  let src = C_SD.get_source op in
			    if List.mem src callStack then 
			      raise RecursionException
			    else
			      begin
				if (List.length callStack = !maxStackLen) 
				then (maxStackLen := !maxStackLen + 1;
				      Message.msg_string Message.Normal ("Max stack size: "^(string_of_int !maxStackLen)));
				let newStack = src:: callStack in
				  ((C_SD.lookup_entry_location _fname),
				      Region.CallStack newStack)
			      end
			end
                      else (C_SD.get_target op,atomic_reg.Region.stack) 
(* failwith ("Undeclared important function"^_fname)*)
		    end
	    | _ -> failwith "what the (post) !"
	end
      else 
        try 
	  if (C_SD.isReturn op) then 
	    begin
	      (*let _ = stackInfo := Some (tos_fname,tos_2_fname,true) in *)
	      let (ret, subs, nextLoc) = Region.reg_lastTarget (Region.Atomic atomic_reg) op in
		Absutil.set_ret_pointers ret subs (C_SD.get_location_fname nextLoc);
		let newStack =
		  List.tl callStack
		in
		  (nextLoc,(Region.CallStack newStack) ) 
	    end
	  else 
	    (* let _ = stackInfo := Some (tos_fname,tos_2_fname,false) in *)
	      (C_SD.get_target op,atomic_reg.Region.stack) 
	with Failure ("tl") -> 
	  (* let _ = stackInfo := Some (tos_fname,tos_2_fname,false) in *)
	    (C_SD.get_target op, atomic_reg.Region.stack ) 
	    


(* foci_post: (Region.t * lvmap_t ) -> Operation.t -> (Region.t * lvmap_t * Predicate.predicate) *)
		      
  let foci_post (reg,orig_lvmap) op =
    let _ = Message.msg_string Message.Debug "Foci Post:" in
    let _ = Message.msg_string Message.Debug (Operation.to_string op) in
    (* 1. compute the pred, next lvmap *)
    let lvmap = lvmap_copy orig_lvmap in
    let cmd = C_SD.get_command op in
     (* 1-a. figure out if this is a return statement *) 
    let ret_o =
      if (C_SD.isReturn op) then
	try
	  let r = (Misc.fst3 (Region.reg_lastTarget reg op)) in
	    if r <> C_SD.junkRet then Some r else None
	with _ -> None (* reg_lastTarget can throw an exception if the stack is empty -- for "main" *)
      else None
    in
       (* RJ: only compute newpred, as the lvmap to be returned is just lvmap which will be updated *)
    let op_constraint =
     let rec _doBlock stmtlst pred i (* RJ: Whats this i for ? *) =
       match stmtlst with
	   [] -> pred
	 | stmt::rest -> 
	     begin
	       let sp_pred = block_sp lvmap ret_o  stmt pred in
		 _doBlock rest sp_pred (i+1)
	     end
     in
       match cmd.C_SD.Command.code with
	   C_SD.Command.Block el -> _doBlock el (Predicate.True) 0

	 | C_SD.Command.Pred e -> 
	     begin 
	       let rv =  (* should take a boolean closure here -- see POPL04 paper *)
		 add_to_lv_table (Absutil.getVarsAndDerefs e); 
	     (* IRRELEVANT. everything of note should already have been added *)
	     (pred_lval_map_a lvmap) 0 e (* Predicate.And (e::pred_l) *)
                in
      Message.msg_string Message.Debug ("bwp pred: "^(Predicate.toString rv));
      rv
	     end
	 | C_SD.Command.FunctionCall fcallExp -> 
	     let (fname,_subs,target) = (C_SD.deconstructFunCall fcallExp) in
	     let exists_fun = 
	       try 
		 let fml = C_SD.lookup_formals fname 
		 in true 
	       with C_SD.NoSuchFunctionException _ -> false
	     in
	     let asgn_list = 
	       if (exists_fun || (target = C_SD.junkRet)) then
		 List.map (fun (x,e) -> C_SD.Command.Expr( Expression.Assignment (Expression.Assign,Expression.Symbol(x),e))) _subs 
	       else
		 [C_SD.Command.Expr( Expression.Assignment (Expression.Assign, target, Absutil.get_new_unknown fname ))]
	     in         
	       _doBlock asgn_list (Predicate.True) 0
	 | C_SD.Command.Skip -> Predicate.True
         | C_SD.Command.SymbolicHook _ -> Predicate.True
         | C_SD.Command.Havoc lv -> 
	     failwith ("foci_post: Havoc unimplemented")
	 | C_SD.Command.HavocAll -> failwith "foci_post: HavocAll unimplemented"
	 | C_SD.Command.Phi (sym, flag) ->
	     failwith "foci_post: Phi unimplemented"
   in 
   match reg with
       Region.Atomic atomic_reg ->
	 let (new_location,new_stack) = location_post atomic_reg op  in
	 let post_reg = Region.Atomic { Region.location = new_location; 
					Region.stack = new_stack; 
					Region.data = Region.Abstract 
							{ Region.bdd = PredTable.bddOne;
                                                          Region.lelement = DataLattice.top ;
							  Region.preds = [] };
				      }
	 in
	   (post_reg,lvmap,op_constraint)
     | _ -> failwith "Unexpected region in foci_post!"





(****************************************************************************************************)
(**** focimodelchecker  refine trace [CONCRETE] *****)
(****************************************************************************************************)

(* focimodelchecker_refine_trace : 
   Region.t array -> Operation.t array -> Predicate.predicate array -> lvmap_t array -> Region.t list *)

 (* where region is the present known region at the nodes. operation is the operation, and the pred list is the 
    constraints of the Trace Formula -- precomputed while model checking, unlike in vanilla-blast *)

(* this function is very much like block_analyze_trace, described below. The difference is that we 
   get the cons_l for free -- Q: is it a better idea to construct that on demand. Hmm. *)


let fociModelChecker_refine_trace 
  (reg_array: Region.t array) 
  (op_array: Operation.t array)
  (block_array: Ast.Predicate.predicate array) 
  (lvmap_array_o : lvmap_t array option)
  =
  let len_reg_array = Array.length reg_array in
  let conj_block_array = Array.create len_reg_array Predicate.True in  
  let _ = 
    let fill_cba i reg_i = 
      let i_conj = 
        match lvmap_array_o with
        Some (lvmap_array) -> 
          let pred_i = extract_pred reg_i in 
          (pred_lval_map_a (lvmap_array.(i)) 0 pred_i)
        | None -> Predicate.True
      in
      Array.set conj_block_array i i_conj
    in
    Array.iteri fill_cba reg_array 
  in
  let _ = Message.msg_string Message.Debug (Region.to_string (reg_array.(len_reg_array -1))) in
  if not (Options.getValueOfBool "blockfmc") then
    begin
      let return_reg_array =  Array.init len_reg_array (fun i -> reg_array.(i)) in
      let all_cons = (Array.to_list block_array) @ [conj_block_array.(len_reg_array - 1)] in
      let ref_start_pos = 
        try
          let itp_list = (* Stats.time "extract foci itpN"*) (TheoremProver.extract_foci_interpolantN_noconv) all_cons in
          let _ = Message.msg_string Message.Debug 
          ("back from itpN size = "^(string_of_int (List.length itp_list ))) 
        in
        let min_i = ref 0 in
        let rec folder i itp_l = 
          match itp_l with 
          [] -> ()
          | h::t -> 
              let new_pred = Predicate.unprime h in
              let p_id_l = List.map PredTable.addPred (Predicate.getAtoms new_pred) in
              let i_reg = return_reg_array.(i) in
              let l_id = C_SD.location_coords (location_coords_of_region i_reg) in
              let _ = add_local_preds l_id p_id_l in
              let _ = if (new_pred <> Predicate.True && !min_i = 0 ) then
                  min_i := i
              in
                  Array.set return_reg_array i 
            (replace_data_region i_reg (Region.Concrete {Region.pred = new_pred;}));
            folder (i+1) t    
          in
        let _ = folder 1 itp_list in    
        !min_i
      with TheoremProver.FociSatException -> -1
          |TheoremProver.FociSatExceptionAsgn _ -> -1
          |e -> raise e
    in
        (ref_start_pos, Array.to_list return_reg_array)    
  end
        else (* use simplify to get useful blocks and then cut *)
  (* This is 1 greater then op_array *)
  let return_reg_array = Array.create len_reg_array Region.bot in
  let _ = Message.msg_string Message.Debug (Predicate.toString (conj_block_array.(len_reg_array -1))) in
    (* now the conj_block_array is also filled *)
  let (ref_start_pos, interesting_block_sets, suppress_stuff) =
    get_useful_blocks op_array block_array conj_block_array 
  in 
    if interesting_block_sets = [] then (-1, [])
    else
      let (_,extract_block,_,_,_) = suppress_stuff in
      let itp_array_l =  Misc.map_partial 
			   (interpolant_refine extract_block conj_block_array reg_array)
			   interesting_block_sets
      in
      let _ = Message.msg_string Message.Debug "itp_array_l made" in
	(* it should be that all the above are of the same length... hmm. how might one prove that ? *)
      let fused_array = Array.make len_reg_array Predicate.True in
      let itp_l_l = List.map (Array.to_list) itp_array_l in
      let _ = Message.msg_string Message.Debug "itp_l_l made" in
	
      let fused_array = Array.of_list (Misc.map_transpose Predicate.conjoinL itp_l_l) in
	(* will worry about normalizing and removing the "True"s in the fullness of time  *)
      let make_reg i i_reg = 
       (* let old_pred = extract_pred i_reg in *)
	(* need to "clean" the interpolants ... i.e. drop the subscripts etc. *)   
	let new_pred = Predicate.unprime (fused_array.(i)) in
	  Array.set return_reg_array i (replace_data_region i_reg 
					  (Region.Concrete {Region.pred = new_pred;}))
      in 
	Array.iteri make_reg reg_array;
	(ref_start_pos, (Array.to_list return_reg_array ))




        

(**********************************************************************************)
(******************** focimodelchecker [ABSTRACT] *********************************)
(**********************************************************************************)



 (************* cache the CONS constr. for each cfa-edge ***************)

let is_pure_skip cmd e_id plist = 
  match cmd.C_SD.Command.code with
        C_SD.Command.Pred e -> not (List.exists PredTable.is_relevant_lval (Predicate.lvals_of_predicate e))
      | _ -> not (List.exists (PredTable.is_pred_affected e_id) plist)
 
let (cons_table : (C_SD.edge_id_t,((int * int ) * Predicate.predicate)) Hashtbl.t) 
        = Hashtbl.create 101

let predicate_update_constraints_cached (cmd,e) postloc (in_plist,out_plist) = 
  let e_id = (fst (C_SD.get_edge_coords e), C_SD.location_coords postloc) in
  let plist = Misc.sort_and_intersect (in_plist @ out_plist) in
  let this_ts = (List.length in_plist, List.length out_plist) in 
  let compute_and_update_cache () = 
    let pl2s x = Misc.strList (List.map Predicate.toString x) in
    let (a,b,c) = predicate_update_constraints (cmd,e) postloc (in_plist,out_plist) in
    let cp = 
      if (Options.getValueOfBool "focisymm" && (is_pure_skip cmd e_id out_plist)) (*OK to check just with "out" list *)
      then Predicate.True
      else Predicate.conjoinL (a @ b @ c) 
    in
    Hashtbl.replace cons_table e_id (this_ts,cp);
    Message.msg_string Message.Debug 
    (Printf.sprintf "pred_upd_cons: %s :\n %s \n %s \n %s" 
    (C_SD.edge_to_string e) (pl2s a) (pl2s b) (pl2s c));
    cp
  in
  try
    let (ts, res ) = Hashtbl.find cons_table e_id in
    if ((compare ts this_ts) = (-1)) (* cache is stale -- more predicates added *)
    then compute_and_update_cache ()
    else res
  with Not_found -> compute_and_update_cache ()
  
  
(* END CACHE *)

(*************************************************************)
(*********** Various Renaming Functions **********************)
(*************************************************************)

(**** timeframe  rename function p[i] 
replaces:
  bp_next_k --> bp_k_i+1,
  s --> s_i
************)
let rename_pred_timeframe i pred =
  let new_name suff suff' l = 
    match l with
      Expression.Symbol s ->
        let new_s =
          if (Misc.is_prefix bp_next_prefix s) then 
          let _s = Misc.chop_after_prefix bp_next_prefix s in
          (bp_prefix^_s^suff')
        else (s^suff)
      in Expression.Symbol new_s
    | _ -> failwith ("Bad lval rename_pred_timeframe: "^(Expression.lvalToString l))    
  in
  Predicate.symbol_alpha_convert (new_name ("_"^(string_of_int i)) ("_"^(string_of_int (i+1)))) pred
       

let remove_pred_timeframe  pred = 
  let new_name l = 
    match l with
      Expression.Symbol s -> 
        let pieces = Misc.chop s "_" in
        let _ = assert (pieces <> []) in
        let s' = if (pieces <> []) then String.concat "_" (Misc.chop_last pieces) else s in
        Expression.Symbol s'
    | _ -> failwith ("Bad lval remove_pred_timeframe: "^(Expression.lvalToString l))    
  in
  Predicate.symbol_alpha_convert new_name  pred

(**** convert a formula using atomic predicates into a boolean formula *)
(* p should use atoms corresponding to existing predicates.
   This function returns pred[atom_i -> bp_prefix_i] *)
  
let lookup_atom_index p = 
  try (PredTable.getPredIndex p,true) 
  with Not_found ->
    begin
    try (PredTable.getPredIndex (Predicate.negate p),false)
    with Not_found -> raise Not_found
  end
 
let convert_pred_to_boolean pred =
  let tf p = 
    match p with
    Predicate.Atom e -> 
        let (i,b) = 
          try 
            lookup_atom_index p 
          with Not_found -> failwith ("unknown atom in conv_pred_to_bool:"^Predicate.toString p) 
        in
        let p' = Predicate.Atom (Expression.Lval (Expression.Symbol (bp_prefix^(string_of_int i)))) in
        if b then p' else (Predicate.negate p')
    | _ -> p
  in
  Predicate.deep_transform tf pred

let convert_boolpred_to_pred j bp =
  let tf p = 
    match p with
    Predicate.Atom e ->
      begin
         match e with 
           Expression.Lval (Expression.Symbol s) -> 
           let (time_s::idx_s::_)  = 
             try
                if (Misc.is_prefix bp_prefix s) then (* sanity check *) 
                   let pieces = (Misc.chop s "_") in
                   List.rev pieces 
                else failwith (Printf.sprintf "Bad symbol in convert_boolpred_to_bdd: %s : %s" s (Predicate.toString bp))
             with _ -> failwith ("Bad arg to convert_boolpred_to_bdd - symbol: "^s)
           in
           let time = int_of_string time_s in
           let idx = int_of_string idx_s in
           if (time = j) then PredTable.getPred idx
           else if (time = j+1) then (Predicate.Next (PredTable.getPred idx))
           else failwith (Printf.sprintf "Bad atom in convert_boolpred_to_pred: %d %s" j (Predicate.toString bp))
          |  _ -> failwith ("Bad arg to convert_boolpred_to_pred -atom: "^(Expression.toString e))
        end
    | _ -> p 
  in
  Predicate.deep_transform tf bp 

(* takes a predicate made of bp_i variables (ONLY) turns it into a bdd
   replace bp_i_j with bdd_i, replace bp_i_{j+1} with bdd_i' *)

let convert_boolpred_to_bdd j pred = 
   let tf p = 
     match p with
        Predicate.Atom e ->
        begin
         match e with 
           Expression.Lval (Expression.Symbol s) -> 
           let (time_s::idx_s::_)  = 
             try
                if (Misc.is_prefix bp_prefix s) then (* sanity check *) 
                   let pieces = (Misc.chop s "_") in
                   List.rev pieces 
                else failwith (Printf.sprintf "Bad symbol in convert_boolpred_to_bdd: %s : %s" s (Predicate.toString pred))
             with _ -> failwith ("Bad arg to convert_boolpred_to_bdd - symbol: "^s)
           in
           let time = int_of_string time_s in
           let idx = int_of_string idx_s in
           if (time = j) then PredTable.getPredIndexBdd idx
           else if (time = j+1) then PredTable.getPredIndexNextBdd idx
           else failwith (Printf.sprintf "Bad timestamp in convert_boolpred_to_bdd: %d %s" j (Predicate.toString pred))
          |  _ -> failwith ("Bad arg to convert_boolpred_to_bdd -atom: "^(Expression.toString e))
        end
    | _ -> failwith ("[convert_boolpred_to_bdd] bad call to tf: "^(Predicate.toString p)) 
  in
  PredTable.convertPredToBdd_map tf pred
 

(* as of now, this turns the thing into DNF so be careful *)  
let convert_bdd_to_boolpred bdd =
   let idx_to_string i = 
     if isCurrentIndex i then (bp_prefix^(string_of_int i)) 
     else (bp_next_prefix^(string_of_int i)) 
   in
   (* PredTable.bdd_to_pred_map 
   (fun i -> PredTable.string_to_atom (idx_to_string i))
   bdd *)
   ac_convertBddToPred 
   (fun i -> Predicate.Atom (Expression.Lval (Expression.Symbol (idx_to_string i)))) 
   bdd 

let trel_bdd_to_string trel_bdd = 
  let bp = convert_bdd_to_boolpred trel_bdd in
  let bp0 = rename_pred_timeframe 0 bp in
  let cp = convert_boolpred_to_pred 0 bp0 in 
  Predicate.toString cp
  


(**************************************) 
 (** This is there in some form in the above, but here's a clearer explanation.
 * Trace: R_0 -- op_0 -> R_1 ... -> R_i -- op_i --> R_i+1 ... -> R_n 
 * Define:
 *   Cons[i] = CHAIN[i] /\ WPC[i] /\ R_i[i]  where:
 *   CHAIN   =   /\_p (b <=> p)
 *   WPC[i]  = /\_p ( (WP.p.op_i)[i] <=> b[i+1])
 * We want: 
 *   1. R'_i = ITP.(A_i,B_i)
 *   where A_i = /\_(0<=j<=i-1) Cons[j] ,
 *         B_i = /\_(i<=k<=n) Cons[k] 
 *   2. T'_i = ITP.(Cons[i], _rest_)
 *)

(* A few global structures *)

let foci_ctr = ref 0

(* Gathering the foci theory clauses -- tautologies that may make reasoning fully boolean *)
let foci_theory_clauses_ref = ref []

(* for now, just accumulate them *)
let accumulate_theory_clauses (tc_l_l :Predicate.predicate list list) =
   let _ = Message.msg_string Message.Debug "In accumulate_theory_clauses" in
   let proc_tc tc =
     let tc_notime = remove_pred_timeframe tc in
     foci_theory_clauses_ref := tc_notime :: (!foci_theory_clauses_ref);
     let s = List.length !foci_theory_clauses_ref in
     Message.msg_string Message.Debug (Printf.sprintf "proc_theory_clauses: %d : %s" s (Predicate.toString tc_notime))
   in
   List.iter proc_tc (List.flatten tc_l_l)
   
let iter_theory_clauses f =
  Message.msg_string Message.Normal 
  ("Num theory clauses: "^(string_of_int (List.length !foci_theory_clauses_ref)));
  List.iter f !foci_theory_clauses_ref
  (* ignore (List.map (fun p -> Message.msg_string Message.Normal
  (Predicate.toString p)) !foci_theory_clauses) *)

let print_counterexample op_array rev_map_array plist =
 let _ = Message.msg_string Message.Normal ("SAT preds: "^(Misc.strList plist))
 in
 let last_two l = 
   let l' = List.rev l in
   match l' with
     h1::h2::_ -> (h2,h1)
   | _ -> failwith "bad arg to last_two"
 in
 let ptable = Hashtbl.create 101 in
 let add_to_table s =
   try
   if (Misc.is_prefix bp_prefix s) then 
     let pieces = Misc.chop s "_" in
     let (ps,ts) = last_two pieces in
     let p = int_of_string ps in
     let t = try rev_map_array.(int_of_string ts) with _ -> failwith "check print_counterexample" in
     Misc.hashtbl_update ptable t p;
   with _ -> ()
 in
 let _ = List.iter add_to_table plist in
 let print_i i op_i = 
   let (l1,l2,rv) = op_to_string op_i in
   let pred_s =
     let pi_l = try Hashtbl.find ptable i with Not_found -> [] in
     let p_l = List.map PredTable.getPred pi_l in
     Misc.strList (List.map Predicate.toString p_l)
   in
   Message.msg_string Message.Normal (Printf.sprintf "Time: %d : %s " i pred_s);
   Message.msg_string Message.Normal (Printf.sprintf "%d :: %s :: %d \n\n" l1 rv l2)
 in
 Message.msg_string Message.Normal "***************Abstract Counterexample**********";
 Array.iteri print_i op_array
   
let check_tautology bp = 
  let gather_timeframes p = 
    let tfms = ref [] in
    let _tf p = 
      match p with
      Predicate.Atom e ->
       begin
        match e with 
        Expression.Lval (Expression.Symbol s) -> 
          (try 
            let ts = int_of_string (Misc.list_last (Misc.chop s "_")) in 
            tfms := ts::!tfms;
            p
          with _ -> (failwith ("Bad atom1 in gather_timeframes  :"^(Predicate.toString p))))
        | _ ->  (failwith ("Bad atom2 in gather_timeframes  :"^(Predicate.toString p)))
       end 
     | _ -> p
   in
   Predicate.deep_transform _tf bp;
   !tfms
 in
 match gather_timeframes bp with
    [] -> (false,PredTable.bddOne)
   |[i] -> 
       let p = Predicate.Not (convert_boolpred_to_pred i bp) in
       let rv = TheoremProver.queryExp (Predicate.Implies (p, Predicate.False)) in
       let rvbdd = if rv then convert_boolpred_to_bdd i bp else PredTable.bddOne in
       (rv,rvbdd)
   | _ -> (false,PredTable.bddZero)
     
let rec process_interpolants_taut i p_list =
    match p_list with
     [] -> ()
   | h::t -> 
     let cp = convert_boolpred_to_pred i h in
    Message.msg_string Message.Normal (Predicate.toString cp);
    let (is_taut,tbdd) = check_tautology h in
    let pflag = if is_taut then Message.Normal else Message.Debug in
    Message.msg_string pflag (Printf.sprintf "Tautology?: %s : %b" (Predicate.toString cp) is_taut);
    if (is_taut) then update_global_inv_bdd tbdd;
    process_interpolants_taut (i+1) t


(* A few "quick" ways to strengthen the trel using foci *)
  
let strengthen_trel i (l_id,l_id') (bdd,bdd') (ibdd,ibdd') (itp,itp') op_cons_timed = 
  (* 1. Copy the values of unchanged predicates *)
  let copy_trel_bdd = 
    let pred_i_list = PredTable.bdd_support ibdd' in
    let unch_p_i_list = 
      List.filter (fun i -> not (PredTable.is_pred_affected (l_id,l_id') i))  pred_i_list 
    in
    Message.msg_string Message.Debug ("Unchanged Preds: "^(Misc.string_of_int_list unch_p_i_list));
    PredTable.make_eq_bdd unch_p_i_list
  in
  let expensive_bdd () = 
  (* 2. A strengthening function -- input: upper bound on postred output*)
  let curr_pred = 
    let joined_bdd = CaddieBdd.bddAnd bdd ibdd in
    let jbd_cartabs_l = PredTable.bdd_cartesian_decomposition joined_bdd in
    let bpred_l = 
      List.map 
        (fun b -> rename_pred_timeframe i (convert_bdd_to_boolpred b)) 
        jbd_cartabs_l 
    in
    Predicate.And bpred_l
  in
  let make_cons_bdd next_pred_bdd = 
    let make_new_curr next_pred_timed = 
     let itp_l = 
      try TheoremProver.extract_foci_interpolantN 
         [curr_pred; op_cons_timed; (Predicate.negate next_pred_timed)]
      with _ -> failwith "Foci says sat or fails -- issue in strengthen"
     in
     List.hd itp_l
    in
    let np_timed = rename_pred_timeframe (i+1) (convert_bdd_to_boolpred next_pred_bdd) in
    let timed_trel_cons = 
      try
        let cp_new = make_new_curr np_timed in
        Predicate.Implies(cp_new, np_timed) 
      with _ -> (Message.msg_string Message.Normal "Strengthen fails!"; Predicate.True)
    in
    let _ = Message.msg_string Message.Debug ("st. constraint:"^(Predicate.toString timed_trel_cons)) in
    convert_boolpred_to_bdd i timed_trel_cons
  in
 (* 3. (De)compose the post region into conjuncts (collection of upper bounds),
        and then strengthen the trel for each conjunct *)
  let _ = Message.msg_string Message.Normal "Now decomposing" in
  let _ = 
    let bb = CaddieBdd.bddEqual bdd bdd' in
    let ibb = CaddieBdd.bddEqual ibdd ibdd' in
    Message.msg_string Message.Debug (Printf.sprintf "[Equal] bdd: %b ibdd: %b" bb ibb)
  in
  let (p'_list) = 
    let c' = CaddieBdd.bddAnd bdd' ibdd' in
    let _ = Message.msg_string Message.Debug 
    (Printf.sprintf "[C'] \n  bdd': %s \n ibdd': %s"
    (PredTable.really_bdd_to_string bdd')
    (PredTable.really_bdd_to_string ibdd'))
    
  in
    let rv = PredTable.bdd_cartesian_decomposition c' in
    let _ = Message.msg_string Message.Debug 
    (Printf.sprintf "[CDEC] \n\n In:\n %s \n\n  Out: \n %s" 
    (PredTable.really_bdd_to_string c') 
    (Misc.string_list_cat "\n" (List.map PredTable.really_bdd_to_string rv))) in
    rv
  in
  
  let _ = Message.msg_string Message.Normal "Now finding copy" in
  (* Q: We could do it for just the predicates in p'_list that are in the support of ibdd'... but *)    
  (*let copied_indices =  
        let b_supp = PredTable.bdd_support ibdd in
        let b'_supp = PredTable.bdd_support ibdd' in 
        Misc.sort_and_intersect b_supp b'_supp
  in      
  let _ = Message.msg_string Message.Normal "Now calling bdd_exist_quantify" in
  let ep' = PredTable.bdd_exist_quantify p' copied_indices in*)
  Misc.accumulate  make_cons_bdd CaddieBdd.bddAnd bddOne (p'_list)
      
  in (* done making expensive bdd *)
  
  if (Options.getValueOfInt "trelc" = 2) then 
    let rv = expensive_bdd () in
    CaddieBdd.bddAnd copy_trel_bdd rv
  else copy_trel_bdd    


  
(* This has the side-effect of updating the edge transition relation in
PredTable.ml, when run in bddpost mode *)
(* TBD:FUNCALL -- constraint generation breaks if you have parameter passing/return *)

let process_inconsistent_ops op_array asm_mask_array op_id_l =
  let proc () i = 
    let op_i = op_array.(i) in
    let (l1,l2,rv) = op_to_string op_i in
    Message.msg_string Message.Normal (Printf.sprintf "[INF%d] %d :: %s :: %d \n\n" !foci_ctr l1 rv l2);
    if (C_SD.isPredicate op_i) then (Array.set asm_mask_array i true);
  in
  List.fold_left proc () op_id_l;
  Message.msg_string Message.Normal (Printf.sprintf "[INF%d] \n\n\n" !foci_ctr)


  
exception TrelUnsatExn of int

let fill_ctrex_cubes cc_array reg_array = 
  Message.msg_string Message.Debug "fill ctrex cubes";
  let n = Array.length reg_array in
  Array.set cc_array (n-1) (PredTable.extract_full_cube (extract_abs_data_reg reg_array.(n-1)).Region.bdd); 
  let idx = ref (n-2) in
  while (!idx >= 0) do
    begin
      let next_cube = cc_array.(!idx + 1) in
      if (PredTable.bdd_cubes next_cube > 1) then 
        (Message.msg_string Message.Error "Warning >1 cube in next_cube!";
        Message.msg_string Message.Normal "Warning >1 cube in next_cube!");
      let trel_bdd = 
        let this_loc = C_SD.location_coords (location_coords_of_region (reg_array.(!idx))) in
        let next_loc = C_SD.location_coords (location_coords_of_region (reg_array.(!idx + 1))) in
        PredTable.get_transrel_bdd (this_loc,next_loc) 
      in
      let this_bdd_int = 
        CaddieBdd.bddAnd
           (PredTable.bdd_preimage next_cube trel_bdd) 
           (extract_abs_data_reg reg_array.(!idx)).Region.bdd 
      in
      if (CaddieBdd.bddEqual this_bdd_int PredTable.bddZero) then
        raise (TrelUnsatExn (!idx))
      else Array.set cc_array !idx (PredTable.extract_full_cube this_bdd_int);
      idx := !idx - 1
    end;
  done;
  () 

  
let build_fmc_constraint_arrays 
        (reg_array : Region.t array) 
        (op_array : Operation.t array) =
 let len_reg_array = Array.length reg_array in 
 let _ = Message.msg_string Message.Debug ("len_reg_array = "^(string_of_int len_reg_array)) in
 assert (len_reg_array = Array.length op_array + 1);
 let printop i a = 
     let (_,_,op_s) = op_to_string a in
     Message.msg_string Message.Debug (Printf.sprintf "%d : %s" i op_s)
 in
 let _ = 
      Array.iteri printop op_array
 in
 let conj_block_array = Array.create len_reg_array Predicate.True in
 let copy_conj_block_array = Array.create len_reg_array Predicate.True in
 let ctrx_cube_bdd_array = Array.create len_reg_array PredTable.bddOne in
 let ctrx_cube_array = Array.create len_reg_array Predicate.True in
 let rel_pred_i_list = PredTable.get_all_pred_indexes () in
 let copy_pred_constr = copy_update_constraints rel_pred_i_list in
 let rev_map_array = Array.create len_reg_array (-1) in 
 let clock_ref = ref 0 in
 (* 1b. For das-dill we need to make a backward pass over the region arrays to
 * whittle them down to a cube array *)
 let cc_posn = 
   if (Options.getValueOfBool "dasdill") then 
     Stats.time "fill ctrx cubes" (fill_ctrex_cubes ctrx_cube_bdd_array) reg_array
 in
 (* 2. build the cons[i] array *)
 let build_cons i op_i = 
   let c_i = C_SD.get_command op_i in
   let thisloc = location_coords_of_region (reg_array.(i)) in
   let postloc = location_coords_of_region (reg_array.(i+1)) in
   let rel_preds_in = 
     if (Options.getValueOfInt "craig" = 2) then PredTable.lookup_location_predicates (C_SD.location_coords thisloc) 
     else rel_pred_i_list
   in
   let rel_preds_out = 
     if (Options.getValueOfInt "craig" = 2) then PredTable.lookup_location_predicates (C_SD.location_coords postloc) 
     else rel_pred_i_list
   in
   let chain_wpc = predicate_update_constraints_cached (c_i,op_i) postloc (rel_preds_in,rel_preds_out) in
   if (chain_wpc <> Predicate.True) then (* Non-trivial edge, generate constr. *)
   begin 
     let r = 
       if (Options.getValueOfBool "regcons") then
          convert_pred_to_boolean (extract_pred (reg_array.(i))) 
       else Predicate.True 
     in
     let _ = 
       if (Options.getValueOfBool "dasdill") then
         let cc = (ctrx_cube_bdd_array.(i)) in
         if (PredTable.bdd_cubes cc > 1) then
           Message.msg_string Message.Error "Warning: 1. bdd-cubes > 1";
         let ccube = convert_pred_to_boolean (convertBddToPred cc) in
         let ccube_i = rename_pred_timeframe (!clock_ref) ccube in
         Array.set ctrx_cube_array (!clock_ref) ccube_i
     in
     let cp = Predicate.conjoinL [chain_wpc;r] in
     let cons_i = rename_pred_timeframe (!clock_ref) cp in
     let copy_i = rename_pred_timeframe (!clock_ref) copy_pred_constr in
        Array.set conj_block_array (!clock_ref) cons_i;
        Array.set copy_conj_block_array (!clock_ref) copy_i;
        Array.set rev_map_array !clock_ref i;
        clock_ref := !clock_ref + 1
   end
 in
 (* fill conj_block_array 0..(len_op_array-1) *)
 let build_cons () = 
   for i = 0 to (Array.length op_array - 1) do
     build_cons i (op_array.(i))
   done;
   (* (Array.iteri build_cons) op_array *) 
 in
 let _ = Message.msg_string Message.Normal "fmc build cons" in
 let _ = Stats.time "fmc build cons" build_cons () in
 let _ = Message.msg_string Message.Normal "done build cons" in
   (* fill cb_array.(len_reg_array-1) *)
 let n = !clock_ref (* len_reg_array - 1 *) in
 let _ = 
   if not (Options.getValueOfBool "dasdill") then 
   let r_n =  rename_pred_timeframe n (convert_pred_to_boolean (extract_pred (reg_array.(len_reg_array -1)))) in   
   Array.set conj_block_array n r_n 
   (* Harmless except in -dasdill in which case its the actual reg, i.e. many cubes.*)
 in
 let _ = 
   if (Options.getValueOfBool "dasdill") then
     let cc = ctrx_cube_bdd_array.(len_reg_array -1) in
     let _ = if (PredTable.bdd_cubes cc > 1) then Message.msg_string Message.Error "Warning: 2. bdd-cubes > 1" in
     let ccube = convert_pred_to_boolean (convertBddToPred cc) in 
     let ccube_n = rename_pred_timeframe n ccube in
     Array.set ctrx_cube_array n ccube_n
 in
 let _ = Message.msg_string Message.Debug "Non-skip constraints" in
 for i = 0 to (!clock_ref - 1) do
   let ac_i = rev_map_array.(i) in
   printop ac_i op_array.(ac_i)
 done;
 (conj_block_array,copy_conj_block_array,rev_map_array,!clock_ref,ctrx_cube_array)

(* TBD:SPEED! *)
let get_cons asm_mask_array copy_conj_block_array conj_block_array num_ops =
  let rec gather idx l =
      if idx > num_ops then (List.rev l)
      else 
        let h = if asm_mask_array.(idx) then copy_conj_block_array.(idx) else conj_block_array.(idx) in
        gather (idx+1) (h::l)
  in
  gather 0 []


let refine_reach_states reg_array op_array conj_block_array copy_conj_block_array asm_mask_array itp_array return_reg_array num_ops =
  let len_reg_array = Array.length reg_array in 
  let all_cons = get_cons asm_mask_array copy_conj_block_array conj_block_array num_ops in
  let itp_list = TheoremProver.extract_foci_interpolantN all_cons in
  let _ = Message.msg_string Message.Debug ("back from itpN size = "^(string_of_int (List.length itp_list ))) in
  let inc_op_list = ref [] in
  let do_rest ()  = 
    let min_i = ref 0 in
    let rec folder i itp_l = 
      match itp_l with
      | [] -> () 
      | h::t ->
          let _ = Array.set itp_array i h in
          let new_i_bdd  = convert_boolpred_to_bdd i h in
          let _ = 
            if (not (CaddieBdd.bddEqual new_i_bdd PredTable.bddOne) && (!min_i = 0)) 
            then  min_i := i
          in
          Message.msg_string Message.Debug ("Access array at:"^(string_of_int i));
          let new_i_reg = insert_bdd_reg (return_reg_array.(i)) new_i_bdd in
          Array.set return_reg_array i new_i_reg;
          let _ =
            let last_bdd = (extract_abs_data_reg (return_reg_array.(i-1))).Region.bdd in
            if not (CaddieBdd.bddEqual last_bdd new_i_bdd) then
              inc_op_list := (i-1)::!inc_op_list;
            in
            let _ = 
              let l_id= C_SD.location_coords (location_coords_of_region (reg_array.(i))) in
              let this_preds = PredTable.bdd_support new_i_bdd in
              PredTable.add_local_preds l_id this_preds
              (*let restricted_itp_bdd = 
                let reach_cube = PredTable.bdd_cube_cover ( (extract_abs_data_reg (reg_array.(i))).Region.bdd) in
                CaddieBdd.bddAnd reach_cube new_i_bdd
              in
              Message.msg_string Message.Normal 
              (Printf.sprintf "itp-cubes: (%d,%d) =  %d restricted %d" (fst l_id)
              (snd l_id) (PredTable.bdd_cubes new_i_bdd)
              (PredTable.bdd_cubes restricted_itp_bdd)); *)
            in
            folder (i+1) t
          in
          let _ = folder 1 itp_list in
          PredTable.print_pred_vector ();
          PredTable.dump_abs ();
          (!min_i,!inc_op_list)
        in 
        Stats.time "refine reach: do_rest" do_rest ()
     

let refine_trel_states reg_array conj_block_array copy_conj_block_array asm_mask_array itp_array return_reg_array num_ops =
  let len_reg_array = Array.length reg_array in 
  let cheap_extract (src_id,dst_id) i cons_i = 
    let itp_b = (extract_abs_data_reg (return_reg_array.(i))).Region.bdd in
    let itp_b' = (extract_abs_data_reg (return_reg_array.(i+1))).Region.bdd in
    let b = (extract_abs_data_reg (reg_array.(i))).Region.bdd in
    let b' = (extract_abs_data_reg (reg_array.(i+1))).Region.bdd in
    let itp = (itp_array.(i)) in
    let itp' = (itp_array.(i+1)) in
    Message.msg_string Message.Debug "Calling strengthen trel";
    strengthen_trel i (src_id,dst_id) (b,b') (itp_b,itp_b') (itp,itp') cons_i
  in
  let rec upd_trel i cons_past cons_future =
    match cons_future with
    [] -> failwith "empty list in refine trel"
    | [h] -> () (* the target region *)
    | cons_i::cons_future' ->
        let _ = 
          Message.msg_string Message.Debug (Printf.sprintf "Refining edge %d of %d" i len_reg_array)
        in
        let rrbdd = (extract_abs_data_reg (return_reg_array.(i+1))).Region.bdd in
        if (not (CaddieBdd.bddEqual rrbdd PredTable.bddOne)) then 
          (* this op participated in proof of unsat *)
          begin
            let src_id = C_SD.location_coords (location_coords_of_region (return_reg_array.(i))) in
            let dst_id = C_SD.location_coords (location_coords_of_region (return_reg_array.(i+1))) in
            let itp_b = 
              let b_cheap = cheap_extract (src_id,dst_id) i cons_i in
              let b_exp =     
                let _ = Message.msg_string Message.Debug "trel:foci" in
                let itp =
                  let a = cons_i in
                  let b = 
                    if (Options.getValueOfInt "trelc" = 0) then
                      (cons_past @ cons_future')
                    else 
                      ([itp_array.(i) ; Predicate.negate itp_array.(i+1)])
                  in
                  try List.hd (TheoremProver.extract_foci_interpolantN [a;(Predicate.And b)]) 
                  with e -> 
                    Message.msg_string Message.Error ("Foci says:"^(Printexc.to_string e));
                    Message.msg_string Message.Error "WARNING! : BAD ITP!";
                    (Predicate.Not (Predicate.And b))
                  in
                convert_boolpred_to_bdd i itp
                in
                CaddieBdd.bddAnd b_cheap b_exp
              in
              PredTable.update_transrel_bdd (src_id,dst_id) itp_b;
              (* lets check if this is a trivial interpolant *)
              let _ = 
                let triv_trel_bdd = 
                let b_i = (extract_abs_data_reg (return_reg_array.(i))).Region.bdd in
                let b_i1' = PredTable.make_primed_bdd ((extract_abs_data_reg (return_reg_array.(i+1))).Region.bdd) in
                CaddieBdd.bddImp b_i b_i1'
              in
                if (CaddieBdd.bddEqual (itp_b) triv_trel_bdd) then 
                  Message.msg_string Message.Normal "TREL-TRIV"
                else Message.msg_string Message.Normal "TREL-NON-TRIV"
              in
              (* lets do a sanity check *)
                let b_i = (extract_abs_data_reg (return_reg_array.(i))).Region.bdd in
                let b_i1 = (extract_abs_data_reg (return_reg_array.(i+1))).Region.bdd in
                let b_i1' = CaddieBdd.replace b_i1 (!PredTable.next_var_map) in
                let b_i' = CaddieBdd.bddAndAbstract b_i itp_b (!PredTable.curr_var_cube)  in
                let leqcheck = CaddieBdd.bddEqual (CaddieBdd.bddImp b_i' b_i1') PredTable.bddOne in
                if not (leqcheck) then 
                  begin
                    Message.msg_string Message.Error ("bdd-itp failure! at i = "^(string_of_int i));
                    Message.msg_string Message.Debug ("bdd-itp failure! at i = "^(string_of_int i));
                    Message.msg_string Message.Debug 
                    (Printf.sprintf "b_i = %s" (PredTable.bdd_to_string b_i));
                    Message.msg_string Message.Debug 
                    (Printf.sprintf "itp_b = %s" (PredTable.bdd_to_string itp_b));
                    Message.msg_string Message.Debug 
                    (Printf.sprintf "b_i' = %s" (PredTable.bdd_to_string b_i'));
                    Message.msg_string Message.Debug
                     (Printf.sprintf "b_i1' = %s" (PredTable.bdd_to_string b_i1'));

                    Message.msg_string Message.Debug "Done bdd print"
                  end;
                () 
              end;
              upd_trel (i+1) (cons_i::cons_past) cons_future'
  in
  Message.msg_string Message.Debug "Now refining transition relation";
  let all_cons = get_cons asm_mask_array copy_conj_block_array conj_block_array num_ops
  in
  upd_trel 0 [] all_cons
 (* end refine_trel_states *)



let check_symm_itp trel_cons itp =               
  if (TheoremProver.interpolant_sat (trel_cons, Predicate.Not itp) = TheoremProver.Sat) 
  then 
    begin
      Message.msg_string Message.Error ("Bogus itp: "^(Predicate.toString itp));
      Message.msg_string Message.Error ("For: ");
      Message.msg_string Message.Error (Predicate.toString trel_cons);
      failwith "Invalid symm interpolant!"
    end


(* Ctrx cube array exists for dasdill ONLY *) 
let refine_trel_symm reg_array conj_block_array copy_conj_block_array ctrx_cube_array asm_mask_array rev_map_array num_ops = 
  let min_i = ref (-1) in
  let inc_op_l = ref [] in
  let curr_bdd1_ref = ref (PredTable.bddOne) in
  let curr_bdd2_ref = ref (PredTable.bddOne) in
  let curr_trel_ref = ref [] in
  let update_skip_edges start_idx stop_idx plist =
    let cp_bdd = PredTable.make_eq_bdd plist in
    for i = start_idx to stop_idx do
      let src_id = C_SD.location_coords (location_coords_of_region (reg_array.(i))) in
      let dst_id = C_SD.location_coords (location_coords_of_region (reg_array.(i+1))) in
      PredTable.update_transrel_bdd (src_id,dst_id) cp_bdd
    done;
  in
  let rec folder i_clock cons_l itp_l = 
    let _ = assert (List.length cons_l = List.length itp_l) in
    match (cons_l, itp_l) with
      (_,[]) -> ()
    | (_,[h]) -> () (* the last fellow is from the constraint due to "target" state --irrelev *)
    | (ch::ct, h::t) ->
        begin
          (* printing *)
          let i = rev_map_array.(i_clock) in
          if (h <> Predicate.True) then
            begin
              let _ = if (Options.getValueOfBool "checktree") then check_symm_itp ch h in
              if !min_i = -1 then min_i := i;
              let _ = inc_op_l := i::!inc_op_l in
              let src_id = C_SD.location_coords (location_coords_of_region (reg_array.(i))) in
              let dst_id = C_SD.location_coords (location_coords_of_region (reg_array.(i+1))) in
              let trel_i_bdd = convert_boolpred_to_bdd i_clock h in
              PredTable.update_transrel_bdd (src_id,dst_id) trel_i_bdd;
              let (inpreds,outpreds) = 
                let _preds = (PredTable.bdd_support trel_i_bdd) in
                  (List.filter PredTable.isCurrentIndex _preds, 
                  Misc.map_partial PredTable.getCIndexFromNIndex _preds)
              in
              let _ =   PredTable.add_local_preds dst_id outpreds in
              let _ =  if (i_clock > 0) then  update_skip_edges (rev_map_array.(i_clock -1) + 1) (i-1) inpreds
              in 
              if (Options.getValueOfBool "checktree" && not (Options.getValueOfBool "dasdill")) then
                begin
                  let trel_bdd = PredTable.get_transrel_bdd (src_id,dst_id) in
                  curr_bdd1_ref := PredTable.bdd_postimage !curr_bdd1_ref trel_bdd;
                  curr_bdd2_ref := PredTable.bdd_postimage !curr_bdd2_ref trel_i_bdd;
                  (* bdd1 => bdd2 *)
                  if not (CaddieBdd.bddEqual (CaddieBdd.bddImp !curr_bdd1_ref !curr_bdd2_ref) PredTable.bddOne)
                  then (failwith "Bdd-image interpolant issue (IMP)");
                  if (Options.getValueOfBool "bddprint") then
                    begin
                      Message.msg_string Message.Debug "Using TREL:";
                      Message.msg_string Message.Debug (trel_bdd_to_string trel_bdd);
                    Message.msg_string Message.Debug (Printf.sprintf "Next abs. state at %d#%d" (fst dst_id) (snd dst_id));
                    Message.msg_string Message.Debug (Predicate.toString (convertBddToPred !curr_bdd1_ref)) 
                    end
                end
            end;
          folder (i_clock + 1) ct t 
        end
  in
  let all_cons = get_cons asm_mask_array copy_conj_block_array conj_block_array num_ops 
  in
  let (all_itp_l,tc_l_l) = 
    if (Options.getValueOfBool "dasdill") then
      let get_res () = 
      let cube_cons = Misc.array_to_list_n ctrx_cube_array (num_ops + 1) in
      let trel_cons = Misc.chop_last all_cons in
      let dasdill_itp_l = TheoremProver.extract_foci_interpolantN_dasdill trel_cons cube_cons in
      if (Options.getValueOfBool "checktree") then List.iter2 check_symm_itp trel_cons dasdill_itp_l;
      (dasdill_itp_l @ [Predicate.True],[])
      in 
      Stats.time "get dd itp" get_res ()
    else (Stats.time "get symm itp" TheoremProver.extract_foci_interpolantN_symm all_cons) 
  in (* may throw SAT exn *)
  let _ = 
    if (Options.getValueOfBool "checktree" && not (Options.getValueOfBool "dasdill") &&
    (TheoremProver.interpolant_sat (Predicate.And (all_itp_l),Predicate.True) = TheoremProver.Sat))
  then (failwith ("invalid symm interpolant! (SAT)"))
  in
  let _ = Stats.time "process thy clauses" accumulate_theory_clauses tc_l_l in
  let _ = Message.msg_string Message.Normal "Cleaned Interpolants:" in
  let _ = process_interpolants_taut 0 all_itp_l in
  let process_itps () = folder 0 all_cons all_itp_l in
  let _ = Stats.time "rename/upd bdds" process_itps () in
  let rsp = if (!min_i = -1) then (-1) else (!min_i + 1) in
  if (Options.getValueOfBool "checktree" 
      && (not (Options.getValueOfBool "dasdill")) 
      && (not (CaddieBdd.bddEqual !curr_bdd2_ref PredTable.bddZero))) then (failwith "Bdd-image interpolant issue (SAT)");
  (rsp,!inc_op_l)
  (* INV: !min_i <> -1 -- the if can't fire as foci will have thrown a sat exn *)

 
let ac_fociModelChecker_refine_trace_abstract 
        (reg_array : Region.t array) 
        (op_array : Operation.t array) =
   let len_reg_array = Array.length reg_array in 
   try 
   let (conj_block_array,
        copy_conj_block_array,
        rev_map_array,
        num_ops,
        ctrx_cube_array) = build_fmc_constraint_arrays reg_array op_array 
   in
   let asm_mask_array = Array.create len_reg_array false in 
   let return_reg_array = Array.init len_reg_array (fun i -> reg_array.(i)) in
   let itp_array = Array.create len_reg_array Predicate.True in
   let get_ref_start_pos () = 
       refine_reach_states reg_array op_array conj_block_array copy_conj_block_array asm_mask_array itp_array return_reg_array num_ops
   in
   let refine_trel () = 
       refine_trel_states reg_array conj_block_array copy_conj_block_array asm_mask_array itp_array return_reg_array num_ops
   in
   let refine_trel_sy () = refine_trel_symm reg_array conj_block_array copy_conj_block_array ctrx_cube_array asm_mask_array rev_map_array num_ops 
   in
   let strengthen_pass () =
     let _ = foci_ctr := !foci_ctr + 1 in
     let (ref_start_pos,inc_op_l) = 
       if (Options.getValueOfBool "focisymm") then
         (Stats.time "refine trel symm" refine_trel_sy ())
       else
         begin
         let rv = Stats.time "itp do_refine" get_ref_start_pos () in
         if ((fst rv) <> -1 && Options.getValueOfBool "bddpost") then (Stats.time "itp refine trel" refine_trel ()) ;   
         rv
         end
     in
     (ref_start_pos,inc_op_l)
   in (* end strengthen_pass *)
   (* EXN: FociSat *)
   let min_rsp = ref (-1) in
   let final_rsp = 
     try 
       while (Options.getValueOfBool "multistr" || !min_rsp = -1) do
         let (rsp,inc_op_l) = Stats.time "strengthen pass" strengthen_pass () in
         if (!min_rsp = -1) || (rsp < !min_rsp) then min_rsp := rsp;
          process_inconsistent_ops op_array asm_mask_array (List.rev inc_op_l);
       done;
       !min_rsp
     with  
        TheoremProver.FociSatException -> !min_rsp
      | TheoremProver.FociSatExceptionAsgn plist -> (* wretched trace is satisfiable *)
            begin
             Stats.time "print ctrx" (print_counterexample op_array rev_map_array) plist;
             !min_rsp
            end
        | e -> raise e
   in
      (final_rsp, Array.to_list return_reg_array)
    with TrelUnsatExn i -> (i, Array.to_list reg_array)

(** this uses the current transition relation bdd to strengthen the regions
    along the path instead of firing a fresh query to foci *)
let dummy_table = Hashtbl.create 2 (* Sigh. Region.post --- and that damn tar... *)
let fociModelChecker_refine_trace_abstract_abstract
        (reg_array : Region.t array) 
        (op_array : Operation.t array) =
      let len_reg_array = Array.length reg_array in 
      let _ = Message.msg_string Message.Debug (" len_reg_array = "^(string_of_int len_reg_array)) in
      (* print the bloody thing! *)
      let pr i r_i = Message.msg_string Message.Debug (Region.to_string r_i) in
      Array.iteri pr reg_array;
        
      assert (len_reg_array = Array.length op_array + 1);
      (* 1. return_reg_array is a copy of reg_array *)
      let return_reg_array = Array.init len_reg_array (fun i -> reg_array.(i)) in
      let back_reg_array = Array.init len_reg_array (fun i -> reg_array.(i)) in
      let rec get_ref_start_pos i =
        if i > 0 then
         begin
           let prereg = pre (back_reg_array.(i)) (op_array.(i-1)) dummy_table in
           let precapreg = (Region.cap (back_reg_array.(i-1)) prereg) in
           let _ = Array.set back_reg_array (i-1) precapreg in
           if Region.is_empty (back_reg_array.(i-1)) then i 
           else get_ref_start_pos (i-1)
         end
        else (-1)
      in
      let rec refine_reg_array i =
        if i > (len_reg_array -1) then () else
          begin
            assert (0 < i && i < len_reg_array); 
            let postreg = post (return_reg_array.(i-1)) (op_array.(i-1)) dummy_table in
            Array.set return_reg_array i postreg;
            refine_reg_array (i+1)
          end   
      in
      (* 2. find the i closest to the error to refine *)
      let ref_start_pos = get_ref_start_pos (len_reg_array - 1) in
      if (ref_start_pos <> -1) then 
      (* 3. strengthen the regions from ref_start_pos to len_reg_array-1 *)  
        refine_reg_array ref_start_pos;
        Message.msg_string Message.Debug ("refine_abstract_abstract from: "^(string_of_int ref_start_pos));
      (ref_start_pos, Array.to_list return_reg_array)

      
(***********************************************************************************************)
(****************** foci_post_abstract :********************************************************) 
(***********************************************************************************************)

let foci_post_abstract reg op =
  let _ = Message.msg_string Message.Debug "Foci Post Abstract:" in
  let _ = Message.msg_string Message.Debug (Operation.to_string op) in
   match reg with
       Region.Atomic atomic_reg ->
	 let (new_location,new_stack) = location_post atomic_reg op  in
	 let post_reg = Region.Atomic { Region.location = new_location; 
					Region.stack = new_stack; 
					Region.data = Region.Abstract 
							{ Region.bdd = PredTable.bddOne;
                                                          Region.lelement = DataLattice.top ;
							  Region.preds = [] };
				      }
	 in
	   (post_reg)
     | _ -> failwith ("Unexpected region in foci_post_abstract!"^(Region.to_string reg))



(* END focimodelchecker ******************************************************************************)
let check_abstract_transition reg op reg' = 
  let b_pred = 
    let p = convert_bdd_to_boolpred (extract_abs_data_reg reg).Region.bdd in
    rename_pred_timeframe 0 p
  in
  let b_pred' = 
    let p =  convert_bdd_to_boolpred (extract_abs_data_reg reg').Region.bdd in
    rename_pred_timeframe 1 p
  in
  let op_cons = 
    let postloc = location_coords_of_region reg' in
    let pred_i_list = (PredTable.get_all_pred_indexes ()) in
    let p = predicate_update_constraints_cached (C_SD.get_command op, op) postloc (pred_i_list,pred_i_list) in
    rename_pred_timeframe 0 p
  in
  try 
    TheoremProver.extract_foci_interpolantN [Predicate.And [b_pred;op_cons];Predicate.Not (b_pred')];
    true
  with e -> 
      begin
      Message.msg_string Message.Normal "[FAIL] Bad transition!";
      Message.msg_string Message.Normal ("[FAIL][IN]: "^(Predicate.toString b_pred));
      Message.msg_string Message.Normal ("[FAIL][OUT]: "^(Predicate.toString b_pred'));
      false
      end



(****************************************************************************)
(******************* Code to combine refinement algorithms ******************)
(****************************************************************************)


(* For -fmc *)
let fociModelChecker_refine_trace_abstract 
        (reg_array : Region.t array)
        (op_array  : Operation.t array) =
       let (rsp,ret_reg_l) = ac_fociModelChecker_refine_trace_abstract reg_array op_array in
       if (rsp <> -1) then (rsp, ret_reg_l) 
       else
          let (block_array,_) = 
            let op_l = Array.to_list op_array in
            let reg_l = Array.to_list reg_array in
            let pre_reg_l = Misc.chop_last reg_l in
            let post_reg_l = List.tl reg_l in
            build_block_at_constraint_arrays op_l reg_l pre_reg_l post_reg_l 
          in
          let (rsp,_) = fociModelChecker_refine_trace reg_array op_array block_array None in  
          if  (rsp <> -1 && Options.getValueOfBool "fmc" && not (Options.getValueOfBool "hybrid")) then 
            ac_fociModelChecker_refine_trace_abstract reg_array op_array
          else  (rsp,ret_reg_l)  

(* WARNING: Does not work for TAR/ETAR *)
(* For vanilla blast *)
(* Could also call the real block_at ===> ac_block_at  and call this below function block_a_t *)
(* could also flip the order below -- if block_at fails && -focisymm is on, try fmc-refine ...*)
let trel_refine_block_analyze_trace reg_l op_l id_l =
  if (Options.getValueOfBool "focisymm") then 
    let (rsp,ret_reg_l) = fociModelChecker_refine_trace_abstract (Array.of_list reg_l) (Array.of_list op_l) in
    if (rsp >= 0) then (rsp,(List.nth ret_reg_l rsp), true)  else (-1,Region.bot,false)
  else (block_analyze_trace reg_l op_l id_l)
    
      
