(*********************************************************************************************************)
(***************************** SKIP FUN Code -- skipping useless functions********************************)
(*********************************************************************************************************)
(* NOTE: some of this code is in PredTable -- where we store what lvals are relevant etc. *)

(* if fname is a skipped function, then returns the original function skipped, otherwise None *)    
  let unskip fname =
    try
      Some(Misc.chop_after_prefix __SKIPFunctionName fname) 
    with _ -> None
      
  let skip_fun_process csd_edge =
  assert (Options.getValueOfBool "skipfun");
    let subst_fname fcexp fname =
      match fcexp with
	  Expression.FunctionCall(Expression.Lval (Expression.Symbol(_)), elist)
	  -> Expression.FunctionCall(Expression.Lval (Expression.Symbol(fname)), elist)
	| Expression.Assignment (Expression.Assign,target, 
				 Expression.FunctionCall(Expression.Lval (Expression.Symbol(_)), elist)) ->
	    Expression.Assignment (Expression.Assign,target, 
				   Expression.FunctionCall(Expression.Lval (Expression.Symbol(fname)), elist))
    in
    let cmd = C_SD.get_command csd_edge in
      match cmd.C_SD.Command.code with
       	  C_SD.Command.FunctionCall fcallExp ->
	    begin
	      let (fname,args) =
		match fcallExp with
		    Expression.FunctionCall(Expression.Lval (Expression.Symbol(fn)), elist) ->
		      (fn,elist)
		| Expression.Assignment (Expression.Assign,target, 
					 Expression.FunctionCall(Expression.Lval (Expression.Symbol(fn)), elist)) -> (fn,elist)
		| _ ->
		    if (Options.getValueOfBool "nofp") then
		      (__NotImplementedFunctionName, [])
		    else failwith ("Strange fcallexp!"^(Expression.toString fcallExp))
	    in
	    let skip = 
	      if (PredTable.always_take fname) then false
	      else
		if (PredTable.always_skip fname) then true
		else
		  (* longer case, when we have to check if the "effects" at this callsite are relevant *)
		  (* TBD *)
		  false
		    (*
		      TBD:FULL-BLOWN-SKIP
		      let actuals = List.map snd _subs in
	      	      let globs_mod = C_SD.global_lvals_modified fname in 
		      let influences = targ :: globs_mod in
		      List.exists PredTable.is_relevant_lval influences
			*)
	    in
	      if skip then
		begin
		  Message.msg_string Message.Normal ("Skipping function: "^fname);
		  let new_fname = __SKIPFunctionName^fname in
		  let new_fcexp = subst_fname fcallExp new_fname in
		  let new_cmd =
		    {cmd with
		       C_SD.Command.code =
		       C_SD.Command.FunctionCall new_fcexp}
		  in
		    C_SD.make_edge
		      (C_SD.get_source csd_edge) (C_SD.get_target csd_edge)
		      new_cmd
		end
	      else csd_edge
	  end
      | _ -> csd_edge

	  
(* This file has various functions used by abstraction/cfAbstraction,
   most of the functions (as of now) pertain to refinement *)

  
(**** interpolation stuff begin ****)
(***********************************************************************************)
(***********************************************************************************)
(*****************     Interpolation based predicate discovery    ******************)
(***********************************************************************************)
(***********************************************************************************)

(**
 * This "module" implements the stuff done for learning predicates by interpolating
 * The main function craig_preds: trace -> list of preds for now
 * Actually we should return a list of locations where the preds should be used
 * but first we shall just return the list of predicates
 *) 

(* the general plan:
   1. set up arrays, reg_array, op_array 
   2. construct constraints using block_concrete_data_pre, put in array
   3. create Atom -> block map 
   4. find interesting block sets -- i.e. those which are unsatisfiable, corresponding to 
      unsatisfiable trace fragments (type: int list list)
   5. List.iter block_refine (4.) ...
   6. return the min of the numbers seen in (4.) as the point from which to refine.
*)



let pred_block_table = Hashtbl.create 109

let foci_pred_table = Hashtbl.create 31


let print_lp_table () = 
  let print_loc l = 
      Message.msg_string Message.Debug (Printf.sprintf "Location (%d,%d)" (fst l) (snd l));
    let p_list = Hashtbl.find PredTable.loc_pred_table l in
      List.iter (fun p -> Message.msg_string Message.Debug (string_of_int p)) p_list
  in
  let print_fun f = 
    Message.msg_string Message.Debug (Printf.sprintf "Function (%d)" f);
    let p_list = Hashtbl.find fun_pred_table f in
      List.iter (fun p -> Message.msg_string Message.Debug (string_of_int p)) p_list
  in
    List.iter print_loc (Misc.hashtbl_keys loc_pred_table);
    List.iter print_fun (Misc.hashtbl_keys fun_pred_table)
    



(* funky stuff you can do with functional programming! *)
let next_pos dir x = x + dir

let toggle_next_pos npf =
  let dir = npf 0 in
    next_pos (-dir)

let next_back = next_pos (-1) 
(* end funky stuff *)

(* block_a_t ... interface to simplify *)
let block_assert = TheoremProver.block_assert 
let block_pop = TheoremProver.block_pop
let block_reset = TheoremProver.block_reset 
let check_unsat = TheoremProver.check_unsat
(* end interface to simplify *)

	   (* this returns an atomic_region *)
let insert_pred_in_reg reg pred =
  match reg with
      Region.Atomic atr ->
	
	{ Region.location = atr.Region.location;
	  Region.data = Region.Concrete {Region.pred = pred};
	  Region.stack = atr.Region.stack;
	}
    | _ -> failwith "bad argument to insert_pred_in_reg!"



let min_reg_id = ref (-1) 
let min_adr = ref None 
  (* this is used by predH 7 *)
let dom_rel l1 l2 = 
  (* true if l1 > l2, false otherwise *)
  if l2 = [] then true
  else
    let (sorted_l1,sorted_l2) = (List.sort compare l1,List.sort compare l2) in
    let (min,max) = (List.hd sorted_l2, Misc.list_last sorted_l2)  in
    let rec dom_check l =
      match l with
	  h1::h2::tl -> if (h1 < min && max < h2) then true else dom_check (h2::tl)
	| [h] -> false
	| [] -> false
    in
      dom_check l1

let lv_temp_table = Hashtbl.create 109 

		      

let extract_foci_preds c = 
  if (Options.getValueOfInt "craig" <= 2) then Absutil.extract_foci_preds c 
  else Absutil.extract_foci_preds_parity c 

let list_max l =
  List.fold_left (fun a b -> if (a > b) then a else b) 0 l




(*******************************************************************************************************)
(******************Vanilla "proof-based" refinement, not using interpolants ***************************)
(******************************************************************************************************)
       
let focus_refine extract_block conj_block_array reg_array (block_l : int list) = 
    (* essentially extract the predicates appearing in those blocks *)
    let _ = PredTable.block_a_t_flag := true in
    let head = Misc.min block_l in
    let head_reg = reg_array.(head) in
    let abst_data_reg = extract_abs_data_reg (head_reg) in
    let trace_pred_l = List.map (fun x -> extract_block x) block_l  in
    let all_trace_pred_l = (conj_block_array.(head))::trace_pred_l in
    let t_empty = check_unsat (Predicate.And trace_pred_l) in
    let _ = block_reset () in    
      begin
	let old_pred_list_len = List.length (abst_data_reg.Region.preds) in 
	let new_adr = 
      do_focus abst_data_reg { Region.pred = 
				 match t_empty with
				     true -> Predicate.And trace_pred_l
					 | false -> Predicate.And (all_trace_pred_l)
			     }
        in
        let new_pred_list_len = List.length (new_adr.Region.preds) in
        if (new_pred_list_len >= old_pred_list_len) then PredTable.block_a_t_flag := true;
        if (head <= !min_reg_id || !min_reg_id = -1) then
          begin
	    min_reg_id := head;
	    Array.set reg_array head (insert_data_reg head_reg (Region.Abstract new_adr)) 
          end;
        ()
      end
  

(***************************************************************************************)
(****************GET USEFUL BLOCKS i.e. inconsistent sets of instructions **************)
(***************************************************************************************)



	  
(*** 1. Exposing functions, not clear that this is really required! *)
let suppress_functions op_array block_array =
  let len_op_l = Array.length op_array in
  let suppress_array = Array.make len_op_l false in (* used to "drop assumes" *)
  let extract_block i = 
    match suppress_array.(i) with
	true -> Predicate.True
      | false -> block_array.(i)
  in
  let drop_constraint i = 
    let is_assume_constraint i = 
      match (C_SD.get_command (op_array.(i))).C_SD.Command.code with
	  C_SD.Command.Pred _ -> true
	| _ -> false
    in
      match is_assume_constraint i with
	  true -> Array.set suppress_array i true
	| false -> ()
  in
  let restore_constraint i = 
    Array.set suppress_array i false
  in (* we may never actually use this function *)
  let restore_all_constraints () = 
    Array.iteri (fun i a -> Array.set suppress_array i false) suppress_array 
  in
    (suppress_array, extract_block, drop_constraint, restore_constraint, restore_all_constraints)

(* 2. the main function, that uses the above *)
    
let get_useful_blocks op_array block_array conj_block_array =
  let len_op_l = Array.length op_array in
  let _ = Message.msg_string Message.Normal ("get_useful_blocks: size ="^(string_of_int len_op_l)) in
let print_op (x : Operation.t) =
    let (l1,l2,rv) = op_to_string x in
    Message.msg_string Message.Debug (Printf.sprintf "%d :: %s :: %d \n\n" l1 rv l2)
 in
let _ = Array.iter print_op op_array in
    (* end_cons is needed by FOCIMODELCHECKER to check for coverability *)
  let end_cons =
   if (Options.getValueOfBool "fmcc")
   then conj_block_array.((Array.length conj_block_array) - 1)
   else Predicate.True
  in
  let suppress_stuff = suppress_functions op_array block_array in
  let (suppress_array, 
       extract_block, 
       drop_constraint, 
       restore_constraint, 
       restore_all_constraints) =  suppress_stuff 
  in
    if len_op_l = 0 then (-1,[],suppress_stuff)
    else

  let check_block_unsat block_l =
    (* this function is used as a sanity check *)
    let _ = Message.msg_string Message.Debug "Checking block unsat" in

    let head = Misc.min block_l in
    let trace_pred_l = List.map (fun x -> extract_block x) block_l  in

    let all_trace_pred_l = (conj_block_array.(head))::trace_pred_l in
    let _ = block_reset () in
    let all_together = 
      let t_empty = check_unsat (Predicate.And trace_pred_l) in
      let all_empty = check_unsat (Predicate.And all_trace_pred_l) in
	(t_empty || all_empty) 
    in
    let one_by_one = 
      block_reset ();
      block_assert (conj_block_array.(head));
      List.iter (fun i -> block_assert (extract_block i)) block_l;
      TheoremProver._is_contra ()
    in
      block_reset ();
      if one_by_one && not (all_together) then
	Message.msg_string Message.Debug "simplify inconsistency!";
      if (all_together) then
	Message.msg_string Message.Normal "blockset unsat as expected"
      else Message.msg_string Message.Normal "WARNING: blockset SATISFIABLE!!";
      all_together
  in
    
  let (ref_start_pos, interesting_block_sets) = 
    begin (* find_interesting_blocks *)
      let pos = ref (len_op_l - 1) in
      (*let contra_flag = ref false *) 
      let rec getUsefulBlocks mustSet dir =
	match mustSet with 
	    [] -> 
	      begin
		(* we dont even know the first point of infeasibility *)
		(* we should be moving backwards -- so dir is  "--" *)
		pos := len_op_l - 1;
		Message.msg_string Message.Debug "Assert end condition";
		block_assert (end_cons);
		(* the above stays on the simplify stack *)

		(* from here on it goes as before *)
		block_assert (conj_block_array.(!pos));
		
		(* let _ = contra_flag := TheoremProver._is_contra () in*)
		  while (not (TheoremProver._is_contra ())  && !pos >= 0) do
		    Message.msg_string Message.Debug ("pos = "^(string_of_int !pos));
		    block_pop ();
		    block_assert (extract_block !pos);
		    block_assert conj_block_array.(!pos);
		    (* contra_flag := TheoremProver._is_contra (); *)
		    pos := dir (!pos)
		  done;
		if (not (TheoremProver._is_contra ())) then
		  begin
		    block_reset ();
		    []
		  end
		else
		  begin
		    let rev_dir = toggle_next_pos dir in
		    let next_dir = 
		      match Options.getValueOfString "tsd" with
			  "zig" -> (rev_dir)
			| _ -> (next_back)
		    in
		      block_reset ();
		      getUsefulBlocks ((rev_dir !pos)::mustSet) next_dir
		  end
	      end
	  | h::t -> 
	      begin
		block_assert (end_cons);
		(* we start zigging from h -- the "head" of the trace is the last elt of h::t *)
		(* head is ONLY used for the forward region ! *)
		let head = Misc.list_last (h::t) in
		  block_assert (conj_block_array.(head));
		  List.iter (fun i -> block_assert (extract_block i)) mustSet;
		  (* contra_flag := TheoremProver._is_contra ();*)
		  if (TheoremProver._is_contra () (*!contra_flag*)) then
		    begin
		      block_reset ();
		      mustSet
		    end
		  else
		    begin
		      let next_dir = 
			match Options.getValueOfString "tsd" with
			    "zig" -> (pos := dir h; toggle_next_pos dir) 
			      (* this is the last place we zigged to *)
			  | _ -> (pos := len_op_l - 1;next_back)
		      in
		      	while (not (TheoremProver._is_contra ()) && !pos >= 0) do
   			Message.msg_string Message.Debug 
			  ("Now asserting block: "^(string_of_int !pos));
			block_assert (extract_block !pos);
			(*contra_flag := TheoremProver._is_contra ();*)
			pos := dir !pos
		      done; 
		      assert (TheoremProver._is_contra ()); 
		      (* if assertion checking was on this fails if !pos = 0 *)
		      if (!pos < 0) then failwith "Error in getUsefulBlocks ... Check Simplify behaviour ...";
		      let rev_dir = toggle_next_pos dir in
			block_reset ();
			getUsefulBlocks ((rev_dir !pos)::mustSet) next_dir
		    end
	      end
      in
	(* predH = 6 trawls the trace and drops assumes in the middle to find "better" blocks *)
	(* predH = 7 drops the blocks that are "dominated" *)
	match (Options.getValueOfInt "predH" >= 6) with
	    true ->
	      begin
		let ubl = ref [] in
		let next_useful_blocks = ref [] in
		  next_useful_blocks := getUsefulBlocks [] next_back;
		  (*if (!next_useful_blocks = []) then failwith "No useful blocks found!";*)
		  while (not (!next_useful_blocks = []) && (List.length !ubl < Options.getValueOfInt "bset")) do
		    Message.msg_string Message.Debug "Another u.b. found";
		    check_block_unsat !next_useful_blocks;
		    ubl := (!next_useful_blocks)::!ubl;
		    List.iter drop_constraint !next_useful_blocks;
		    pos := 0; (* must start it off at the end! *)
		    (* this WILL do the suppress-assume trick *)
		    (* adds the assume indices to a table, when doing getUseful, skip over !pos
		       if !pos is in the table. Elementary. *)
		    next_useful_blocks := getUsefulBlocks [] next_back
		  done;
		  match (!ubl) with
		      [] -> (-1, [])
		    | l -> 
			let head = Misc.min (List.flatten l) in
			let l' = 
			  if (Options.getValueOfInt "predH" = 7) then 
			    let max_l = Misc.maximal dom_rel l in
			      Message.msg_string Message.Debug ("Maximal blocks: "^(string_of_int (List.length max_l)));
			      Message.msg_string Message.Debug ("Maximal blocks: "^(string_of_int (List.length max_l)));
			      Message.msg_string Message.Debug ("Orig blocks: "^(string_of_int (List.length l)));
			      Message.msg_string Message.Debug ("Orig blocks: "^(string_of_int (List.length l)));
			      max_l
			  else l
			in
			  (head, l')
	      end
	  | false ->
	      let useful_blocks = getUsefulBlocks [] next_back in
		match useful_blocks with
		    [] -> 
		      Message.msg_string Message.Debug "No useful blocks found!";
		      (-1, [])
		  | l -> 
		      let head = Misc.min l in
			(head,(* List.nth reg_l head, *) [l])
      (* if the trace is feasible then ref_start_pos = -1 and ref_start_reg is the 
	 region for the root from which the trace happens *)
    end
  in
    restore_all_constraints ();
    Message.msg_string Message.Normal "get_useful_blocks done.";
    (ref_start_pos,  interesting_block_sets, suppress_stuff)





 
(******************************************************************************************************************)
(********************************************* TRACE PROJECTION ***************************************************)
(******************************************************************************************************************)

(* We attempt to find a subset of the operations of the trace that actually matter in some way.
   Clearly, if the resulting TF clauses are UNSAT, then the whole TF is unsat.
   Dually, we require that if the TF clauses are SAT, then, modulo termination assumptions, the whole thing is SAT.
*)

(* Expected use: block_analyze_trace --> to minimize counterexamples *)


(* The relevant set of LVALS *)
    (* this is used as a global and is an implicit parameter to many of the functions below *)

exception UnskipException of int
exception RefineFunctionException of string

let error_variable = Expression.Symbol (Options.getValueOfString "errorvar")
let exit_variable = Expression.Symbol (Options.getValueOfString "exitvar")

let rel_field_table : (string,bool) Hashtbl.t = Hashtbl.create 37
				
		  
let rel_variable_table : (Ast.Expression.lval, bool) Hashtbl.t = Hashtbl.create 37
let rel_variable_size = ref 0


let add_rel_fields l = 
  (* INV: see invariant for flush_rel_fields *)
  let l_fields = (Expression.fields_of_lval l) in
    List.iter (fun fld -> 
		 let added = C_SD.add_field_to_unroll true fld in
                 if added then Hashtbl.replace rel_field_table fld true) l_fields
      (* TBD:FIELDS -- mixing tproj/image relevant fields *) 
 

let add_rel_variable l =
  if (Hashtbl.mem rel_variable_table l) then ()
  else
    begin
      Message.msg_string Message.Debug
	(Printf.sprintf "Adding relevant lval %s" (Expression.lvalToString l));
      Hashtbl.replace rel_variable_table l true;
      rel_variable_size := 1 + !rel_variable_size;
      add_rel_fields l
    end
  
let del_rel_variable l =
  if (Hashtbl.mem rel_variable_table l) then 
    begin
      Message.msg_string Message.Debug
	(Printf.sprintf "Removing relevant lval %s" (Expression.lvalToString l));
      Hashtbl.remove rel_variable_table l;
      rel_variable_size := !rel_variable_size - 1 
    end

let flush_relevant_locals fname = 
  let del lv _ = 
        let sc = C_SD.scope_of_lval lv in
        if List.mem fname sc then (del_rel_variable lv)
  in
  Hashtbl.iter del rel_variable_table

let rel_variable_reset () =
  Hashtbl.clear rel_variable_table;
  Hashtbl.clear rel_field_table;
  rel_variable_size := 0;
  add_rel_variable error_variable;
  if (Options.getValueOfBool "exit") then add_rel_variable exit_variable;
  ()
  (* TBD:EXIT *)


let hit_check_cache = Hashtbl.create 1001
let sub_lval_cache = Hashtbl.create 1001

(* alas, we have to take closure here. 
   take the closure of lv_wr ... for each item, check if it is aliased to lv_rd -- such a fellow is lv_rd' == lv_wr^i s.t. lv_rd' aliases lv_rd *)
(* CHECK: hmm is caching ok in the presence of changing closures ? I believe so because if you cached with lv_rd, 
   then the result computed at that point contains all relevant fields of lv_rd.
   I can construt pathological situations where the aliasing happens via some field NOT in lv_rd in which case you get the wrong
   answer -- but that can happen without caching as well :) *)
let stats_nb_tproj_alias_queries = ref 0
        
let hit_check aliasfn lv_wr lv_rd =
  Message.msg_string Message.Debug 
        (Printf.sprintf "In Hit check (lv_wr = %s ,lv_rd = %s)"        
        (Expression.lvalToString lv_wr) (Expression.lvalToString lv_rd));
  let in_cache = Hashtbl.mem hit_check_cache (lv_wr,lv_rd) in
  if (in_cache) then Hashtbl.find hit_check_cache (lv_wr,lv_rd)
  else
    begin
      let clos = C_SD.lvalue_closure_stamp true lv_wr in (* caching done inside C_SD *)
      Message.msg_string Message.Debug (Printf.sprintf "Closure lv_wr = %s"
      (Misc.strList (List.map Expression.lvalToString clos)));
      let get_res () = 
	try
	  Some
	    ((List.find 
		(fun lvwri -> 
		   let (e1,e2) = ((Expression.Lval lvwri),(Expression.Lval lv_rd)) in
		     if e1 = e2 then true else
		       begin 
                               stats_nb_tproj_alias_queries := 1 +
                               !stats_nb_tproj_alias_queries;
                               Absutil.queryAlias_cache e1 e2
                       end      
                   ))
	       clos)
	with Not_found -> None
      in
      let get_res_aliasfn () = (* throwing this back in for time comparisons *)
        let get_aliases () = aliasfn lv_rd in
        let rd_aliases = Stats.time "get aliases aliasfn" get_aliases () in
        (* RJ: this is cached inside aliasfn -- so we don't cache again *)
        let a_table = Misc.hashtbl_of_list  7 rd_aliases in
        try 
           Some (List.find (Hashtbl.mem a_table) clos)
        with Not_found -> None
      in                
      let res = 
              if (Options.getValueOfBool "aiter") then 
                      Stats.time "get res aliasfn" get_res_aliasfn () 
              else    
                      Stats.time "get res" get_res () 
      in
	Hashtbl.replace hit_check_cache (lv_wr,lv_rd) res;
        let lvos lvo = match lvo with None -> "None" | Some lv ->
          Expression.lvalToString lv in
         Message.msg_string Message.Debug 
        (Printf.sprintf "Hit check (lv_wr = %s ,lv_rd = %s,res = %s)"        
        (Expression.lvalToString lv_wr) (Expression.lvalToString lv_rd) (lvos
        res));
       res
    end

let gen_is_relevant aliasfn l = 
  let _mapper lv_read _ = 
    match hit_check aliasfn l lv_read with
	None -> None
      | Some lv_read' -> Some (lv_read,lv_read')
  in
    Misc.hashtbl_map_partial _mapper rel_variable_table

let max_relevant_size = ref 0

let dump_relevant () =
  let lvs = (Misc.hashtbl_keys rel_variable_table) in
  let slvs = List.map (fun lv -> Expression.lvals_of_expression (Expression.Lval
  lv)) lvs
  in 
  let rv = List.flatten slvs in
  if (List.length rv > !max_relevant_size) then
    begin
      max_relevant_size := List.length rv;
      Message.msg_string Message.Normal ("Max rel-cone size"^(string_of_int
      !max_relevant_size))
    end; 
  rv
        
  
let mods_on_path_table = Hashtbl.create 101
let mods_on_path_reset () = Hashtbl.clear mods_on_path_table


(* when you add a new "relevant" lv -- means you are taking the op --> means you are resetting the mods_on_path table.
   Hence, you ONLY do the add_and_hit_check when you are adding a mod_on_path *)
let add_and_hit_check aliasfn mod_lv =
  if Hashtbl.mem mods_on_path_table mod_lv then false
  else
    begin
      Hashtbl.replace mods_on_path_table mod_lv true;
      let find_rel = Stats.time "gen is relevant" (gen_is_relevant aliasfn) mod_lv in 
      (find_rel <> [])
    end	


(* NOTES on proc_write:
   when  you do an assign lv = e,
   then we cal gen_is_relevant to get a list:
   [ ... (lv_rd,lv_rd')...] where lv_rd is "read" in the future and thus  is relevant now,
   and lv_rd' == lv^i (in the closure of lv) such that lv_rd' aliases lv_rd.
   then, corresponding to this "hit", the new "read" value (add) is e^i,
   which is lv^i[e/lv] which is exactly e^i ... which is the new "value" of lv_rd,
   and "lv_rd" is itself irrelevant now (killed) and hence put in del.
   if lv_rd' == lv_rd then we can definitely delete lv_rd from the relevant set, and replace with e^i
   otherwise not. We must do the above for all lv' that are must aliases of lv (including lv)
*)
		       
let rec process_asgn_block aliasfn fname a_block add_c del_c take =
  (* first, lets put add_c/del_c in the table etc -- so we dont have to fuss with them when checking relevant etc. *)
  let aliasfn_f = aliasfn fname in    
  let _ = 
  List.iter del_rel_variable del_c; 
    List.iter add_rel_variable add_c 
in      
  let proc_write (lv,e) =
    (* returns a 4-tuple. (must_add, must_del, may_add,may_del) *)
    let (must_add,must_del,may_add,may_del) = 
      (ref [],ref [],ref [],ref []) 
    in
    let must_aliases_lv = lv::(AliasAnalyzer.get_all_must_aliases lv) in    
    let get_hit_list () = 
      Misc.map_partial 
	(fun lv' -> 
	   let lv'_hits = gen_is_relevant aliasfn_f lv' in
	     if lv'_hits = [] then None else Some (lv',lv'_hits))
	must_aliases_lv
    in
    let hit_list = Stats.time "get hit list" get_hit_list () in
    let proc_hit lv' (lv_rd,lv_rd') = 
      let spairs = [((Expression.Lval lv'),e)] in
      let update al_ref dl_ref =
	let subs_lvals = 
	  Expression.lvals_of_expression 
	    (Expression.substituteExpression spairs (Expression.Lval lv_rd')) 
	    (* we compute e^i by subst. lv_rd' == lv' ^i using [e/lv'] *) 
	in
        Message.msg_string Message.Debug (Printf.sprintf "Subs (lv' , e) in lv_rd' : %s , %s, %s " 
        (Expression.lvalToString lv') (Expression.toString e)
        (Expression.lvalToString lv_rd'));
        Message.msg_string Message.Debug (Printf.sprintf "subs_lvals: %s"
        (Misc.strList (List.map Expression.lvalToString subs_lvals)));
	  al_ref :=  subs_lvals @ !al_ref;
	  dl_ref := lv_rd :: !dl_ref
      in
	(* INV: its a MUST hit if lv_rd == lv_rd', o.w. its a may hit *)
	if (lv_rd' = lv_rd) then
	  update must_add must_del
	else update may_add may_del
    in
    let process_all_hits () = 
      List.iter (fun (lv',hits) -> List.iter (proc_hit lv') hits) hit_list
    in
    let _ = Stats.time "process all hits" process_all_hits () in
      (!must_add,!must_del,!may_add,!may_del)
  in
    match a_block with
	(target,value)::tl ->
	  begin
	    let _ = Message.msg_string Message.Debug
		      (Printf.sprintf "Hit assign %s := %s"
			 (Expression.lvalToString target)
			 (Expression.toString value))
	    in
	    let (tadd,tdel,alias_add,alias_del) = 
	      Stats.time "proc write targ" proc_write (target,value) 
	    in
	      (* (alias_del) = [] empty in the worst case, use it only to see if this op was relevant,ie to compute "take" *)
	    let read = Misc.sort_and_compact (tadd @ alias_add) in
	    let written = (target::tdel) in
	    let take' = ((tdel@alias_del) <> []) in 
	      (* take' == true IFF some update made was relevant. if some update was relevant,
		then tdel@alias_del <> [], as every relevant update sends something into the del bin *)
	      (* OLD CODE: let written = [target] (* TBD:SOUNDNESS *) in
	            let read = Expression.allVarExps_deep value in
	         TBD:TPROJ -- do we really need allVarExps_deep ? *)
	    let (add_c',del_c') =
	      if take' then
		(Misc.union (Misc.difference add_c  written)  read,
		 Misc.difference (Misc.union del_c  written)  read)
	      else
		(add_c,del_c)
	    in
	      process_asgn_block aliasfn fname tl add_c' del_c' (take || take')
	  end
      | [] -> ((add_c,del_c),take)


	        
(* unsat_flag == true means we just want to gather the "residual" funcalls *)	  
let process_trace_op unsat_flag aliasfn op_array left last i =
  (* ABC INV: 0 < i < Array.length op_array *)
  (* INV: the two locations MUST be in the same CFA *)
  let level = Message.Debug in
  let _ = Message.msg_string Message.Normal ("TPROJ: at op "^(string_of_int i)) in
  let op = op_array.(i) in
  let call_i = left i in
  let call_op = op_array.(call_i) in
  let op_fname = C_SD.get_location_fname (C_SD.get_source op) in
  (*let this_lvalue_closure =
     if (Options.getValueOfBool "fldunroll") then
     C_SD.lvalue_closure_stamp (* (is_relevant_field_tproj []) *)
     else C_SD.lvalue_closure
     in
    let lvalue_hits lv =
    List.flatten (List.map (aliasfn op_fname) (this_lvalue_closure lv))
    in -- used to "close" mods_on_path w.r.t. aliasing etc *)
  let last_op = op_array.(last) in
  let _ =
    Message.msg_string Message.Debug "proc_trace_op: ";
    Message.msg_printer Message.Debug C_SD.Command.print (C_SD.get_command op)
  in    
  let process_return  callsite_op call_triple caller_name retexp = 
    let (targ_option,fname,args) = call_triple in
    let cb_asgns =
      let fmls =
	match C_SD.lookup_formals fname with
	    C_SD.Variable l -> l
	  | C_SD.Fixed l -> l
	in
	copy_back_asgns
	  (Options.getValueOfBool "incref")
	  (Misc.partial_combine fmls args)
	  fname
    in
    let all_asgns = 
      match targ_option with
	  None -> cb_asgns
	| Some(targ) -> (targ,retexp)::cb_asgns
    in
    let ((add,del),ret_take) =
      process_asgn_block aliasfn caller_name
	all_asgns [] [] false
    in
    let _ = Message.msg_string Message.Debug "Now check glob_take" in
    let get_glob_take () = 
      if (Options.getValueOfBool "dmod") 
      then 
        let (glob_lvals,ng_lvals) = 
          Misc.filter_cut C_SD.is_global (dump_relevant ()) 
        in
        let _ = Message.msg_string Message.Normal "check mod_on_edge" in
        (List.exists (C_SD.local_mod_on_edge callsite_op) ng_lvals
        || List.exists (C_SD.global_mod_on_call fname) glob_lvals)
      else
        begin
          let globals_mod = (C_SD.global_lvals_modified fname) in (* closed under alias *) 
          let _ = 
              if (level = Message.Normal) then ()
              (*Message.msg_string level ("globs mod: "^(Misc.strList
                (List.map Expression.lvalToString globals_mod)))*)
          in
           List.exists (fun lv -> (gen_is_relevant (aliasfn "") lv) <> []) globals_mod
        end
    in
    let _ = Message.msg_string Message.Debug "done checking glob_take" in
    let glob_take = Stats.time "get glob take" get_glob_take () in
    let take = glob_take || ret_take in
    let i' = if take then i-1 else (call_i - 1) in	      
      ((add,del),take,i')
  in
    
  let skipcheck (targ_option,fname,args) =
        Message.msg_string Message.Debug "in skipcheck" ;
    if (not (Options.getValueOfBool "skipfun") 
    || (Options.getValueOfBool "skipcheck" = false)) then ()
      else
	match unskip fname with
	    None -> ()
	  | Some(orig_fname) ->
              if (not (C_SD.is_defined orig_fname)) then ()
              else
	        let dummyretexp = Absutil.get_new_unknown orig_fname in
	        let (_,take,_) = process_return op  (targ_option,orig_fname,args) op_fname dummyretexp in
	        if take then 
                        begin
                        Message.msg_string Message.Normal ("Whoops: skipped --need to take: "^orig_fname);

                        if (Misc.is_substring orig_fname "__initialize__" &&
                            not (Options.getValueOfBool "initialize")) then 
                              raise (UnskipException (-1));
                        PredTable.add_take_fun orig_fname; 
                        raise (UnskipException (i-1))
                        end
                else (Message.msg_string Message.Debug ("Can still skip"^orig_fname))
  in
    match Operation.get_info op with
	Operation.Ret retexp ->
	  begin
	    if (unsat_flag) then (([],[]),false,call_i - 1)
	    else
	      match Operation.get_info call_op with
		  Operation.Call (_, targ_option,fname,args) ->  (* aka: *)
		    process_return call_op (targ_option,fname,args) 
		    (C_SD.get_location_fname (C_SD.get_source call_op))
		    retexp
		| _ -> failwith "bad call-op in process_trace_op"
	  end
      | Operation.Call (_, targ,fname,args) ->
	  (* note: as we are actually processing the call, we have to take it *)
	  begin
            if (unsat_flag) then (([],[]),true,i-1)
	    else
	    let ac_formals =
	      List.map (fun s -> (Expression.Symbol s))
		(match C_SD.lookup_formals fname with
		     C_SD.Variable fml -> fml
		   | C_SD.Fixed fml -> fml)
	    in
	    let asgn_list = Misc.partial_combine ac_formals args in
	      (* INV: List.length.formals = List.length args *)
	    let ((add,del),_) = process_asgn_block aliasfn fname asgn_list [] [] false in
	      (* NOTE: that we have used fname -- the callee_name NOT the op_fname here *)
            let _ = flush_relevant_locals fname in
            if (add = [] && !rel_variable_size = 0) then
	      begin
		Message.msg_string Message.Error ("CONE EMPTY! -- probably an ERROR:"^fname);
		Message.msg_string Message.Normal ("CONE EMPTY! -- probably an ERROR:"^fname);
		if (Options.getValueOfBool "tpbug") then (failwith "CONE EMPTY UNSAFE ? :-(")
	      end;
            let ((add,del), take, next_i) =
              let ac_call_i = left (i-1) in  
              let current_loc = C_SD.get_source op in
              let current_fname = C_SD.get_location_fname current_loc in
              if (Options.getValueOfBool "projfun" 
                        && ac_call_i <> -1 
                        && not (Options.is_errorseed current_fname))
              then
              begin
                let _ = 
                  List.iter (del_rel_variable) del;
                  List.iter (add_rel_variable) add;
                in      
                let last_loc = current_loc in 
                let this_loc = C_SD.lookup_entry_location current_fname in  
                let rel_lvals = dump_relevant () in
                let _ = Message.msg_string Message.Normal 
                "dmod: projf: check may_be_modif" in
                let cant_jump = Stats.time "projf : may be modified" (C_SD.may_be_modified rel_lvals this_loc) last_loc  in
                if (not cant_jump) then (Message.msg_string Message.Normal
                ("TP: Jumping!:"^(string_of_int i)));
                (([],[]),true,if cant_jump then (i-1) else ac_call_i)
              end
            else
	      ((add,del), true, i-1)
            in
                ((add,del),take,next_i)
	  end
      | Operation.Normal -> (* split cases on whether its an assign or an assume *)
	  begin
	    if (unsat_flag) then (([],[]),false,i - 1)
	    else
	      match (C_SD.get_command op).C_SD.Command.code with
		C_SD.Command.Block l ->
		  let peel_stm stm =
		    match stm with
			C_SD.Command.Expr (Expression.Assignment
			  (Expression.Assign, targ,valu)) -> (targ,valu)
		      | _ -> failwith "Bad stm in peel_stm!"
		  in
		  let ((add,del),take) =
		    process_asgn_block aliasfn op_fname
		      (List.rev (List.map peel_stm l)) [] [] false
		  in
		    ((add,del),take,i-1)
	      | C_SD.Command.Pred p ->
		  begin
		    let this_loc = C_SD.get_source op in
		    let last_loc = C_SD.get_source last_op in
                    (*let _ =
                            if (level = Message.Normal) then 
		      (Message.msg_string level
			("Mods Between "^(string_of_int (List.length
                        mods_between)))) 
                    in*)
                    let hit_checker = (add_and_hit_check (aliasfn op_fname)) in
                    let this_pred_lvals = Predicate.lvals_of_predicate p in
                    let this_pred_relevant = 
                      List.exists hit_checker this_pred_lvals 
                    in
                    let path_mods_relevant = 
                      if (Options.getValueOfBool "dmod") 
                      then
                         begin
                          let rel_lvals = dump_relevant () in
                          let _ = Message.msg_string Message.Normal 
                          "dmod: check may_be_modif" in
                          Stats.time "may be modified" (C_SD.may_be_modified rel_lvals this_loc) last_loc                         
                        end
                     else
                        begin
                        let mods_between =  Stats.time "mods_on_path" (C_SD.mods_on_path this_loc) last_loc
		    
                        in
                        Message.msg_string Message.Debug 
                        ("Mods between#:"^(string_of_int (List.length
                        mods_between)));
                        Stats.time "add and check hits" 
                        (List.exists hit_checker) mods_between 
                        end     
                    in (* note: ok to STOP at first "hit", ie use List.exists, because the set will be reset if the exists is true *)
                    let not_always_reach = (* false in *)
                     (not
			    (Stats.time "Always Reach" (C_SD.always_reach this_loc) last_loc))
                    in 
                    let take =
		      (i+1 = Array.length op_array) (* always take the last op! *)
		      || not_always_reach || this_pred_relevant || path_mods_relevant
		    in
		    let (add,del) = ((if take then this_pred_lvals else []), []) in
		      ((add,del),take,i-1)
		  end
	      | C_SD.Command.Skip -> (([],[]),false,i-1)
              | C_SD.Command.SymbolicHook _ -> (([],[]),(not unsat_flag),i-1)
              (* take it if we see it!*)
              | C_SD.Command.FunctionCall e -> (* an unknown function *)
		  begin
		    match e with
			Expression.FunctionCall (fn, el) ->
			  let fname = C_SD.get_fname fn in
			  let _ = (skipcheck  (None,fname,el)) 
                          in
                          let _ = Message.msg_string Message.Debug "after skch"
                          in
                          
			    (* skipcheck will throw an exception to be caught in block_analyze_trace if reqd.,
			       AFTER changing the various tables etc.*)
			  Message.msg_string Message.Normal
			    ("Unknown fun "^fname^" in trace_project : optimistic");
			    (([],[]),false,i-1)
		      | Expression.Assignment
			  (Expression.Assign, targ, Expression.FunctionCall(fn, el)) ->
			  let fname = C_SD.get_fname fn in
			  let _ = skipcheck  (Some(targ),fname,el) in
                          let _ = Message.msg_string Message.Debug "after skch2"
                          in
                          Message.msg_string Message.Normal
			  ("Unknown fun "^fname^" in trace_project : optimistic");
			  let asgn_block = [(targ,Absutil.get_new_unknown fname)] in
			  let ((_,del),take) =
			    process_asgn_block aliasfn op_fname asgn_block [] [] false
			  in
			    (([],del),take,i-1)
		  end
	      | _ -> failwith ("unexpected value in process_trace_op")
	  end

 (* For strange dependancy reasons, this has to be passed the get_info function -- but will it work ! *)
let print_trace op_l =
  let print_op op = 
    let (l1,l2,rv) = op_to_string op in
  Message.msg_string Message.Normal (Printf.sprintf "%d :: %s :: %d" l1 rv l2) 
in
  List.iter print_op op_l

let print_trace_reg level op_l reg_l = 
        let print_op_reg i (op,reg) = 
            let (l1,l2,rv) = op_to_string op in
            let d = Predicate.toString (extract_pred reg) in
            Message.msg_string level ("Operation: "^(string_of_int i));
            Message.msg_string level (Printf.sprintf "%d :: %s :: %d"
            l1 rv l2);
            Message.msg_string level ("post-reg: "^d);
            i+1
        in
        let zipped = try List.combine op_l reg_l with _ -> failwith "combine error in print_trace_reg" in
        List.fold_left print_op_reg 0 zipped;
        ()
  
let trace_project op_l reg_l =
  let aliasfn =
    if (Options.getValueOfBool "cf") then AliasAnalyzer.get_lval_aliases_scope
    else
      if (Options.getValueOfBool "talias") then
	(fun _ -> trace_lv_table_get_lval_aliases)
      else (fun _ -> get_lval_aliases)
  in
  let trace_size = List.length op_l in
  Message.msg_string Message.Major
	    ("Trace Project: IN size "^(string_of_int trace_size));
  Message.msg_string Message.Error
	    ("Trace Project: IN size "^(string_of_int trace_size));
  print_trace_reg Message.Normal op_l (List.tl reg_l) ;
  let op_array = Array.of_list op_l in
  let reg_array = Array.of_list reg_l in
  let _ = Message.msg_string Message.Debug "Building tp_array" in
    (* Splicing in code that will do the projection only till the point where the thing is sat *)
    (* 1. initialize lvmap/simplify *)
  let _ = block_reset () in
    (* 2. initialize lvmap *)
  let _ = Stats.time "make_lv_stack [tproj]"
	    (make_lv_stack (reg_function_name (List.hd reg_l))) op_l
  in 

    (* compute and assert the constraints for operation i.
       Returns TRUE iff the thing becomes unsat -- FALSE ow *)
  let add_unsat_check_op i =
      begin
	Message.msg_string Message.Debug ("Tproj: asserts op "^(string_of_int i));
	let op = op_array.(i) in
	let pre_reg = reg_array.(i) in
	let post_reg = reg_array.(i+1) in
	let pre_pred = extract_pred pre_reg in
	let atomic_post_reg = 
	  match post_reg with
	      Region.Atomic atr -> atr
	    | _ -> failwith "bad region given to add_unsat_check_op!"
	in
        let fname = reg_function_name pre_reg in
	let block_cons = (block_concrete_data_pre fname 0 atomic_post_reg
			    (C_SD.get_command op) op)
	in
	let conj_block_cons = pred_lval_map 0 pre_pred in
	let is_unsat = 
	  block_assert block_cons;
	  block_assert conj_block_cons;
	  let rv = TheoremProver._is_contra () in
	    block_pop ();
	    rv
	in
	  if (is_unsat) then
	    Message.msg_string Message.Debug  ("Found unsat at op"^(string_of_int i));
      	  is_unsat
      end
  in
    (* end splicing *)	
  (* first we have to make a function that given an operation,
      returns the call-frame-creater of the function *)
  let (left,_) = Misc.paren_match Operation.paren_fun op_array in
  let tp_array = Array.make trace_size false in
  let _ = rel_variable_reset ()

  in
  let last = ref (trace_size - 1) in	      
  let unsat_flag = ref false in
  (* now a strange way to do a decreasing for-loop in OCAML *)
  let i = ref (trace_size - 1) in
  let _ =
    while (!i >= 0) do
      Ast.check_time_out () ;
      let _ = Message.msg_string Message.Debug ". " in
      let ((add,del),take,i') =
	Stats.time "process_trace_op"
	  (process_trace_op !unsat_flag aliasfn op_array left (!last)) (!i)
      in
	Array.set tp_array !i take;
	if (take) then 
          begin
            last := !i; 
            mods_on_path_reset ()
          end;
	i := i'; (* proceed to the next operation -- if !i was a return, then we may have  made a big jump *)
        
	List.iter del_rel_variable del;
	List.iter add_rel_variable add; (* these are done internally in process_asgn_block *)
	  (* now lest check if we need to go on *)
	  (* note that unsat_flag is set to true ONLY if predH < 6 o.w. we just do the entire trace_project *)
	  if (not !unsat_flag) && not (Options.getValueOfInt "predH" >= 6) then 
	    begin
	      if (take) then 
		unsat_flag := add_unsat_check_op !last
	    end
    done
  in
  let _ = block_reset () in (* dont forget to reset the thmprover *)
  let _ = Message.msg_string Message.Debug "tp_array built" in
(*
  let _ = Array.iteri (fun i a -> Message.msg_string Message.Normal (Printf.sprintf  "tp_array: %d -> %b\n" i a)) tp_array in
*)
  let proj_size = Misc.array_counti (fun i a -> a) tp_array in
  let _ = Message.msg_string Message.Error
	    ("Trace Project: OUT size "^(string_of_int (proj_size)))
  in
  let _ = Message.msg_string Message.Major
	    ("Trace Project: OUT size "^(string_of_int (proj_size)))
  in
  let fwd_map_array = Array.create trace_size (-1) in
  let rev_map_array = Array.create proj_size (-1) in
  let idx = ref 0 in
  let _ = 
    for i = 0 to trace_size - 1 do
      if tp_array.(i) then
	begin
	  Array.set fwd_map_array i (!idx);
	  Array.set rev_map_array (!idx) i;
	  idx := !idx + 1
	end
    done
  in
  let _ = assert (!idx = proj_size) in
  let revmap = (fun i -> if (0 <= i) && (i < proj_size) then rev_map_array.(i) else -1) in
  let fwdmap = (fun i -> if (0 <= i) && (i < trace_size) then fwd_map_array.(i) else -1) in
  let mask_fn = fun i -> if 0 <= i && i < trace_size then tp_array.(i) else false in
  let op_l' = Misc.list_mask mask_fn op_l in
  let pre_reg_l = Misc.list_mask mask_fn reg_l in
  let post_reg_l = Misc.list_mask (fun i -> mask_fn (i-1)) reg_l in
  let _ =
    Message.msg_string Message.Debug "Out Trace";
    print_trace op_l'
  in
    (pre_reg_l,post_reg_l,op_l',revmap,fwdmap)


(*******************************************************************************************************)
(******** CF INTERPOLANT REFINE -- takes useful blocks, cuts the trace, adds new predicates ************)
(*******************************************************************************************************)
let stats_nb_non_trivial_functions = ref 0

let cf_interpolant_refine (left,right) extract_block conj_block_array reg_array
  (block_l : int list) = 
  Message.msg_string Message.Debug ("Inside cf interpolant refine:"^(string_of_int (List.length block_l)));
  (* assert (not (block_l = [])); *)
  let len_op_l = Array.length conj_block_array in
  let ord_block_l = List.sort compare block_l in
  let start = List.hd ord_block_l in
  let cons_l = List.map extract_block  ord_block_l in
  let (start_cons,start_cons_pos) = 
    if check_unsat (Predicate.And cons_l) then (Predicate.True,-1)
    else (conj_block_array.(start),start) 
  in
      (* for debugging purposes, we shall dump the interesting blocks *)
  let _ =
          let level = Message.Debug in
          if (level = Message.Normal) then
    begin
      Message.msg_string level (Printf.sprintf "num blocks = %d" (List.length block_l));
      Message.msg_string level  (Printf.sprintf "start cons = [%d] \n "  start_cons_pos );
      Message.msg_printer level  Predicate.print (start_cons);
      Message.msg_string level ("Constraints");
      List.iter2 (fun p i -> Message.msg_string level  ((string_of_int i)^" :  "^(Predicate.toString p)) ) 
	cons_l ord_block_l
    end
  in
  let cut ls i = 
    let l_i = left i in
    let r_i = right i in
    let (l_i,r_i) = if (r_i = i) then (left (i+1), right (i+1)) else (l_i,r_i) in
    let (context,hole) = 
      if (l_i = -1) then ([],ls) 
      else
	List.partition (fun x -> (x <= l_i || (x >= r_i && r_i <> -1))) ls 
    in
    let pre_context = List.filter (fun x -> x <= l_i) context
    in (* to get the symbolic constants *)
    let (phi_minus,phi_plus) = List.partition (fun x -> x <= i) hole in
    let b = (l_i < start_cons_pos) && (start_cons_pos <= i) in 
      (* true means the "start_cons" goes to the past, false means it goes into the future *)
      
    (* RJ:
       Q: can you have simultaneously :
       1. a nontrivial start_cons in the past
       2. a "closed" trace ie all calls return
       3. a context containing constraints from before the call at i ?
       A: I really dont think so !
    *)
    let rv =
      if r_i <> -1 then 
	(i,pre_context,phi_minus,context@phi_plus,b) 
      else 
	(* RJ: can you have a r_i = -1 AND the start_cons in the future ... NO. return true for the b coordinate ... *)
	(i,pre_context,context@phi_minus, phi_plus,true) 
    in
    let b = b || (r_i = -1) in
      Message.msg_string Message.Debug (Printf.sprintf "scp, i, l_i, r_i, b = %d %d %d %d %b" start_cons_pos i l_i r_i b);
      rv
  in
  let trace_cuts = Misc.list_tabulate (cut ord_block_l) 0 (len_op_l) in
    let _ = Hashtbl.clear foci_pred_table in
      (* now for each trace cut, get the predicates and add them to their resp. functions. for now only craig 1 *)
    let process_cut (i,call_past,past,future,b) =

      let _ = Message.msg_string Message.Debug ("In cf process cut at:"^(string_of_int i)) in
      let _ = 
	Message.msg_string Message.Debug ("b : "^(string_of_bool b));
	Message.msg_string Message.Debug ("start_cons : "^(Predicate.toString start_cons))
      in
      let _ = 
	Message.msg_string Message.Debug ("Past Constraints:");
	List.iter (fun p -> Message.msg_string Message.Debug (Predicate.toString (extract_block p))) past
      in
      let _ = 
	Message.msg_string Message.Debug ("Future Constraints:");
	List.iter (fun p -> Message.msg_string Message.Debug (Predicate.toString (extract_block p))) future
      in

	
      let (cons_l_pre,cons_l_post) = 
	match b with
	    false -> (List.map extract_block past, start_cons :: (List.map extract_block future))
	  | true -> (start_cons::(List.map extract_block past),List.map extract_block future)
      in
      let l_id = C_SD.location_coords (location_coords_of_region (reg_array.(i+1))) in (*RJ : check*)
      let callee_name =
	C_SD.get_location_fname (C_SD.lookup_location_from_loc_coords l_id)
      in
      let foci_preds = 
	try 
	  Hashtbl.find foci_pred_table (past)
	with
	    Not_found ->
	      let rv = 
		(* JHALA: temporary patch for segfaults! *)
		match (check_unsat (Predicate.And cons_l_pre)) || (check_unsat (Predicate.And cons_l_post)) with
		    true -> []
		  | false ->
		      let pre_call_cons =
			Predicate.conjoinL (List.map extract_block call_past) in
			Stats.time "extract foci preds (cf)"
			  (Absutil.extract_foci_preds_cf
			  (callee_name,pre_call_cons))
		      	  (cons_l_pre,cons_l_post)
			  
	      in 
		Hashtbl.add foci_pred_table past rv;
		rv
      in
      let _ = Message.msg_string Message.Debug "extraction done" in
      let _ = List.iter (fun i -> Message.msg_string Message.Debug (string_of_int i)) foci_preds
      in
     (* 
      let _ = Message.msg_string Message.Normal ("Accessing i= "^(string_of_int i)) in
      let _ = Message.msg_string Message.Normal  (Printf.sprintf "loc_id (%d, %d)" (fst l_id) (snd l_id)) 
      in 
	Message.msg_string Message.Normal ("#Found preds" ^ string_of_int
        (List.length foci_preds)); *)
	let present_p_l = try Hashtbl.find loc_pred_table l_id with Not_found -> [] in
	let present_fun_p_l = try Hashtbl.find fun_pred_table (fst l_id) with Not_found -> [] in
	let to_add = List.filter
		       (fun p_i -> not( List.mem p_i present_p_l)) foci_preds 
in 
  Message.msg_string Message.Debug ("#Added preds " ^ string_of_int (List.length to_add)); 
  List.iter (fun p_i -> refined ();block_a_t_flag := true; 
	       Message.msg_string Message.Debug (string_of_int p_i)) to_add;
  Hashtbl.replace loc_pred_table l_id (to_add@present_p_l);
  Hashtbl.replace fun_pred_table (fst l_id) (Misc.compact (to_add@present_fun_p_l));
  if (present_fun_p_l = [] && to_add <> []) then 
      stats_nb_non_trivial_functions := !stats_nb_non_trivial_functions + 1
      
      (* RJ: some of the above compacting seems redundant ... *)
in
  List.iter process_cut trace_cuts



(*******************************************************************************************************)
(********** INTERPOLANT REFINE -- takes useful blocks, cuts the trace, adds new predicates *************)
(*******************************************************************************************************)

let interpolant_refine extract_block conj_block_array reg_array (block_l : int list) = 
  Message.msg_string Message.Debug "Inside interpolant refine";
  (*b assert (not (block_l = [])); *)
  (* focimodelchecker -- piggybacking on this ! *)
  let len_reg_array = Array.length reg_array in
  let itp_array = Array.make len_reg_array Predicate.True in (* set the default ITP to be true *)
  let last_block = list_max block_l in
  let _ = Array.iteri (fun i _ -> if i > last_block then itp_array.(i) <- Predicate.False else ()) itp_array in
    (* end focimodelchecker *)
  let ord_block_l = List.sort compare block_l in
    
    let start = List.hd ord_block_l in
    let endp = Array.length conj_block_array - 1 in
    let cons_l = List.map extract_block ord_block_l in
    let end_cons = if (Options.getValueOfBool "fmcc") then conj_block_array.(endp) else Predicate.True
in    
    let start_cons = 
      if check_unsat (Predicate.And (cons_l@[end_cons])) then Predicate.True
      else conj_block_array.(start) 
    in	
      (* for debugging purposes, we shall dump the interesting blocks *)
(*
    let _ = 
      begin
	Message.msg_string Message.Debug (Printf.sprintf "num blocks = %d" (List.length block_l));
	Message.msg_string Message.Debug (Printf.sprintf "start cons = \n %s" (Predicate.toString start_cons));
	Message.msg_string Message.Debug (Printf.sprintf "end cons = \n %s" (Predicate.toString end_cons));
	Message.msg_string Message.Debug ("Constraints");
	List.iter (fun p -> Message.msg_string Message.Debug (Predicate.toString p) ) cons_l
	  
      end
    in
*)

    let accum_pred = ref [] in
      (* let proof = getFociProof (start_cons::cons_l) in *)
    let rec cut before_l after_l =
      (* INV: before_l + after_l = block_l *)
      if (after_l = [] && end_cons = Predicate.True) then () (* all is done *)
      else
        let (refine_begin,refine_end) =
	  match after_l with
	      [] -> (List.hd before_l + 1, endp)
	    | end_pos::tl -> if (before_l = []) then (start,start) else (List.hd before_l + 1, end_pos)
              	
	    in
	    let cons_l_pre = start_cons::(List.map extract_block before_l) 
	    and cons_l_post = end_cons ::(List.map extract_block after_l) in
	   	    let (foci_preds) = 
	      if (Options.getValueOfBool "fmcc") then []
	      else
		(* JHALA: temporary patch for segfaults! *)
		match (check_unsat (Predicate.And cons_l_pre)) || 
		  (check_unsat (Predicate.And cons_l_post)) with
		      true -> []
		    | false -> extract_foci_preds (cons_l_pre,cons_l_post) 
	    in 

	    let (foci_interpolant) = 
	      if (Options.getValueOfBool "fmcc") then
		match (check_unsat (Predicate.And cons_l_pre)) || 
		  (check_unsat (Predicate.And cons_l_post)) with
		      true -> Predicate.True (* hmmm.. how wise is this ? -- this should only be if the post is unsat ... *)
		    | false -> TheoremProver.extract_foci_interpolant (cons_l_pre,cons_l_post)
	      else Predicate.True
	    in
	     
	    let _ = Message.msg_string Message.Debug "extraction done" in
	    let _ = List.iter (fun i -> Message.msg_string Message.Debug (string_of_int i)) foci_preds
	    in
	    let _ = accum_pred := foci_preds @ !accum_pred in
	      (* loop begin *)
	      for i = refine_begin to refine_end do
		let _ = Message.msg_string Message.Debug ("Accessing i= "^(string_of_int i)) in
		(* focimodelchecker: store the interpolants now *)
		let _ = Array.set itp_array i foci_interpolant in 
		let _ = Message.msg_string Message.Debug ("Writing interpolant:"^(Predicate.toString foci_interpolant)) in
		(* end focimodelchecker *) 
		let (l_id(*,_*)) = C_SD.location_coords (location_coords_of_region (reg_array.(i))) in
		let _ = Message.msg_string Message.Debug (Printf.sprintf "loc_id (%d, %d)" (fst l_id) (snd l_id)) in
		let present_p_l = try Hashtbl.find PredTable.loc_pred_table l_id with Not_found -> [] in
		let to_add = 
		List.filter
		  (fun p_i -> not( List.mem p_i present_p_l)) foci_preds 
		     in 
let _ = Message.msg_string Message.Debug "Adding pred "; List.iter (fun p_i -> PredTable.block_a_t_flag := true; Message.msg_string Message.Debug (string_of_int p_i)) to_add
in
  Hashtbl.replace PredTable.loc_pred_table l_id (to_add@present_p_l)
              done; 
              (* recur on the remainder of the block-set *)
              match after_l with
		  [] -> ()
		| end_pos :: tl -> cut (end_pos::before_l) tl
   in
     let _ = cut [] ord_block_l in
     let _ = 
       if (Options.getValueOfString "checkRace" <> "") then
	 let _ = accum_pred := Misc.compact !accum_pred in
	 let (min,max) = (List.hd ord_block_l, List.hd (List.rev ord_block_l)) in
	 let _ = Message.msg_string Message.Normal "Adding all preds now..." in
	   for i = min to max do
	     Message.msg_string Message.Debug ("Accessing i= "^(string_of_int i));
	     
	     let (l_id(*,_*)) = C_SD.location_coords (location_coords_of_region (reg_array.(i))) in
	       PredTable.add_local_preds l_id !accum_pred;
	   (* so the "head" of the list is the min of the list, should also be the last in 
	      the list *)
	   done

     in
       if (Options.getValueOfBool "fmcc") then Some (itp_array) else None



(********************************************************************************************************)
(******************* The main refinement function for seq programs called by lmc*************************)
(********************************************************************************************************)

let refined_ops ops = 
  let proc op = 
    let sloc = C_SD.get_source op in
    let fn = C_SD.get_location_fname sloc in
    Hashtbl.replace refined_fun_table fn true
  in
  Hashtbl.clear refined_fun_table;
  List.iter proc ops


let build_block_at_constraint_arrays op_l reg_l pre_reg_l post_reg_l =
  (* returns block_array, conj_block_array *)
  let foci_mode = Options.getValueOfBool "fmc" in
  let ptrace_size = List.length op_l in
  let conj_block_array = Array.make ptrace_size Predicate.True in
  let block_array = Array.make ptrace_size Predicate.True in
  let posn = ref (ptrace_size - 1) in 
  let _ = Message.msg_string Message.Debug "block_a_t: building blocks start" in  
  let unsat_flag = ref false in
  let _ = block_reset () in
  let rec cons_folder opRegs preds =
    if (not (!unsat_flag)) then
      match (opRegs, preds) with
	  (((op, post_reg) :: opRest), (pre_pred :: predRest)) -> 
	    begin
	      Message.msg_string Message.Debug
		("cons_folder !posn = "^(string_of_int !posn));
   	      let atomic_post_reg = 
		match post_reg with
		    Region.Atomic atr -> atr
		  | _ -> failwith "bad region given to cons_folder!"
	      in
              let fname =
		match opRest with
		    (_, entry_reg)::_ -> reg_function_name entry_reg
		      | _ -> atomic_reg_function_name atomic_post_reg 
	      in
	      let block_cons = (block_concrete_data_pre fname 0 atomic_post_reg
				  (C_SD.get_command op) op)
	      in
	      let conj_block_cons = pred_lval_map 0 pre_pred in
	      let is_unsat =
                if foci_mode then false
                else
                  begin
                    block_assert block_cons;
                    block_assert conj_block_cons;
                    let rv = TheoremProver._is_contra () in
                    block_pop ();
                    rv
                  end
	      in
		Array.set block_array !posn block_cons;
		Array.set conj_block_array !posn conj_block_cons;
		posn := !posn - 1;
		if (is_unsat) then
		  Message.msg_string Message.Debug 
		    ("Found unsat at level"^(string_of_int (!posn +1)));
		if (is_unsat && not (Options.getValueOfInt "predH" >= 6))
		then unsat_flag := true;
		cons_folder opRest predRest 
	    end
	| _ -> ()
  in
    (* for cf -- well, actually whenever recursion is enabled, we need to preprocess the lvmap appropriately *)
   let _ =
     Stats.time "make lv_stack" (make_lv_stack (reg_function_name (List.hd reg_l))) op_l
   in
   let conj_l = List.map (fun r -> if foci_mode then Predicate.True else extract_pred r) pre_reg_l in
   let op_reg_l = try List.combine op_l (post_reg_l) with _ -> failwith "lists done have the same length" in
     (* RJ: lv_table is not used anywhere ..! *)
  let _ =
    if (Options.getValueOfBool "talias" || Options.getValueOfBool "fldunroll") then
    (Stats.time "load lv table" load_lv_table op_l)
  in
  let _ = Stats.time "Cons folder" (cons_folder (List.rev op_reg_l)) (List.rev conj_l) in
  let _ = block_reset () in
  let _ = 
    Message.msg_string Message.Debug "THE CONSTRAINTS ARE:\n";
    Array.iter
      (fun x -> (Message.msg_printer Message.Debug Predicate.print x))
      block_array;
    Message.msg_string Message.Debug "END CONSTRAINTS \n";
    Message.msg_string Message.Debug "block_a_t: building blocks end"
  in
  (block_array,conj_block_array)
  

let block_analyze_trace reg_l op_l _ =
  assert(List.length op_l > 0);
  print_string "block_analyze_trace\n";
  (* List.iter (fun op -> Message.msg_printer Message.Debug Operation.print op) op_l;*)
  let trace_size = List.length op_l in
  Message.msg_string Message.Normal ("In block_a_t:"^(string_of_int trace_size));
  let rec printer preregl opl postregl ctr = 
    match (preregl,opl,postregl) with
	(prereg::pre_tl,op::op_tl,postreg::post_tl) ->
	  Message.msg_string Message.Debug ("BNum: "^(string_of_int ctr));
	  Message.msg_string Message.Debug ("Pre reg: ");
	   Message.msg_printer Message.Debug Region.print prereg ;
	    Message.msg_string Message.Debug ("OP: ");
	  (Message.msg_printer Message.Debug Operation.print op);
	  Message.msg_string Message.Debug ("Post reg: ");
	  Message.msg_printer Message.Debug Region.print postreg ;
	    printer pre_tl op_tl post_tl (ctr+1)
      | ([],[],[]) -> ()
      | _ -> failwith "Bad list lengths in block_at"
  in
    try
    (* Phase 1: Load trace information into arrays *) 
    (* TRACE PROJECTION *)
  let orig_op_l = op_l in
  let orig_op_array = Array.of_list orig_op_l in (* needed in CF for left,right *)

  let (pre_reg_l,post_reg_l,op_l,revmap,fwdmap) =
    if (Options.getValueOfBool "tproj") then
      Stats.time "Trace Project" (trace_project op_l) reg_l 
    else (Misc.chop_last reg_l, List.tl reg_l, op_l,(fun x -> x),(fun x -> x))
  in
     (* INV: List.length pre_reg_l = List.length post_reg_l = List.length op_l *)
  let _ = Message.msg_string Message.Debug
	    (Printf.sprintf "Projected lengths: pre_reg_l %d post_reg_l %d op_l %d "
	       (List.length pre_reg_l)
	       (List.length post_reg_l)
	       (List.length op_l))
  in
    
  let _ = printer pre_reg_l op_l post_reg_l 0 in
  let _ = PredTable.block_a_t_reset () in
  let op_array = Array.of_list op_l in
  (* to check if any new info has been added *)
  (* Phase 2: we've set up the various arrays ... now to get the block constraints *)
  let (block_array,conj_block_array) = build_block_at_constraint_arrays op_l reg_l pre_reg_l post_reg_l in
  (* Phase 3: Constraints set up, the various arrays are loaded! Compute the useful ie contradictory blocks *)
    (* notice that here the last cell of conj_block_array is always true *)
  let (_ref_start_pos, _interesting_block_sets, suppress_stuff) =
    Stats.time "get useful blocks" (get_useful_blocks op_array block_array) conj_block_array 
  in
  let (_, _extract_block,_,_,_) =  suppress_stuff in
  (* TRACE PROJECT : now we shall go back into the "real" trace *)
  let interesting_block_sets = List.map (List.map revmap) _interesting_block_sets in 
  let _ =
    let printop i =
     let a = orig_op_array.(i) in
     let (_,_,op_s) = op_to_string a in
     Message.msg_string Message.Normal (Printf.sprintf "[INF] %d : %s" i op_s)
    in
    let print_bs l = 
      Message.msg_string Message.Normal "Conflicting Blocks";
      List.iter printop (List.sort compare l) 
    in
    List.iter print_bs interesting_block_sets
  in
  let ref_start_pos = revmap _ref_start_pos in
  let extract_block =
    fun i -> if fwdmap i = -1 then Predicate.True else (_extract_block (fwdmap i))
  in
    
  let _conj_block_array =
    if (Options.getValueOfBool "tproj") then
      begin
	let arr = Array.make trace_size Predicate.True in
	  Array.iteri (fun i a -> Array.set arr (revmap i) a) conj_block_array;
	  arr
      end
    else conj_block_array
  in
  let conj_block_array = _conj_block_array in
  let reg_array = Array.of_list reg_l in
    (* RJ: this is reqd for 1) oldschool focus, 2) to know what loc *)
    (* DONE setting up arrays for cutting *)
  let _ = Message.msg_string Message.Normal "[BAT] Done setting arrays for cutting" in
    (* more vicious hacks *)
  let _ = min_reg_id := (-1) in
  let _ = min_adr := None in
  let pre_flag = predicate_fixpoint_reached () in
  let _ = Message.msg_string Message.Debug ("#block sets = "^(string_of_int (List.length interesting_block_sets)))
  in
    (* Phase 4: refine using the blocks *)
  let refine_fun =
    if (Options.getValueOfBool "cf") then
	cf_interpolant_refine (Misc.paren_match Operation.paren_fun orig_op_array)
    else
      fun a b c d -> (interpolant_refine a b c d; ())
  in
  let refiner =   
    (match (Options.getValueOfInt "craig" <> 0)  with
	 true -> Misc.ignore_list_iter (refine_fun extract_block conj_block_array reg_array)
       | false ->  Misc.ignore_list_iter (focus_refine extract_block conj_block_array reg_array)
    )
  in
  let _ = Message.msg_string Message.Normal "[BAT] Calling refiner" in
  let _ = Stats.time "Interpolant Refine" refiner interesting_block_sets in
  let _ = Message.msg_string Message.Normal "[BAT] Done refiner" in
  let _ = Message.msg_string Message.Debug
	    (Printf.sprintf "Leaving block_a_t refine from %d" ref_start_pos)
  in
  let _ = block_reset () in 
    (* Phase 5: Check if we added something new *)
  let _ = bat_counter := !bat_counter + 1 in
  let _ = Message.msg_string Message.Normal ("Non-trivial functions"^(string_of_int !stats_nb_non_trivial_functions))
  in
  let cleanup_and_go () = 
    if (ref_start_pos = -1 || !PredTable.block_a_t_flag) then
      (* bonafide refine or error *)
      begin
       Hashtbl.clear cfb_local_refine_table; 
      (ref_start_pos, (if ref_start_pos < 0 then Region.bot else reg_array.(ref_start_pos)), true)
      end
    else if (Options.getValueOfInt "craig" >= 2) then
      begin	  
      if (Options.getValueOfBool "localrestart") then
        (print_lp_table ();printTableOfPreds (); dump_abs ();
        raise NoNewPredicatesException) 
      else 
        (Options.setValueOfBool "localrestart" true;dump_abs ();
         ((ref_start_pos), reg_array.(ref_start_pos),true))
      end
    else if (Options.getValueOfBool "cfb") then
      begin
        let fname = reg_function_name (reg_array.(ref_start_pos)) in
        if Hashtbl.mem cfb_local_refine_table fname then 
          raise NoNewPredicatesException
        else 
          (Hashtbl.replace cfb_local_refine_table fname true;
          raise (RefineFunctionException fname))
      end
    else if (Options.getValueOfBool "cf") then
       begin (* TBD:COMPLETENESS this is v. premature -- but for dfs its ok *)
       print_lp_table ();
       printTableOfPreds (); 
       dump_abs ();
       raise NoNewPredicatesException 
       end
     else
       begin
         Options.setValueOfBool "localrestart" false;
         dump_abs ();
         (ref_start_pos, 
         (if ref_start_pos < 0 then Region.bot else reg_array.(ref_start_pos)),
         true)
       end
  in
    Stats.time "cleanup and go" cleanup_and_go ()
 with UnskipException ref_start_pos ->
   begin
     (* we skipped an important function! *)
     block_reset ();
     Options.setValueOfBool "localrestart" false;
     dump_abs ();
     ((ref_start_pos), (if ref_start_pos < 0 then Region.bot else
			  List.nth reg_l ref_start_pos), true)
   end
      | e -> 
              begin 
               Message.msg_string Message.Error ("block_at hits unexpected expression"^(Printexc.to_string e));
               block_reset ();
               raise e
      end
      
(*** interpolation stuff end ****)    


(* Misc *)

  (* this next function is duplicated in cfAbstraction! *)

  let print_debug_info oplist = 
    let rec get_funs ops lst = 
      match ops with 
	  [] -> lst
	| op::rest ->
	    begin
              match (C_SD.get_command op).C_SD.Command.code with
		  C_SD.Command.FunctionCall fcexp -> 
		    begin
		      match fcexp with
			  Expression.FunctionCall(Expression.Lval (Expression.Symbol(fname)), _) -> 
			    if (C_SD.is_defined fname) then get_funs rest lst 
			    else get_funs rest (fname::lst)
			| Expression.Assignment (_, _, Expression.FunctionCall (Expression.Lval (Expression.Symbol fname), _)) -> 
			    if (C_SD.is_defined fname) then get_funs rest lst 
			    else get_funs rest (fname::lst)
			| _ -> get_funs rest (("FnPtr::"^(Expression.toString fcexp)) :: lst)
		    end
		| _ -> get_funs rest lst
	    end
    in
    let get_skip op = 
      match (C_SD.get_command op).C_SD.Command.code with
          C_SD.Command.FunctionCall fcexp -> 
	    begin
	      match fcexp with
		  Expression.FunctionCall(Expression.Lval (Expression.Symbol(fname)), _) 
		| Expression.Assignment (_, _, Expression.FunctionCall
					   (Expression.Lval (Expression.Symbol fname), _)) -> 
		    let (n,_,_) = C_SD.deconstructFunCall fcexp in
		      if n = __SKIPFunctionName then Some fname
		      else None
		| _ -> (*TBD: function pointer call *) None 
	    end
	| _ -> None
    in
      
    let impfun = get_funs oplist [] in
    let skipped_funs = Misc.compact (Misc.map_partial get_skip oplist) in
      if (List.length impfun > 0) then 
	begin
	  Message.msg_string Message.Normal "The following are suspect funs:\n";
	  Message.msg_string Message.Normal (Misc.strList impfun); 
	  Message.msg_string Message.Normal "The following are skipped funs:\n";
	  Message.msg_string Message.Normal (Misc.strList skipped_funs) 
	end

  (*******************************************************************************************)

let reset_decision_procedures () = 
    TheoremProver.kill_simplify_server ()

(* enabled_ops *)
(* This should really go to abs-region.ml, but it depends on
   skip_fun_process that is defined here --Rupak
*)
(* enabled_ops *)
  (* the function giving the possibly enabled operations *)
  let enabled_ops reg =
    let disambiguate_function_pointers edge =
      let (srcloc,op, targetloc) = (C_SD.get_source edge, C_SD.get_command edge, C_SD.get_target edge) in 
      match op.C_SD.Command.code with
        C_SD.Command.FunctionCall(e) ->
          begin
            let is_fnptr_call fcexp = match fcexp with Expression.Lval (Expression.Symbol _) -> false | _ -> true
            in
            let gen_eq fname fcexp = Predicate.Atom (Expression.Binary (Expression.Eq, Expression.Lval fname, fcexp)) in
            (* if this function call is a function pointer call, disambiguate it *)
            match e with
              Expression.FunctionCall(fcexp, args) -> 
                if is_fnptr_call fcexp then
                 begin
                  Message.msg_string Message.Normal ("Disambiguating " ^ (Expression.toString fcexp)) ;
                  (* edge :: *) (List.map (fun fname -> 
                    C_SD.make_edge srcloc targetloc
                      (C_SD.Command.gather_accesses 
                        (C_SD.Command.GuardedFunctionCall (gen_eq fname fcexp, Expression.FunctionCall(Expression.Lval fname, args)) )  ) )
                         (AliasAnalyzer.getFunctionsPointedTo fcexp))
                 end
                else [ edge ]
            | Expression.Assignment(aop, target, Expression.FunctionCall(fcexp, args)) ->
                if is_fnptr_call fcexp then
                 begin
                  Message.msg_string Message.Normal ("Disambiguating " ^ (Expression.toString fcexp)) ;
                  (* edge :: *) ( List.map (fun fname -> 
                    C_SD.make_edge srcloc targetloc
                            (C_SD.Command.gather_accesses 
                               (C_SD.Command.GuardedFunctionCall (gen_eq fname fcexp, 
                                           Expression.Assignment(aop, target, Expression.FunctionCall(Expression.Lval fname, args)) ) )))
                         (AliasAnalyzer.getFunctionsPointedTo fcexp) )
                 end
                else [ edge ]
            | _ -> failwith "enabled_ops: Internal invariant broken! Strange function call!" 
          end
      | _ ->  [ edge ]
      
    in
    let rv = 
      match reg with
          Region.Empty -> []
	| Region.Atomic atomic_reg ->
            if (Options.getValueOfBool "nofp" = false) then
              List.fold_right (fun thisedge sofar -> (disambiguate_function_pointers thisedge @ sofar))
                (C_SD.get_outgoing_edges atomic_reg.Region.location) []
            else (C_SD.get_outgoing_edges atomic_reg.Region.location)
	| Region.AtomicTI atomicTI_reg ->
            if (Options.getValueOfBool "nofp" = false) then
              List.fold_right (fun thisedge sofar -> (disambiguate_function_pointers thisedge @ sofar))
                (C_SD.get_outgoing_edges atomicTI_reg.Region.ti_location) []
            else (C_SD.get_outgoing_edges atomicTI_reg.Region.ti_location)
	| Region.ParallelAtomic _ -> failwith "Not implemented yet!"
	    
	| Region.Union atomic_reg_map ->
            failwith "enabled_ops: argument is a union: unimplemented"
	| Region.ParallelUnion p_atomic_reg_map ->
            failwith "enabled_ops: argument is a parallel union: unimplemented"
    in
    if not (Options.getValueOfBool "skipfun") then rv
    else
      List.map (skip_fun_process) rv
	



