(* This file has code for lvmaps, block_wp, block_sp, wp, sp,
   block_sp is only used by foci_model_checker, which also uses several lvmaps.
   Hence, only block_sp takes an lvmap parameter.
   The others just use the default lvmap.
*)

   (***************************************************************************)
   (*         weakest preconditions and strongest postconditions              *)
   (***************************************************************************)

(* A function to "simplify" address equality predicates *)
(* TBD: generalize this to a "unification" style procedure that spits out the
 * list outstanding obligations *)
let rec simplify_address_equality (a,a') =
  (* for now, just the array case *)
  match (a,a') with
   (Expression.Binary(Expression.Plus, e1,e2),
    Expression.Binary(Expression.Plus, e1',e2')) -> 
      if e1 = e1' then (e2,e2') else (a,a')
  |(Expression.Binary (Expression.FieldOffset,e1,f1),
    Expression.Binary (Expression.FieldOffset,e2,f2))->
      if f1 = f2 then simplify_address_equality (e1,e2) else (a,a') 
  | _ -> (a,a')
  
  



   (* compute WP for expressions involving pointers.  in particular, we have
      to consider aliasing if we are given an assignment statement.
      otherwise, we do the standard WP computation. 
      possibleAliases is the set of alias pairs that need to be considered for an
     assignment statement *)

  (* It does explicit substitutions when the flag is true.
     *)
 (* RJ: my head hangs in shame re: the compress-flag *)
 let compress_flag = ref false
 (* note that when wp is called with compress_flag set to true, it must be that
    explicit_flag is also true *)
 (* this is really only to be called when doing the interpolant thing *)

 (* pred_l is a list of predicates -- the first of which is the one where we will add stuff 
 while the others correspond to constraints further down the trace and so we merely substitute
 and so on in them. It returns a pred_list as well ... *)
 let make_exp_lval e = 
   match e with
       Expression.Lval l -> l
     | _ -> failwith ("bad call to make_exp_lval: "^(Expression.toString e))


 
(* THE NEXT TWO FUNCTIONS ARE NOT USED ANYWHERE! *)
 (* VVVVIMP : TBD make this trace check faster ! -- RJ: this is NOT used ANYWHERE right now *)
 let trace_lv_table = Hashtbl.create 109  (* this is actually a table of expressions *)
(* WARNING! this table is not actually READ anywhere ... *)

 let trace_field_table = Hashtbl.create 17 (* just stores the fields of lvs in the trace *)

 let is_relevant_field_trace fld =
   Hashtbl.mem trace_field_table fld
		
 let this_trace_field_table = Hashtbl.create 17 (* store the fields added BY THIS TRACE, FLUSHED at the end of the refinement *)
	   
 (* lval list -> unit *) 
 let add_to_lv_table candidates = 
    (*let candidates = Absutil.getVarsAndDerefs pred in*)
    Message.msg_string Message.Debug "Adding to lv_table:";
      List.iter (fun x ->
		   Message.msg_string Message.Debug (Expression.toString x);
		   Hashtbl.replace trace_lv_table x true) candidates;
      let c_fields = List.flatten (List.map (Expression.fields_of_expr) candidates) in
	Message.msg_string Message.Debug ("Adding Fields "^(Misc.strList c_fields));
	List.iter 
	  (fun fld ->
	     Hashtbl.replace trace_field_table fld true;
	     let added = C_SD.add_field_to_unroll true fld in
             if added then Hashtbl.replace this_trace_field_table fld true
	  ) c_fields

	
let lv_of_expr = Expression.lv_of_expr

(* this function has been moved to aliasAnalyzer.ml *)
let ac_get_lval_aliases a_flag lv =
  if a_flag then AliasAnalyzer.get_lval_aliases lv
  else [lv]

let get_lval_aliases lv = ac_get_lval_aliases (not (Options.getValueOfBool "incref")) lv

 
(* TBD:ARRAYS *)
let aliases_in_scope fname target alias_candidates =
  let relevant_scopes = (* TBD:CF+pointer?CHECK for now, lets just keep it simple *)
    Misc.cons_if fname
      (List.filter (fun s -> s <> "") (C_SD.scope_of_lval target))
  in
  let formals = match C_SD.lookup_formals fname with
      C_SD.Fixed args -> args
    | C_SD.Variable args -> args
  in
   let fml_table = Hashtbl.create 17 in
   let _ = List.iter (fun x -> Hashtbl.replace fml_table x true) formals in
   let _ = Message.msg_string Message.Debug ("Formals : "^(Misc.strList formals)) in
   let f_conv lv =
     match lv with
	 Expression.Symbol x ->
	   let lv' =
	     if (Hashtbl.mem fml_table x) && not (Options.getValueOfBool "incref")
	     then (Expression.make_symvar fname lv) else lv in
	     Expression.Lval(lv')
       | _ -> Expression.Lval lv
   in
   let proc_cand lv = 
     if not (C_SD.is_in_scope fname lv)
       (*(List.exists (fun fn -> C_SD.is_in_scope fn lv) relevant_scopes) *)
     then []
     else
       let lv' = lv_of_expr (Expression.deep_alpha_convert f_conv (Expression.Lval lv)) in
	 if lv' = lv then [lv]
	 else [lv;lv'] (* RJ:CF+pointer -- this isn't quite whats in the popl-paper.. or isit ? *)
   in
   let rv = List.flatten (List.map proc_cand alias_candidates) in
     Message.msg_string Message.Debug
       (Printf.sprintf "Aliases in scope of %s : %s"
	  (Expression.lvalToString target) (Misc.strList (List.map (Expression.lvalToString) rv)));
	rv


	  
(************************************************************ ***************************************************)
(************************************************ LVALUE MAPS ***************************************************)
(************************************************************ ***************************************************)
			 
module LvalMap = Map.Make(struct
  type t = int * Expression.lval
  let compare (i1,lv1) (i2,lv2) =
    if (compare i1 i2) = 0
    then
      compare (Expression.lvalToString lv1) (Expression.lvalToString lv2)
    else compare i1 i2
end)


type base_lvmap_t = Expression.expression LvalMap.t

type lvmap_t = (base_lvmap_t * (base_lvmap_t list) * (int list)) ref

let lvmap_empty = (LvalMap.empty, [], [0])

let frame = ref 0
let get_new_frame () =
  frame := !frame + 1;
  !frame

let make_lval_tuple tid lval = 
   if C_SD.is_global lval then (0,lval) else (tid,lval)
(* RJ: I shall nevertheless call them lvmap_ref to rub in the fact that they are pointers *)		    
let update_map_a_b new_var_fun lvmap_ref tid lval =
  let (cur,hist,frames) = !lvmap_ref in
  let lval_tup = make_lval_tuple tid lval in
  let new_lval = new_var_fun lval in
    (*
      Message.msg_string Message.Normal (Printf.sprintf "update_map: %s -> %s" (Expression.lvalToString lval) (Expression.toString new_lval));
    *) (* Hashtbl.add map_table lval_tup new_lval *)
    lvmap_ref := (LvalMap.add lval_tup new_lval cur, hist ,frames)
      
      
let update_map_a = update_map_a_b Absutil.getNewTmpVar
let update_map_clock_a = update_map_a_b Absutil.getNewTmpVar_clock


let in_lval_map_a lvmap_ref tid lval = 
  let (cur,hist,frames) = !lvmap_ref in
  let lval_tup = make_lval_tuple tid lval in
  LvalMap.mem lval_tup cur

let _lval_map_a lvmap_ref tid lval =
  let (cur,hist, frames) = !lvmap_ref in
  let lval_tup = make_lval_tuple tid lval in
    try
      LvalMap.find lval_tup cur
    with Not_found -> 
      begin (* TBD:WHY? we need this to be deterministic -- WHY ? -- is it to do with "clock" ? *)
	let id =
	  if (Options.getValueOfBool "lvframe") then
	    try
	      (string_of_int (List.hd frames))
	    with _ -> failwith "BAD FRAMES!"
	  else "0"
	in
	let new_lval = Expression.Chlval (lval,id) in
	  lvmap_ref := (LvalMap.add lval_tup new_lval cur, hist,frames);
	  new_lval
      end

let lval_map_a lvmap_ref tid e =
  Expression.address_normalize 
    (Expression.deep_alpha_convert (_lval_map_a lvmap_ref tid) e)


let pred_lval_map_a lvmap_ref tid pred =
  let p' = (Predicate.deep_alpha_convert (_lval_map_a lvmap_ref tid ) pred) in
  Message.msg_string Message.Debug "plvmap: now normalize";
  Predicate.address_normalize p'
    
let new_lvmap_t tid = 
  let rv = ref lvmap_empty in
  let upd_f = if Options.getValueOfBool "clock" then update_map_clock_a rv tid else update_map_a rv tid in
  let dummy_lookup lv = ignore (_lval_map_a rv tid lv) in
  C_SD.iterate_all_lvals dummy_lookup; (* adds all the lvals of the program to this table *)
    Absutil.clock_tick ();
    rv

let default_lvmap_ref = ref lvmap_empty

let lvmap_reset () =
  default_lvmap_ref := lvmap_empty
			  
(* NOT NEEDED ANYWHERE !*)
(* let _lval_map = _lval_map_a default_lvmap_ref *)

let update_map = update_map_a default_lvmap_ref
let update_map_clock = update_map_clock_a default_lvmap_ref
			  
let lval_map = lval_map_a default_lvmap_ref 
			
let in_lval_map = in_lval_map_a default_lvmap_ref

let pred_lval_map = pred_lval_map_a default_lvmap_ref


(** copy_globals m_s m_d returns m' s.t. m'.lv = (lv \in (m_s \cap globals)) ? m_s.lv : m_d.lv *) 		      
let copy_globals map_src map_dst =
  let folder lv _ m' =
    (* INV: lv in map_src *)
    try
      if C_SD.is_global (snd lv) then
	LvalMap.add lv (LvalMap.find lv map_src) m' 
      else m'
    with _ -> failwith "INV smashed in copy globals"
  in
    LvalMap.fold folder map_src map_dst
      
let lvmap_push_a lvmap_ref =
  if (Options.getValueOfBool "cf" || Options.getValueOfBool "enablerecursion")
  then
    let (curr,hist,frames) = !lvmap_ref in
    let new_map = copy_globals curr LvalMap.empty in
      lvmap_ref := (new_map,curr::hist,((get_new_frame ())::frames))

let lvmap_pop_a lvmap_ref =
  if (Options.getValueOfBool "cf" || Options.getValueOfBool "enablerecursion")
  then
    match !lvmap_ref with
	(curr, top::rest, _ ::rest_fr) ->
	  let new_curr = copy_globals curr top in
	    lvmap_ref := (new_curr,rest,rest_fr)
      | _ -> failwith "Empty lvmap history in lvmap_pop"
	  
(** ADAMS's old code -- don't delete for reference -- but its quite messy. 		      
(* TBD:SOUNDNESS now only used in CF mode, in general, should be used EVERYWHERE --  *)
let lvmap_push_a lvmap_ref = 
  if (Options.getValueOfBool  "cf" || Options.getValueOfBool "enablerecursion")
  then 
     let (cur,hist,frames) = !lvmap_ref in
       lvmap_ref := (cur, cur::hist,((get_new_frame ())::frames))
  
let lvmap_pop_a lvmap_ref =
if (Options.getValueOfBool "cf" || Options.getValueOfBool "enablerecursion")
then
  match !lvmap_ref with
      (cur, top::rest, t_fr::rest_fr) ->
	let isGlobal =
	  function
	    Ast.Expression.Symbol sym ->
	      (try let _ = String.index (Ast.Symbol.toString sym) '@' in false
	       with Not_found -> true)
	  | l -> 
	      begin
		Message.msg_string Message.Debug ("Warning: Unsupported lval in lvmap_pop: "^(Expression.lvalToString l));
		false
		  (*failwith ("Unsupported lval in lvmap_pop_a: "^(Expression.lvalToString l)) *)
 	      end
	in
	let retainGlobal lv lv' lvmap =
	  (*fprintf "retainGlobal called.@.";*)
	  let _lv = snd lv in
	  if isGlobal _lv then lvmap
	  else
	    (try LvalMap.add lv (LvalMap.find lv top) lvmap
	     with Not_found ->
	       Message.msg_string Message.Debug
	       ("lvmap_pop: removing "^(Expression.lvalToString _lv));
	       LvalMap.remove lv lvmap)
     in
	  (*fprintf "POPPING!@.";*)
	  lvmap_ref := (LvalMap.fold retainGlobal cur cur, rest, rest_fr)
 | _ -> failwith "Empty lvmap history in lvmap_pop"
 *)

	
let lvmap_push () = lvmap_push_a default_lvmap_ref      

let print_frames () =
  let (_,_,frames) = !default_lvmap_ref in
    Message.msg_string Message.Debug
      ("Frames: "^(Misc.strList (List.map (string_of_int) frames)))



let lvmap_pop () =
      lvmap_pop_a default_lvmap_ref
		     
let lvmap_copy lvmap_ref =
  ref (!lvmap_ref)

    
(* TBD:CF+POINTERS -- this whole business about lvmap push/pop rename needs to  be looked at *)
let lvmap_rename_locals first_push_flag upd fname  =
  if (Options.getValueOfBool "cf" || Options.getValueOfBool "enablerecursion")
  then
    if (Options.getValueOfBool "lvframe") then 
      lvmap_push ()
    else
      begin
	Message.msg_string Message.Debug (Printf.sprintf "Tweak locals of %s.@." fname);
	let all_local_lvs = Stats.time "lookup locals" C_SD.lookup_local_lvals fname in
	let _ = Message.msg_string Message.Debug
		  ("lvmap_rename_locals: "^
		   (Misc.strList (List.map Expression.lvalToString all_local_lvs)))
	in
	  if first_push_flag then
	    begin
	      lvmap_push ();
	    List.iter upd all_local_lvs
	    end
	  else
	    begin
	      List.iter upd all_local_lvs;
	      lvmap_push ()
	    end
      end

(** here f: ( int * lval) -> expr -> unit *)    
let lvmap_iter f lvmap_ref =
  LvalMap.iter f (Misc.fst3 !default_lvmap_ref)

let print_lvmap () =
  let _pr (_,lv) e =
    Message.msg_string Message.Debug
      (Printf.sprintf "%s ----> %s" (Expression.lvalToString lv) (Expression.toString e))
  in
    lvmap_iter _pr default_lvmap_ref



let make_lv_stack start_fname op_l =
  let stack_fold stack op =
    match Operation.get_info op (* (C_SD.edge_to_command op).C_SD.Command.code *) with
	Operation.Ret _ -> 
	  begin
	    Message.msg_string Message.Debug  "Return\n";
	      List.tl stack
	  end
      | Operation.Call (_, _, fname,_) ->
	  begin
	    Message.msg_string Message.Debug (Printf.sprintf "Call %s\n" fname);
	    fname :: stack
	  end
      | Operation.Normal ->  stack
  in
  let stack = List.fold_left stack_fold [start_fname] (* [reg_function_name (List.hd reg_l)] *) op_l in
    (* We have to find the stack state at the error to prepare the lvmap correctly with
       * initial renamings of local variables. *)
  let update_map =
    if (Options.getValueOfBool "clock") then update_map_clock 0 else update_map 0
  in
    lvmap_reset ();
    if (Options.getValueOfBool "cf" || Options.getValueOfBool "enablerecursion") then
      List.fold_right (fun fn () -> lvmap_rename_locals false update_map fn) stack () 
      


      
(* only functions used are:
   ...
1. lvmap_push
   2. lvmap_pop
   3. lvmap_rename_locals
   4. update_map/_clock
   5. (pred_)lval_map
   6. lvmap_empty
   7. lvmap_copy
   8. lvmap_rename_locals -- NOT REALLY used -- should be suppressed.
   9. lvmap_iter
   10. make_lv_stack
*)
    
(********************************* END OF LVMAP MODULE ****************************************)
(**********************************************************************************************)
    

(* Misc. predicate helpers for block_wp/sp -- see popl04 paper *)

let ite (cond,t1,t2) = 
  Predicate.disjoinL 
    [ Predicate.conjoinL [ cond;t1] ; 
      Predicate.conjoinL [Predicate.Not cond; t2]]
    

       
(* RJ: THERE CAN BE BUGS WITH THE LVAL_MAP INTERFACE !! PLEASE CHECK IF YOU GET WIERD ANSWERS THAT THE CONSTRAINTS GENERATED 
   ARE GOOD ! -- better still PROVE this interface correct. PROVE ???!!! HA HA.*)


let trace_lv_table_get_lval_aliases =
  let trace_lv_table_iter f =
    let _iter e _ =
      try
	let lv = lv_of_expr e in
	  f lv
      with _ -> Message.msg_string Message.Major
	("Warning: tlv_get_alias skipping "^(Expression.toString e))
    in
    Hashtbl.iter _iter trace_lv_table
  in
    AliasAnalyzer.get_lval_aliases_iter Misc.traceLvals_t_c trace_lv_table_iter 
  

       
(* TBD:SOUNDNESS -- in cf mode, only use lvals IN SCOPE! not all lvals *)
(* fname = name of function to which e belongs *)
              
let block_wp lvmap_operation_option fname tid e_list pred =
  let _ = print_frames () in
  (* to deal with multiple threads *)
  let update_map = if (Options.getValueOfBool "clock") then update_map_clock tid else update_map tid
  in (* RJ: this used to be update_map *)
  let lval_map =  lval_map tid in (* i.e. we use the default lvmap *)
  let doAssign_list targets_fname (target_valu_list) = 
    (* make the clock tick!! *)
    let _ = Absutil.clock_tick () in
    let num = List.length target_valu_list in
      (* first, lets "close" the assignments under must aliasing ... moohaha MUST ALIASING in ONE LINE !! *)
    let must_close (target,valu) =
      let must_aliases = AliasAnalyzer.get_all_must_aliases target in
      let _ = Message.msg_string Message.Debug 
		(Printf.sprintf "Must aliases of %s are %s" 
		   (Expression.lvalToString target) 
		   (Misc.strList (List.map Expression.lvalToString must_aliases)))
      in
	List.map (fun lv -> (lv,valu)) (target::must_aliases)
    in
    let target_valu_list =
      List.flatten (List.map must_close target_valu_list)
    in (* thats it. *)
    let target_valu_list =
      List.map (fun (target,valu) -> (Expression.push_deref target,valu))
	target_valu_list
    in  
    let (target_list,valu_list) = Misc.unzip target_valu_list in
      
    
      (* shouldnt really need to do this latter operation as its subsumed by the one later but ... *)
    
    let wp_get_lval_aliases =
      if (Options.getValueOfBool "talias") then trace_lv_table_get_lval_aliases
	(* occurring in the trace *)
      else get_lval_aliases
	(* in the whole program *)
    in
      (* this is only used for refinement so we shall just use the fields in the trace_lv_table *)
    let wp_closure_fun =
      if (Options.getValueOfBool "fldunroll") then
	C_SD.expression_closure_stamp true (* (is_relevant_field_trace) *)
      else C_SD.expression_closure
    in
    let possibleAliases_list =
      List.map
	(fun (target,_) ->
	   if (Options.getValueOfString "alias" = "" )
	   then [target]
	   else
             let candidates = List.map Expression.push_deref (wp_get_lval_aliases target) in
	     let _ = Message.msg_string Message.Debug
		       ("Alias pushed candidates: "^(Misc.strList (List.map Expression.lvalToString candidates)))
	     in
	       (* only lvals in the same "scope"  ?? *)
	       if (Options.getValueOfBool "cf")
	       then
		 Misc.cons_if target (aliases_in_scope targets_fname target candidates)
		 else candidates
	)
	target_valu_list
    in
    let lv_triple_list_list = 
      let make_triple lv = 
	let eclos = wp_closure_fun (Expression.Lval lv) in
	  (lv,eclos,List.map lval_map eclos)
	  (* RJ: the 3rd co-ord is the names of lvalues _AFTER_ assignment *)
      in
      List.map (List.map make_triple) possibleAliases_list
    in
      (* implicit INV:possibleAliases have the same closure set as the target *)
      (* FALSE!!! The aliases may have a "smaller" closure, due to heuristics in
       * closure that prevent unrolling a field if it already is in the prefix!! *)
      (* now updating the map so that all the things that might possibly have changed have new values for 
	 _BEFORE_ the operation *)
    let relevant_lvals =
      List.flatten
	(List.map (fun l -> List.flatten (List.map Misc.snd3 l)) lv_triple_list_list)
    in
      (*let old_map = copy_lval_map_table () in*)	
    let _ = List.iter (fun x -> update_map (lv_of_expr x)) relevant_lvals in
      (* after this we are using the updated lval_map *) 
    let _ =
      match lvmap_operation_option with
	  None -> ()
	| Some (f) -> f ()
    in
      (* USED to be: if push then lvmap_rename_locals true update_map fname *)
      (* change to -- do whatever must be done between reading/writing --
	 can be:
	 lvmap_push -- if you are processing a return,
	 lvmap_pop -- if these are assignments of a funcall,
	 (fun () -> ()) otherwise
      *)
      (* push indicates that this assignment is saving a return value.
       * We'll save the renaming AFTER changing names for aliases of the target but
       * BEFORE renaming local variables of the function we return from, to make
       * sure that these local variable renamings go out of scope outside this
       * function call. TBD:SOUNDNESS -- do it for *)
    
      (* implicit invariant, the closures come out in the right order ... *)
    let _mapper (target,valu) lv_triple_list =
      let valu_closure_vals =
	let valu_closure =
	  let target_closure = wp_closure_fun (Expression.Lval target) in
	    List.map 
	      (fun x ->
		 let vx = 
		   (Expression.substituteExpression
		      [(Expression.Lval target, valu)] x)
		 in 
		   try Expression.Lval (Expression.push_deref (lv_of_expr vx))
		   with _ -> vx)
	      target_closure
	in
	  List.map lval_map valu_closure
	    (* note: the above is done _AFTER_ the lval_map has been updated *)
      in
      let target_addr = Expression.addressOf (lval_map (Expression.Lval target)) in
	(* RJ: this order is VERY important. First substitute, THEN take addressOf *)
      let process_alias (lv,lv_closure,new_vals) = 
	if lv = target then
	  try Predicate.min_equate new_vals valu_closure_vals with excn -> failwith ("eq1: "^(Printexc.to_string excn))
	else
	  begin
	    let old_vals = List.map lval_map lv_closure in
	    let eq_new = 
              try Predicate.min_equate new_vals valu_closure_vals with excn -> failwith ("eq2: "^(Printexc.to_string excn))
            in
	    let eq_old = try Predicate.min_equate new_vals old_vals with excn ->
              failwith ("eq3: "^(Printexc.to_string excn))
            in
	    let lv_addr = Expression.addressOf (lval_map (Expression.Lval lv)) in
	    let cond =
              let (a,a') = simplify_address_equality (target_addr,lv_addr) in
	      Predicate.Atom (Expression.Binary(Expression.Eq,a,a')) 	    
            in
	      ite (cond,eq_new,eq_old)
	  end
      in
      let new_cons = List.map process_alias lv_triple_list in
	new_cons
    in
    let ac_new_cons = List.flatten (List.map2 _mapper target_valu_list lv_triple_list_list) in
      ac_new_cons
  in
  let doReturn retexp =
    Message.msg_string Message.Debug "WP of return";
    let (return_asgn_option,copy_back_asgns) = Absutil.asgn_of_return fname  in
      (* copy_back_asgns is [] if not "cf" *)
      (* TBD:ARRAYS
	 actually, you have to make a pass over the code and see what cells may have been modified.
 	 if this includes things like arrays, this DOES NOT WORK...
	  unless, you repace the arrays wholesale... but this is not sound, as you may create
	  sharing where there is none... sigh. Ideally, you have to know which lvals have been "hit"
	 inside the function call, and copyback all those lvals at the return site.
	 For now, we assume we only have to copy back the lvals passed in as args and their transitive derefs.
	 One has to use trace_lv_table -- to copy back all the fellows that may be looked at in the future.
	 One has to think about this...
      *)
    let all_ret_asgn =
      match return_asgn_option with
	  Some (return_target) ->
	    Message.msg_string Message.Debug
		(Printf.sprintf "Return: %s <- %s"
		   (Expression.lvalToString return_target) (Expression.toString retexp));
	    (return_target,retexp)::copy_back_asgns
	| None -> copy_back_asgns
    in
      doAssign_list  (!Absutil.retCallerName) all_ret_asgn
	(* CONCURRENT ASSIGNMENT *)     
  in
  let new_cons =
    match e_list with
	[C_SD.Command.Expr (Expression.Assignment(Expression.Assign, target, valu))] ->
	  doAssign_list fname [(target,valu)] (* target "belongs" to fname *)
	  (* INV: in this case the lvmap_oper_option = None *)
      | [C_SD.Command.Return exp] -> doReturn exp
	  (* INV: in this case the lvmap_oper_option = Some(lvmap_push) *)
      | _ ->
	  (* INV: in this case the lvmap_oper_option = Some(lvmap_pop) *)
	  begin
	  try
	    let extract_tv e =
	      match e with
		  C_SD.Command.Expr (Expression.Assignment(Expression.Assign, target, valu))
		    -> (target,valu)
		| _ -> failwith "bad call to compressed_wp --extract_tv"
	    in
	    let t_v_list = List.map extract_tv e_list in
	      doAssign_list fname t_v_list
		
	  with _ -> failwith("bad call to compressed_wp!")
	  end
  in
    Predicate.conjoinL (new_cons@[pred])


 (* Weakest Preconditions. *)   
 let rec wp explicit_flag sat_flag stmt pred (* arg  (edge,i) *) =
   match stmt with
       C_SD.Command.Expr e->
	 begin
	   match e with
	       (* Jhala : am restructuring the wp code *)
	       Expression.Assignment(Expression.Assign, target, valu) -> (* operation target = value *)
		 let assigned_value = 
		   if explicit_flag then (* EXPLICIT *)
		     begin
		       Message.msg_string Message.Debug "explicit wp \n";
		       let newvar = (Absutil.getNewTmpVar target) in
			 newvar
		     end
		   else (* IMPLICIT *)
		     begin
		       Message.msg_string Message.Debug "implicit wp \n";
		       valu
		     end
		 in
		 (* is the value assigned a boolean expression? *)
		 let is_boolexp = match valu with Expression.Binary (op, _, _) when Expression.isRelOp op -> true | _ -> false
		 in
		 let _print_info_flag = (!wpCapRegion = Predicate.True) in
		 let possibleAliases = Stats.time "getPossibleAliases" (Absutil.getPossibleAliases target) pred in
		   let aliasSets = Misc.powerSet (List.filter (function al -> al <> (Expression.Lval target)) possibleAliases)
		   in
		   let wp_one aliasSet =
		     (* In the implicit version, we would construct substitutions here.
			But here we should not substitute: the aliasExps will take care of that.
		     *)
		     begin
		       let aliasExps = List.map 
			   (function (b,e) ->
			     let targetAddr = Expression.addressOf (Expression.Lval target)
			     and eAddr = Expression.addressOf e in 
			     let (a,a') = simplify_address_equality (targetAddr,eAddr) in
			     match b with 
			       true -> Predicate.Atom (Expression.Binary(Expression.Eq,a ,a'))
			     | false -> Predicate.Atom (Expression.Binary(Expression.Ne,a ,a'))) 
			   aliasSet 
		       in

			   (* compute the weakest precondition *)
			   (* without pointers, the following line would be the entire
			      weakest precondition computation, with an empty substSet *)
		       if is_boolexp then (* the value being assigned is a boolean expression *)
			 begin
			   if explicit_flag then 
			     begin
			       let substSet = 
				 List.map (function (_,e) -> (e,assigned_value))
				   (List.filter (function (b,e) -> b=true) aliasSet) in

			       let _s1 =
				 Predicate.substitute ( (Expression.Lval target, 
							 assigned_value )
						       :: substSet) pred in
Message.msg_string Message.Normal (Predicate.toString _s1) ;

			       Predicate.conjoinL (
						     (Predicate.Or
							[(Predicate.conjoinL [ Predicate.Atom valu; 
									      Predicate.Atom (Expression.Binary (Expression.Eq, assigned_value, 
														 Expression.Constant (Constant.Int 1)))]
							    );
							  (Predicate.conjoinL [ Predicate.Atom (Expression.negateRel valu); 
										Predicate.Atom (Expression.Binary (Expression.Eq, assigned_value, 
														   Expression.Constant (Constant.Int 0)))])
							]
							) :: (_s1 :: aliasExps))
			     end
			   else
			     begin
			       let substSet1 = 
				 List.map (function (_,e) -> (e, Expression.Constant (Constant.Int 1)))
				   (List.filter (function (b,e) -> b=true) aliasSet) in
			       let substSet0 = 
				 List.map (function (_,e) -> (e,Expression.Constant (Constant.Int 0)))
				   (List.filter (function (b,e) -> b=true) aliasSet) in

			       let _s1 =
				 Predicate.substitute ( (Expression.Lval target, 
							 Expression.Constant (Constant.Int 1)) 
						       :: substSet1) pred in

			       let _s0 =
				 Predicate.substitute ( (Expression.Lval target, 
							 Expression.Constant (Constant.Int 0)) 
						       :: substSet0) pred in
Message.msg_string Message.Normal (Predicate.toString _s1) ;
Message.msg_string Message.Normal (Predicate.toString _s0) ;
			       Predicate.conjoinL ( 
			       (Predicate.disjoinL
				  [
				Predicate.conjoinL [Predicate.Atom valu ; _s1 ] ;
				Predicate.conjoinL [Predicate.Atom (Expression.negateRel valu); _s0 ]
			      ] )
			       :: aliasExps)
			     end

			 end
		       else 
			 let substSet = 
			   List.map (function (_,e) -> (e,assigned_value))
			     (List.filter (function (b,e) -> b=true) aliasSet) in

			 let _s1 = 
			   Predicate.substitute ((Expression.Lval target, assigned_value) :: substSet) pred in
			   (*	if (not explicit_flag) then Message.msg_string Message.Normal ("_s1 = "^(Predicate.toString _s1)) ; *)
			   (* conjoin the weakest precondition result to the expressions
			      representing this particular aliasing scenario *)
			 Predicate.conjoinL (List.append (_s1:: aliasExps) 
					     (if explicit_flag then
					       [Predicate.Atom (Expression.Binary
								  (Expression.Eq,
								   assigned_value, valu))]
					     else [])
					       )
		     end
		   in
		   let actual_pre = (Predicate.disjoinL (List.map wp_one aliasSets)) in
		     (*
		       if (not explicit_flag) then 
		       Message.msg_string Message.Normal ("Actual pre = "^(Predicate.toString actual_pre)) ;
		     *)
		     if sat_flag then 
		       sat_filter (actual_pre) Predicate.True
		     else 
		       actual_pre

	     |        Expression.Unary(Expression.Assume, e) -> Predicate.conjoinL [(Predicate.Atom e); pred]
	     |        _ -> failwith "wp: not implemented"
	 end
     | C_SD.Command.Return exp ->
	 Message.msg_string Message.Debug "WP of return";
	 if ((!retTarget) <> C_SD.junkRet) then
	   begin
	     Message.msg_printer Message.Debug C_SD.Command.print { C_SD.Command.code = C_SD.Command.Block [(C_SD.Command.Expr(Expression.Assignment(Expression.Assign, !retTarget, exp)))];
								    C_SD.Command.read = []; (* printing ignores .read and .written *)
								    C_SD.Command.written = []}; (* ditto *)
	     (wp explicit_flag sat_flag (C_SD.Command.Expr(Expression.Assignment(Expression.Assign, !retTarget, exp))) pred (* arg *))
	   end
	 else pred
	   (* Jhala: wp RETURN, here. Devilish hack. Sorry ! *)

   (*
     Strongest postcondition
   *)
	   (* Jhala: Need to fold in rupaks changes to SP *)
   (*
     Strongest postcondition
     We do not need an explicit flag because sp is always explicit
   *)
   let rec sp sat_flag stmt pred =
     match stmt with
	 C_SD.Command.Expr e ->
	   begin
	     match e with  
		 Expression.Assignment(Expression.Assign, target, value) ->
		   begin 
		     let targetExp = Expression.Lval target 
		     in
		     let possibleAliases = Absutil.getPossibleAliases target pred in
		       (* compute all possible alias scenarios.  we can remove the case when the target
			  appears in the pred, because these are always aliased (and there is no point of
			  doing the usual wp substitution in this case). *)
		     let aliasSets = Misc.powerSet (List.filter (function al -> al <> (Expression.Lval target)) possibleAliases) in
		     let newvar = (Absutil.getNewTmpVar target) in
		     let sp_one aliasSet =
		       (* create the pairs for substituting expressions in pred for
			  value, when pred and target are aliased *)
		       let substSet = List.map (function (_,e) -> (e,newvar))
					(List.filter (function (b,e) -> b=true) aliasSet) in
			 (* create expressions representing the alias pairs *)
		       let aliasExps = List.map 
					 (function (b,e) ->
					    let targetAddr = Expression.addressOf (Expression.Lval target)
					    and eAddr = Expression.addressOf e in
                                            let (targetAddr,eAddr) =
                                              simplify_address_equality
                                              (targetAddr,eAddr)
                                            in
					      match b with 
						  true -> Predicate.Atom (Expression.Binary(Expression.Eq,targetAddr,eAddr))
						| false -> Predicate.Atom (Expression.Binary(Expression.Ne,targetAddr,eAddr))) aliasSet 
		       in
			 (* compute the strongest postcondition *)
			 (* without pointers, the following line would be the entire
			    strongest postcondition computation, with an empty substSet *)
		       let s1 = Predicate.substitute ((targetExp, newvar) :: substSet) pred in
		       let s2 = Expression.substituteExpression [(targetExp, newvar)] value in
			 Predicate.conjoinL (List.append 
					       [s1; Predicate.Atom (Expression.Binary (Expression.Eq, targetExp, s2))] 
					       aliasExps)
		     in
		     let actual_sp =  Predicate.disjoinL (List.map sp_one aliasSets) in
		     if sat_flag then
		       sat_filter (actual_sp) Predicate.True else actual_sp
		   end
	       |        Expression.Unary (Expression.Assume, e) -> (* Predicate.normalize *)
		     (Predicate.And [(Predicate.Atom e); pred])
	       |        _ -> failwith "sp: not implemented"
	   end
       |  C_SD.Command.Return exp ->
	   if ((!retTarget) <> C_SD.junkRet) then
	     begin
	       sp sat_flag (C_SD.Command.Expr (Expression.Assignment (Expression.Assign, !retTarget, exp))) pred
	     end
	   else
	     pred

(************************************************************************************************************************)
(***********************************  block_sp, for the foci_model_checker **********************************************)
(************************************************************************************************************************)


(* Calling this changes the lvmap *)
(* the first param is a function that should return an overapproximaiton of the aliases of an lv, w.r.t. whatever
   is of interest -- you could return, for ex. either ALL aliases of lv, or those aliases that occur in the future,
   or those that occur in a predicate etc etc.
*)

let rec ac_block_sp closure_fun aliases_of_fun lvmap_ref retTarget_o e pred =
  let update_map = if (Options.getValueOfBool "clock") then update_map_clock_a lvmap_ref 0 else update_map_a lvmap_ref 0
  in (* RJ: this used to be: let update_map = update_map_a lvmap_ref 0 in *)
  let lval_map =  lval_map_a lvmap_ref 0 in
    match e with
	 C_SD.Command.Expr (Expression.Assignment(Expression.Assign, target, valu)) ->
	   begin
	     let _ = Absutil.clock_tick () in
	     (* i'm doing this again for no good reason *)
	     let target_closure = closure_fun (Expression.Lval target) in
	     let valu_closure = 
	       List.map 
		 (
		   fun x -> 
		     let vx = 
		       (Expression.substituteExpression [(Expression.Lval target, valu)] x)
		     in 
		       try Expression.Lval (Expression.push_deref (lv_of_expr vx))
		       with _ -> vx
		 )
		 target_closure
	     in
	       (* shouldnt really need to do this latter operation as its subsumed by the one later but ... *)

	     let possibleAliases = List.map Expression.push_deref (aliases_of_fun target) in
	       (* 1. store the old NAMES *)
	     let lv_triple_list = 
	       let make_triple lv = 
		 let eclos = closure_fun (Expression.Lval lv) in
		 let lv_addr = lval_map (Expression.addressOf (Expression.Lval lv)) in
		   ((lv,lv_addr),eclos,List.map lval_map eclos) (* these are names BEFORE the operation *)
	       in
		 List.map make_triple possibleAliases
	     in
	       (* 2. store the old VALUES *)
	     let valu_closure_vals = List.map lval_map valu_closure in 
	     let target' = lval_map (Expression.Lval target) in
	     let target_addr = Expression.addressOf target' in
	       (* this is the OLD name of target -- i.e. BEFORE the operation *)
	       (* implicit INVARIANT: the possibleAliases have the same closure set as the target *)

	       (* 3. now updating the map for things that may have changed *)
	     let relevant_lvals = List.flatten (List.map Misc.snd3 lv_triple_list) in
	     let _ = List.iter (fun x -> update_map (lv_of_expr x)) relevant_lvals in
	       
	     (* after this we are using the updated lval_map *)
	     (* implicit INVARIANT: the closures come out in the right order ... *)

	     let process_alias ((lv,lv_addr),lv_closure,old_vals) = 
	       let new_vals = List.map lval_map lv_closure in
               if lv = target then
		   try Predicate.equate new_vals valu_closure_vals with _ -> failwith "pa 1" 
               else
		 begin
		   let eq_new = try Predicate.equate new_vals valu_closure_vals with _ -> failwith "pa 2" in
		   let eq_old = try Predicate.equate new_vals old_vals with _ -> failwith "pa 3" in
		   let lv_addr = Expression.addressOf  (lval_map (Expression.Lval lv)) in
		   let cond =
                     let target_addr,lv_addr = simplify_address_equality (target_addr,lv_addr) in
		     Predicate.Atom (Expression.Binary(Expression.Eq,target_addr,lv_addr))
		   in
		     ite (cond,eq_new,eq_old)
		 end
	     in
	     let new_cons = 
	       List.map process_alias lv_triple_list 
	     in 
(*
	       Message.msg_string Message.Debug
		 ("bsp: "^(Predicate.toString (Predicate.conjoinL new_cons)));
*)
	       Predicate.conjoinL (new_cons@[pred])
	   end
       | C_SD.Command.Return exp ->
          begin
	   Message.msg_string Message.Debug "bSP of return";
	   match retTarget_o with
	       None -> pred
	     | Some retTarget ->
		 begin
		   Message.msg_printer 
		     Message.Debug C_SD.Command.print 
		     { C_SD.Command.code = 
			 C_SD.Command.Block 
			   [(C_SD.Command.Expr(Expression.Assignment
						 (Expression.Assign, retTarget, exp)))];
		       C_SD.Command.read = []; (* printing ignores .read and .written *)
		       C_SD.Command.written = []}; 
		   (ac_block_sp closure_fun aliases_of_fun lvmap_ref None 
		      (C_SD.Command.Expr
			 (Expression.Assignment(Expression.Assign, retTarget, exp))) 
		      pred)
		 end
        end	
     | _ -> failwith("bad call to block_sp!")


 let block_sp = ac_block_sp (C_SD.expression_closure) get_lval_aliases 
		  
		  
(* a new implementation of strongest postcondition that uses the above -- the regular one
   is flaky in the presence of pointers *)
(* for now, we only do it for blocks of assignments *)
	 (* asgn list : (lval * expression) list *)

 let sp_via_block closure_fun (relevant_future_lvals) pred asgn_list  =
   (* 1. make new lvmap
      2. p = lval_map lvmap pred
      3. p' = block_sp my_alias_fun lvmap pred asgn_list
      4. p'' = Predicate.deep_transform_top f p'
      where
      f (Chlv_exp) =
      if is_latest chlv_exp then Exp.Lval (removechlv chlvexp)
      else chlv_exp
      is_latest chlv_exp = ((lval_map lvmap (removeChlval chlvexp)) = chlvexp)
   *)
 let level = Message.Debug in
 let _ = 
        if (level <> Message.Debug) then
                begin
                let asgn_string = Misc.strList
                        (List.map
			  (fun (lv,e) -> Printf.sprintf "%s := %s " (Expression.lvalToString lv) (Expression.toString e))
			  asgn_list
		       )
                in
                Message.msg_string level ("sp_via_block called with: ") ;
                Message.msg_printer level  Predicate.print pred ;
                Message.msg_string level ("asgns: "^asgn_string)
                end
 in
   let relevant_future_lvals = List.map Expression.push_deref
   relevant_future_lvals 
 in 
   (* 1. make a new lvmap *)
   let sp_lvmap_ref = ref lvmap_empty in
     (* now, lets fill the lvmap with the relevant_future_lvals *)
   (* 2. compute p *)
   let p =
     let updater =
       if (Options.getValueOfBool "clock") then Stats.time "update_map_clock" (update_map_clock_a sp_lvmap_ref) 0
       else Stats.time "update_map" (update_map_a sp_lvmap_ref) 0
     in
       List.iter updater relevant_future_lvals;
       Stats.time "pred_lval_map_a" pred_lval_map_a sp_lvmap_ref 0 pred
   in
   (* 3. compute p' *)
   let p' =
     let sp_aliases_fun lvmap_ref lv =
       (* find all the aliases of lv, in the present lvmap *)
       let cand_ref = ref [] in
       let f (_,lv') _ =
	 if Absutil.queryAlias_cache (Expression.Lval lv) (Expression.Lval lv')
	 then cand_ref := lv'::(!cand_ref)
       in
	 lvmap_iter f lvmap_ref;
	 (lv::!cand_ref)
     in
     let this_sp pred (target,valu) =
       ac_block_sp closure_fun (sp_aliases_fun sp_lvmap_ref) sp_lvmap_ref None
	   (C_SD.Command.Expr (Expression.Assignment(Expression.Assign, target, valu)))
	   pred
     in
       Predicate.conjoinL [p;(List.fold_left this_sp (Predicate.True) asgn_list)]
   in
     (* 4. finally,  compute p'' *)
   let p'' =
     let transformer  = 
       let subs e' =
	 match e' with
	     Expression.Chlval _ ->
	       let dechlv = Expression.replaceChlvals e' in
		 if (lval_map_a sp_lvmap_ref 0 dechlv) = e' (* i.e. e' is the latest name of dechlv *)
		 then dechlv else e'
	   | _ -> e' (* this is strictly an optimization (?) ...
			the above would ensure that if e' was not a chlv, then the lhs of the if above IS a chlv,
			hence NOT eq the rhs, and so you get e' anyway ... *)
       in
	 Expression.deep_transform_top subs
     in
       Predicate.atom_transform transformer p'
   in
(*
     Message.msg_string Message.Debug ("sp_via returns: "^(Predicate.toString p''));
*)
     p''
	 
   (*******************************************************************************************)      

		   (** block-concrete_data_pre *)
 (* to do block_concrete_data_pre we need to know all the lvals that appear in a trace -- so we define load_lv_table here *)
 (* RJ: changing this to only track the values READ along a trace -- the other lvalues are useless to us CHECK!*)
let load_lv_table op_l =
  Message.msg_string Message.Debug "In load_lv_table";
  let closure_fun =
    if (Options.getValueOfBool "fldunroll") then
      C_SD.expression_closure_stamp true (* is_relevant_field_trace *)
    else C_SD.expression_closure
  in
    
  let add_write e = 
    let clos  = closure_fun e in (*TBD:CLOSURE*)
      add_to_lv_table (List.flatten (List.map Absutil.getVarsAndDerefs_exp clos))
  in
let proc_asgn asgn = 
  match asgn with
      C_SD.Command.Expr (Expression.Assignment(Expression.Assign, target, valu)) ->
	add_to_lv_table (Absutil.getVarsAndDerefs_exp valu)
	  (*
	let target_closure = closure_fun (Expression.Lval target) in
	let valu_closure = 
	  List.map 
	    (
	      fun x -> 
		let vx = 
		  (Expression.substituteExpression [(Expression.Lval target, valu)] x)
		in 
		  try Expression.Lval (Expression.push_deref (lv_of_expr vx))
		  with _ -> vx
	    )
	    target_closure
	in
	  add_to_lv_table (List.flatten (List.map Absutil.getVarsAndDerefs_exp (target_closure@valu_closure))) *)
| C_SD.Command.Return e -> (add_to_lv_table (Absutil.getVarsAndDerefs_exp e))
    | _ -> ()
 in
let process_op op = 
     match (C_SD.get_command op).C_SD.Command.code with
	 C_SD.Command.Skip -> ()
       | C_SD.Command.Pred e ->  (add_to_lv_table (Absutil.getVarsAndDerefs e))
       | C_SD.Command.Block el -> List.iter proc_asgn el
       | C_SD.Command.FunctionCall fcallExp -> 
	   let (_,_,target) = C_SD.deconstructFunCall fcallExp in
	   let asgn_list = Absutil.asgn_of_funcall (true,false) None fcallExp in
	   (*let _ =  if (target <> C_SD.junkRet) then add_write (Expression.Lval target) in*)
	     List.iter proc_asgn asgn_list
       | C_SD.Command.GuardedFunctionCall (p, fcallExp) ->  (* treat as a Pred followed by a function call *)
           (add_to_lv_table (Absutil.getVarsAndDerefs p)) ;
	   let (_,_,target) = C_SD.deconstructFunCall fcallExp in
	   let asgn_list = Absutil.asgn_of_funcall (true,false) None fcallExp in
	   List.iter proc_asgn asgn_list
       | _ -> ()
   in
  Hashtbl.clear trace_lv_table;
  Hashtbl.clear trace_field_table;
  List.iter process_op op_l

       
 let blocks_of_pred p = 
   match p with 
       Predicate.And l -> l
     | _ -> failwith "bad call to blocks_of_pred"



	 

   (* Pre and Post *)

 (* concrete Pre for a function call *)
 (* take the region, and replace all the variables that are the formals
    of the function you just called with the expressions that were used
    to call it *)	

  let disambiguate_fnptr_call a fname =
    begin
      (* if this is a function pointer call, disambiguate it:
      disambiguation is done by looking at the fname passed in *)
      match ((C_SD.get_command a).C_SD.Command.code) with
        C_SD.Command.FunctionCall (Expression.FunctionCall(Expression.Lval (Expression.Symbol _), _)) 
      | C_SD.Command.FunctionCall (Expression.Assignment(_, _, Expression.FunctionCall(Expression.Lval (Expression.Symbol _), _)))  ->
        a
      | C_SD.Command.FunctionCall (Expression.FunctionCall(_, args)) ->
        C_SD.make_edge (C_SD.get_source a) (C_SD.get_target a) 
		(C_SD.Command.gather_accesses 
                  (C_SD.Command.FunctionCall(Expression.FunctionCall(Expression.Lval (Expression.Symbol fname), args))))
      | C_SD.Command.FunctionCall (Expression.Assignment(aop, t, Expression.FunctionCall(_, args)))  ->
         C_SD.make_edge (C_SD.get_source a) (C_SD.get_target a) 
	   (C_SD.Command.gather_accesses 
              (C_SD.Command.FunctionCall(Expression.Assignment(aop, t, Expression.FunctionCall(Expression.Lval (Expression.Symbol fname), args)))))
      | _ -> failwith "disambiguate_fnptr_call is called with something that is not a function call!"
    end

 (* lastTarget looks at the stack of call points and returns the
    following info: (target := if the call was x = f(y), f, the
    successor location) *)

(* RJ: REMOVE ALL OCCURENCES OF LAST_LOC_POINTER !!! ITS VILE AND DANGEROUS *)

 let last_loc_pointer = ref (None)
 (* Jhala: Another evil hack as greg doesnt let  me look at the top of thestack !! *)

 let lastTarget () = 
   match (!last_loc_pointer) with
       None -> (Message.msg_string Message.Debug "lastTarget called with errorLocation  as top of stack !"); failwith "Empty Call Stack"
     | Some(callLoc) ->
	 let outEdges = C_SD.get_outgoing_edges callLoc in
	 let _ = if (List.length outEdges <> 1) then (Message.msg_string Message.Error "Call point with bad successor !!") in
	 let callEdge = (List.hd outEdges) in
	 let fcallExp = match (C_SD.get_command callEdge) with 
	       {C_SD.Command.code = C_SD.Command.FunctionCall fcexp} -> fcexp 
	     | _ -> failwith "bad command at funcall point !"
	 in
	 let (fname,subs,target) = C_SD.deconstructFunCall fcallExp in
	 let retLoc = C_SD.get_target callEdge in
	 (target,fname,retLoc,subs)


(* Jhala: Misc functions required for the pre of function calls *)
 let getRetexp op =
   let cmd = C_SD.get_command op in
     match cmd.C_SD.Command.code with
	 C_SD.Command.Block l ->
	   begin
	     try
	       begin
		 let last_stm = Misc.list_last l in
		   match last_stm with
		       C_SD.Command.Return e -> e
		     | _ -> failwith "Bad call to getRetexp 1"
	       end
	     with _ -> failwith "Bad call to getRetexp 2"
	   end
       | _ -> failwith "Bad call to getRetexp3"

	   
let isReturn op = 
   let cmd = C_SD.get_command op in
   match cmd.C_SD.Command.code with
       C_SD.Command.Block l -> (List.exists (function x -> match x with C_SD.Command.Return e -> true | _ -> false) l)
   | _ -> false

let isBlock op =
   let cmd = C_SD.get_command op in
   match cmd.C_SD.Command.code with
          C_SD.Command.Block _ -> true
      | _ -> false

let isPredicate op =
   let cmd = C_SD.get_command op in
   match cmd.C_SD.Command.code with
          C_SD.Command.Pred _ -> true
      | _ -> false

let isFunCall op = 
   let cmd = C_SD.get_command op in
   match cmd.C_SD.Command.code with
       C_SD.Command.FunctionCall _ -> true 
   |   C_SD.Command.GuardedFunctionCall _ -> true 
   | _ -> false

	   
 let opFunName op = 
    let cmd = C_SD.get_command op in
    match cmd.C_SD.Command.code with 
    C_SD.Command.FunctionCall fcallExp -> Misc.fst3 (C_SD.deconstructFunCall fcallExp)
    | _ -> Message.msg_string Message.Error "opFunName called with bad arg !"; failwith "opFunName called with bad arg !"

	
 (* INV: p is of the form Forall s. p' *)
 let rec instantiate_quantifiers p = 
   let tf p s lv' = 
     let p' = Predicate.substitute [(Expression.Lval (Expression.Symbol s)),(Expression.Lval lv')] p 
     in
     instantiate_quantifiers p'
   in
   match p with
   Predicate.All (s,p') ->
     let lv'_l = 
       List.filter C_SD.is_skolem
       (C_SD.globals_of_type (C_SD.get_type_of_lval (Expression.Symbol s))) 
     in
     Message.msg_string Message.Normal (Printf.sprintf "[QI]: %s --> %s" s
     (Misc.strList (List.map Expression.lvalToString lv'_l)));
     Predicate.And (List.map (tf p' s) lv'_l)
   | _ -> p
 
   
   (* this function expects the conc_data_reg to look like And ( pred_list) and returns something with that
    form as well *)

 (* this function pretends he's called with the predicate true -- its the callers job to collect 
    all the constraints! 
 *)


(* concrete Pre for a function call *)
(* take the region, and replace all the variables that are the formals
   of the function you just called with the expressions that were used
   to call it *)	

   

 let rec block_concrete_data_pre fname tid atomic_reg cmd edge (* phi_table *) =
   Message.msg_string Message.Debug "In block_concrete_data_pre -- arguments are:" ;
   Message.msg_string Message.Debug ("fname is " ^ fname) ;
   Message.msg_printer Message.Debug (Region.print_atomic_region (Options.getValueOfBool "bddprint")) atomic_reg ;
   Message.msg_printer Message.Debug C_SD.Command.print cmd ;
   Message.msg_printer Message.Debug C_SD.print_edge edge ;
   
   let lvmap_operation_option
     = 
     if (C_SD.isReturn edge) then
       begin
	 Message.msg_string Message.Debug "block c_ Pre: Hit isReturn\n";
	 (* recall that we put the call points in the stack *)
	 (* we assume that a function call edge is the only edge
	    going to its target location *)
	 Message.msg_string Message.Debug "Ingoing edges : ";
	 List.iter (fun a -> Message.msg_string Message.Debug (Operation.to_string a)) (C_SD.get_ingoing_edges atomic_reg.Region.location);
	 Message.msg_string Message.Debug "\nNow matching...\n";
	 let rec get_fn_call_edge lst =
	   match lst with
	       a::rest -> 
		 begin
		   if (C_SD.isFunCall a) then
		     (* check there are no other function calls *)
		     if (List.exists C_SD.isFunCall rest) then
		       failwith "bc pre: More than one function call in incoming edges!"
		     else  
                       disambiguate_fnptr_call a fname  
		   else
		     get_fn_call_edge rest
		 end
	     | [] -> failwith ("pre: No corresponding function call found")
	 in
	 let callEdge = get_fn_call_edge (C_SD.get_ingoing_edges atomic_reg.Region.location) in
         Message.msg_string Message.Normal (C_SD.edge_to_string callEdge) ;
	 let callLoc = C_SD.get_source callEdge in
	 last_loc_pointer := Some(callLoc);
	 let (rt,_,_,subs) = 
           try lastTarget () with C_SD.FunctionPointerException -> 
             begin
               let fcallExp = match (C_SD.get_command callEdge) with
                 { C_SD.Command.code = C_SD.Command.FunctionCall fc } -> fc
               in
               let (_, subs, target) = C_SD.deconstructFunCall fcallExp in
               (target, fname, C_SD.get_target callEdge, subs)
               
             end
         in
         Message.msg_string Message.Normal ("rt is " ^ (Expression.lvalToString rt)) ;
         Message.msg_string Message.Normal ("fname is " ^ fname) ;
	 Absutil.set_ret_pointers rt subs (C_SD.get_location_fname callLoc);
	 last_loc_pointer := None;
	 Some(lvmap_push)
       end
     else
       None
   in
     (* this should be a list of blocks of constraints *)
   let newpred =
     let rec _doBlock stmtlst pred  =
       (* INV: if lvmap_operation_option <> None then stmtlist = [Return] *)
       match stmtlst with
	   [] -> (*if (Options.getValueOfBool "wpsat" && sat_flag) then sat_filter pred (!wpCapRegion) else*) 
	     pred
	 | stmt::rest -> 
	     begin
	       let wp_pred = block_wp lvmap_operation_option fname tid  [stmt] pred in
		 _doBlock rest wp_pred 
	     end
     in
     match cmd.C_SD.Command.code with
	   C_SD.Command.Block el -> _doBlock (List.rev el) (Predicate.True) 
	 | C_SD.Command.Pred e -> 
               let qf_e = instantiate_quantifiers e in
               pred_lval_map tid qf_e 
         | C_SD.Command.GuardedFunctionCall(_, fcallExp)  (* TBD: LOOK AT PREDICATE HERE!! *) 
	 | C_SD.Command.FunctionCall fcallExp ->
	     let callee_name = Misc.fst3 (C_SD.deconstructFunCall fcallExp) in
             let glv_filter lv = in_lval_map tid lv in
             let asgn_list = Absutil.asgn_of_funcall (false,false) (Some glv_filter) fcallExp in
	     let _ =
	       Message.msg_string Message.Debug "asgn_list is:";
	       List.iter (Message.msg_printer Message.Debug C_SD.Command.print_statement)
		 (asgn_list)
	     in
	     let (targ_fname,lvmap_operation_option) = 
	       try 
		 let _ = C_SD.lookup_entry_location callee_name 
		 in (callee_name, Some(lvmap_pop))
	       with C_SD.NoSuchFunctionException _ -> (fname , None)
	     in
	     let rv = block_wp lvmap_operation_option targ_fname tid asgn_list (Predicate.True) in
	       (* RJ: pop only if its a real function that we entered ... *)
		  Message.msg_string Message.Debug ("bwpred is : ");
		  Message.msg_printer Message.Debug Predicate.print rv;
	       rv
	
	 | C_SD.Command.Skip -> Predicate.True
	 | C_SD.Command.Havoc lv -> 
             let block = [C_SD.Command.Expr(Expression.Assignment
             (Expression.Assign, lv, Absutil.get_new_unknown "havoc"))] in
             _doBlock block (Predicate.True)
	 | C_SD.Command.HavocAll -> failwith "Abstraction.block_concrete_data_pre: FunctionCall: TBD"
	 | C_SD.Command.Phi (sym, flag) ->
	     failwith "block_cdp: Phi unimplemented"
	     (*
	       let e = phi_command_to_pred sym flag phi_table in
	       let cmd' = {cmd with C_SD.Command.code = C_SD.Command.Pred e} in
	       let res = concrete_data_pre explicit_flag sat_flag conc_data_reg cmd' edge phi_table in
	       res.Region.pred
	     *)
         | C_SD.Command.SymbolicHook fname ->
                (* RJ: i believe it suffices to just plug these in for
                 * the lvals that are "relevant" i.e. appear in preds,
                 * not for every bloody lvalue mod. 
                 * Reason: if its mod in the body of this trace, then no need.
                 * if its not mod in the body of this trace, then WAIT
                 * first iter -- pred gets added -- lv becomes relevant.
                 * second iter -- after mod-smashing in summary_post gives a
                 * ctrex, you will add this constraint and thus the symvar gets
                 * added... *)
             let make_asgn lv = 
                C_SD.Command.Expr ( Expression.Assignment
                                        (Expression.Assign, lv, Expression.Lval
                                        lv))
              in
              let globs_smashed = (PredTable.relevant_globals_mod fname) in
              let _ = Message.msg_string Message.Debug 
                        ("Hook-plug: "^(Misc.strList (List.map
                        Expression.lvalToString globs_smashed)))
              in
              let fmls =
                let acf = match C_SD.lookup_formals fname with C_SD.Fixed fml -> fml |
                C_SD.Variable fml -> fml in
                List.map (fun s ->  Expression.Symbol s) acf
              in
              let asgn_list = List.map make_asgn (globs_smashed @ fmls) in
              let block_pred = _doBlock (asgn_list) (Predicate.True) 
		in
		 block_pred
   in
   let _ = Absutil.reset_ret_pointers () in
   newpred
     (* { Region.pred = newpred ; } *)


   (* Concrete predecessor: Given a pair (C_SD location, predicate), and a C_SD command operation,
      compute the set of concrete predecessor states.

      This is used in the counterexample-based refinement process.
   *)

   (*******************************************************************************************)
       
   let isPredicate op = C_SD.isPredicate op

   let can_escape lv = C_SD.can_escape lv

   let is_lock v = C_SD.is_a_lock v

   let is_global = C_SD.is_global 

   let is_foo foo_check r = 
     try
       let (loc,stack) = Region.getPcStack r in
       let stack_locs = loc::(locations_of_stack stack) in
	 List.exists foo_check stack_locs 
     with _ -> false (* failwith "is_foo fails!"*)

   let is_atomic = is_foo C_SD.is_atomic

   let is_event = is_foo C_SD.is_event

   let is_task = is_foo C_SD.is_task

   let is_spec = is_foo C_SD.is_spec

		   
   (*******************************************************************************************)

		   
   (* function to substitute tid -> notTid and locals with fresh copies.
      Uses the global phi_com_counter to get new names for locals 
      *)

   (******************************************************************************)
   (* new : with aliasing *) (** HEREHERHEREHEREHERE *)
   let phi_command_to_pred sym flag phi_table =
     assert (flag == true) ;

     let phi_pred newname sym1 sym2 =
       let phi = 
	 try
	   Region.rename_vars 
	     !(Hashtbl.find phi_table sym2) 
	     (newname) 
	 with Not_found ->

	   begin
	     (* RUPAK: check that this is sound.
		If the name sym2 is not in the phi_table, this means 
		that the syntactic lvalue name sym2 is never used in
		the program to read an address or to update an address.
		Hence we can assume that the phi is false.
		*)
	     Message.msg_string Message.Error ("phi_command_to_pred raised not found on lvalue "^(Expression.lvalToString sym2)) ;
	     Region.bot
	   end
       in
       begin
	 match phi with
	   Region.Empty -> Predicate.False
	 |	_ ->
	     begin
	       if (sym1 = sym2) then Region.region_to_predicate phi
	       else
		 begin
		   Predicate.And 
		     [ Predicate.Atom (Expression.Binary(Expression.Eq, 
							 Expression.addressOf (Expression.Lval sym1),
							 Expression.addressOf (Expression.Lval sym2)));
		       Region.region_to_predicate phi ] 
		 end
	     end
       end
     in
     if can_escape sym then
       begin
	 if flag then 
	   begin
	     let aliases = AliasAnalyzer.getAllAliases sym in
	     let newname = new_name (Absutil.brand_new_suffix ()) in
	     let p = Predicate.normalize (Predicate.Or (List.map (phi_pred newname sym) (sym :: aliases)))
	     in
	     Message.msg_string Message.Normal ("Phi for "^(Expression.lvalToString sym)^" is "^(Predicate.toString p)) ;
	     p
	   end
	 else 
	   begin
	     failwith "phi_command_to_pred : flag should not be false "
	   end
       end
     else
       Predicate.False

 (* Jhala: End of misc functions ... *)




       
       (* Type : Region.concrete_data_region -> C_SD.Command.t -> C_SD.edge -> Region.concrete_data_region *)
   let rec concrete_data_pre explicit_flag sat_flag conc_data_reg cmd edge phi_table =
     Message.msg_string Message.Debug "In concrete_data_pre -- arguments are:" ;
     if explicit_flag then (Message.msg_string Message.Debug " EXPLICIT WP ");
     Message.msg_printer Message.Debug Region.print_concrete_data_region conc_data_reg ;
     Message.msg_printer Message.Debug C_SD.Command.print cmd ;
     let pred = conc_data_reg.Region.pred in
     let newpred =
       let rec _doBlock stmtlst pred i =
		 match stmtlst with
		   [] -> if (Options.getValueOfBool "wpsat" && sat_flag) then sat_filter pred (!wpCapRegion) else pred
		 | stmt::rest -> 
		     begin
		       let wp_pred = wp explicit_flag sat_flag stmt pred (*  (edge,i) *) in
		       if (rest = []) then 
			  if (Options.getValueOfBool "wpsat" && sat_flag) then 
				  sat_filter wp_pred (!wpCapRegion)
			       else 
				  wp_pred
		       else
			  _doBlock rest wp_pred (i+1)
		     end
       in
       match cmd.C_SD.Command.code with
	   C_SD.Command.Block el ->
	     begin         
	       (* PREV: List.fold_right (fun a -> fun b -> wp false a b (edge,0)) el pred *)
	       _doBlock (List.rev el) (* predicate *)pred (* end of block *)0 
	     end
	 | C_SD.Command.Pred e -> Predicate.conjoinL  [e; pred]
	 | C_SD.Command.FunctionCall fcallExp ->
	     let asgn_list = Absutil.asgn_of_funcall (true,false) None fcallExp in
	     let _ =
	       Message.msg_string Message.Debug "asgn_list is:";
	       List.iter (Message.msg_printer Message.Debug C_SD.Command.print_statement)
		 (asgn_list)
	     in
               _doBlock asgn_list pred 0
	 | C_SD.Command.Skip -> pred
	 | C_SD.Command.Havoc lv ->
	     let new_var = 
	       match Absutil.getNewTmpVar (Expression.Symbol "__foo_havoc") with 
		 Expression.Chlval (Expression.Symbol x, s) -> s^x
	       | _ -> failwith "Abstraction.concrete_data_pre: Havoc: TBD"
	     in
	     let new_name v = 
	       if (v = lv) 
	       then Expression.Symbol new_var
	       else v
	     in 
	     Predicate.alpha_convert new_name pred 
	 | C_SD.Command.HavocAll -> failwith "Abstraction.concrete_data_pre: FunctionCall: TBD"
	 | C_SD.Command.Phi (sym, flag) -> 
	     let e = phi_command_to_pred sym flag phi_table in
	     let cmd' = {cmd with C_SD.Command.code = C_SD.Command.Pred e} in
	     let res = concrete_data_pre explicit_flag sat_flag conc_data_reg cmd' edge phi_table in
	     res.Region.pred
         | C_SD.Command.SymbolicHook _ -> pred
     in
     { Region.pred = newpred ; }

  
 (** block-wp end *)

 let rec concrete_data_post conc_data_reg cmd edge phi_table =
     Message.msg_string Message.Debug "In concrete_data_post -- arguments are:" ;
     Message.msg_printer Message.Debug Region.print_concrete_data_region conc_data_reg ;
     Message.msg_printer Message.Debug C_SD.Command.print cmd ;

     let pred = conc_data_reg.Region.pred in
     let flag = Options.getValueOfBool "wpsat" in
     let newpred =
       match cmd.C_SD.Command.code with
	   C_SD.Command.Block el -> List.fold_left (fun a -> fun b -> sp flag b a) pred el
	 | C_SD.Command.Pred e -> Predicate.normalize (Predicate.And [e; pred])
	 | C_SD.Command.FunctionCall fcallExp -> 
	     begin
	       let (_, _subs, _) = C_SD.deconstructFunCall fcallExp in
	       let asgn_list = List.map (fun (x,e) -> 
		 C_SD.Command.Expr(Expression.Assignment 
				     (Expression.Assign,Expression.Symbol(x),e))) _subs 
	       in         
	       List.fold_left (fun a -> fun b -> sp flag b a) pred asgn_list
	     end
	 | C_SD.Command.Skip -> pred
	 | C_SD.Command.Havoc sym -> 	    
	     let new_var = (* sym should be assigned a fresh variable *)
	       match Absutil.getNewTmpVar (Expression.Symbol "__foo") with 
		 Expression.Chlval (Expression.Symbol x, s) -> Expression.Lval (Expression.Symbol (s^x))
	       | _ -> failwith "Abstraction.concrete_data_post: Havoc of lval: TBD"
	     in
	     let assign = Expression.Assignment(Expression.Assign, sym, new_var) in
	     let stmt = C_SD.Command.Expr assign in
	     sp flag stmt pred
	 | C_SD.Command.HavocAll -> failwith "Abstraction.concrete_data_post: HavocAll: TBD"
	 | C_SD.Command.Phi (sym, flag) -> 
	     let e = phi_command_to_pred sym flag phi_table in
	     Predicate.normalize (Predicate.And [e; pred])
         | C_SD.Command.SymbolicHook _ -> pred
       in
     { Region.pred = newpred ; }




