(*
 * Description:
 *
 * Copyright (c) 2000 by
 *  David Blei			blei@cs.berkeley.edu
 *  Chris Harrelson		chrishtr@cs.berkeley.edu
 *  Ranjit Jhala		jhala@cs.berkeley.edu
 *  Rupak Majumdar		rupak@cs.berkeley.edu
 *  George C. Necula	necula@cs.berkeley.edu
 *  Shree P. Rahul		sprahul@cs.berkeley.edu
 * 	Wes Weimer			weimer@cs.berkeley.edu
 *  Dror Weitz			dror@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1.  Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products derived from 
 * this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)
module E = VampyreErrormsg
module I = Input
module P = Proof
module S = Satproc
module D = Dag
    
open VampyrePretty
open Dag
open UnionFind

let ccid = ref ""
let debug = (ref false)
let rflag = ref false     
    
let sysProveIt = S.proveIt
    
let mesgToProofInfo m =
  let mbodyToInfoT x = match x with
    S.Equality (p,q) -> Equality (p.id,q.id)
  | S.Literal p -> Literal p.id
  in
  {body = mbodyToInfoT m.S.body;
    owner = m.S.owner;
    token = m.S.id;
    goal = m.S.goal}

let proofInfoToMesg pi =
  let infoTToMbody x =
    match x with 
      Equality (i,j) ->
	(try 
	  S.Equality (findDag i,findDag j)
	with
	  Not_found -> print_string "CC: error equality ";print_int i;print_string "  ";E.s (E.bug "Dammit Beavis equality !! "))
    | Literal i ->
	(try 
	  S.Literal (findDag i)
	with
	  Not_found -> print_string "CC: error literal ";print_int i;print_string "  ";E.s (E.bug "Dammit Beavis literal !! ")
	)      
  in
  {S.body = (infoTToMbody pi.body);
    S.owner = pi.owner;
    S.id = pi.token;
    S.goal = pi.goal;}
let infRule op =
  match op with
    BFun(s) -> ("not_FI",false)
  | NotBFun(s) -> ("not_FI",true)
  | Eq -> ("eqNeq_FI",false)
  | Neq -> ("eqNeq_FI",true)
  | (Geq | Gt | Leq | Lt ) ->  ("NONSENSE_AXIOM",false)
	 (*(E.s (E.bug "CC: weird atom Gt")) *)
  | _ -> E.s (E.bug "CC: Asserting non atom")

    
let myId = ref "CC"
 
let idf x = List.map (function p -> p.id) x

let nodeProofItem x = 
  let xv = nodeVal x in
  match xv with
    DAtom a -> P.F (I.Atom a)
  | DExp e -> P.E e
	
    
let reflex x =
  let xv = nodeVal (findDag x) in
  match xv with
    DAtom a -> P.G ("b=id",[P.F (I.Atom a)])
  | DExp e -> P.G ("=id",[P.E e])
	
let trans bflag x y  =
  if bflag then P.G ("b=tr",[(P.P x);(P.P y)]) else   P.G ("=tr",[(P.P x);(P.P y)])  


let sym bflag x =
  (* Poor man's fix of the =sym problem *)
  if bflag then
    match x with
      P.G ("b=sym", [P.P a]) -> a
    | _ -> P.G ("b=sym", [(P.P x)])
  else
    match x with
      P.G ("=sym", [P.P a]) -> a
    | _ -> P.G ("=sym", [(P.P x)])
(*Was:  if bflag then P.G ("b=sym",[(P.P x)]) else  P.G ("=sym",[(P.P x)]) *)

let trans3 bflag x y z = trans bflag (trans bflag x y) z

    
(* the above need to be changed once they are put into dag.ml *)
let proofTableSize = 109
(*let proofTable = Hashtbl.create proofTableSize    *)

exception Bad_call_to_pathPf
         
let rec pathPf x y =
  let bflag = not (isExp !(Hashtbl.find dagTable x)) in
  let px = (Hashtbl.find ufTable x).UF.parent.UF.idt in
  let _ =
    if (!debug) then (
      print_string "pathPf " ;print_int x;print_string "..."; print_int y;print_string "..."; print_int px;print_string "......\n") in 
    
    if x = y then reflex x
    else (
      if px = y  then pPf x
      else (
	let p1 = pPf x in
	let p2 = pathPf px y in
	  trans bflag p1 p2
      )
    )
and pPf x =
  let bflag = not (isExp !(Hashtbl.find dagTable x)) in
  let x_node = (Hashtbl.find ufTable x) in
  let px_node = x_node.UF.parent in
  let px = px_node.UF.idt in
  let _ =
    if (!debug) then (
      print_string "ppf " ;print_int x;print_string "..."; print_int px;print_string "......\n" ) in 
    
    if x = px then reflex x
    else (
      let (a1,b1) = x_node.proof in
      let flip =
	if (x_node.sP.owner = "CC") then false
	else
	  match x_node.sP.body with
	      Equality (i1,i2) ->
		
		(if (!debug) then print_string "CC: got here instead") ;
		
		if (a1,b1) = (i1,i2) then false
		else (
		  if (a1,b1) = (i2,i1) then true else E.s (E.bug "CC: Strange info in x, px node -- Eq"))
	    | Literal i ->
		let anode = findDag i in
		  if not ((anode.op = Eq) || (anode.op = Neq))  then E.s (E.bug "CC: Strange info in x px node-- op not Eq")
		  else ( if (a1,b1) = ((List.nth anode.args 0).id ,(List.nth anode.args 1).id) then false
			 else ( if (b1,a1) = ((List.nth anode.args 0).id,(List.nth anode.args 1).id) then true
				else E.s (E.bug "CC: Strange info in x px node -- Literal ")))
      in
      let _ =
	if (!debug) then (
	  print_string "ppf1 " ;print_int a1;print_string "..."; print_int b1;print_string "......\n") in 
	(*let (how,who) = x_node.sP.owner in*)
      let a1b1_Pf_temp = if (not (x_node.sP.owner = "CC")) then ((!sysProveIt) (proofInfoToMesg x_node.sP)) else proveCong (!(Hashtbl.find dagTable a1)) (!(Hashtbl.find dagTable b1))
	
      in
      let a1b1_Pf = if flip then (sym bflag a1b1_Pf_temp) else a1b1_Pf_temp in
	if (a1,b1) = (px,x) then (sym bflag a1b1_Pf)
	else (
	  
	  let a1_p_Pf = pathPf a1 px in
	  let b1_p_Pf = pathPf b1 x in
	    trans3 bflag (sym bflag  b1_p_Pf) (sym bflag  a1b1_Pf) (a1_p_Pf)
	)
    )
and
  proveCong n1 n2 = 
  let cong_axiom n = 
    match n.op with
	OpName(s) -> (E.s (E.bug "BUG: cong_axiom error ..."); failwith ("cong_axiom: error ! OpName"))
  | Int(i) -> (E.s (E.bug "BUG: cong_axiom error ..."); failwith ("cong_axiom: error ! Int"))
  | Var(v) -> (E.s (E.bug "BUG: cong_axiom error ..."); failwith ("cong_axiom: error ! Var"))                     (* An integer constant *)
  | Plus -> "=~+"
  | Minus -> "=~-"
  | Times -> "=~*"
  | Div _ -> "=~div"
  | Mod _ -> "=~mod"
  | Floor _  -> "=~floor"
  | Ceil _ -> "=~ceil"
  | Fun (s) -> "=~"^s

                                        (* Some boolean nodes *)
  | BFun(s) -> "b=~"^s
  | NotBFun(s) -> "notb=~"^s
  | Eq -> "=~eq"
  | Neq -> "=~neq"
  | Geq -> "=~geq"
  | Gt -> "=~gt"
  | Leq -> "=~leq"
  | Lt -> "=~lt"
  in
  try 
    P.G (cong_axiom n1, (List.map2 (function x -> function y -> P.P (proveEq x y)) (idf n1.args) (idf n2.args)))
  (*  P.G ("cong",[(nodeProofItem n1);(nodeProofItem n2)]@	(List.map2 (function x -> function y -> P.P (proveEq x y)) (idf n1.args) (idf n2.args))) *)
  with Invalid_argument _  -> E.s (E.bug "CC: bad call to proveCong!!")
				
and  proveEq x y =
  let bflag = not (isExp !(Hashtbl.find dagTable x)) in
  let xnode = (Hashtbl.find ufTable x) in
  let ynode = (Hashtbl.find ufTable y) in
  (* let _ = print_int x;print_string "  ";print_int y;print_string "---" in*)
  if (UF.onPathtoRoot xnode ynode) then pathPf x y
  else (
    if (UF.onPathtoRoot ynode xnode) then (sym bflag (pathPf y x))
    else (
      let rootx = UF.find xnode in
      let rooty = UF.find ynode in
      if rooty = rootx then(
	let px = pathPf x rootx in
	let py = pathPf y rootx in
	trans bflag px (sym bflag py)
	  )
      else E.s (E.bug "CC: Incorrect call to proveEq")
	  )
	)
      
let checkEq x y =
  try
    let _ = proveEq x y in
    true
  with Jhala -> false
    (* the two nodes supplied as ids - not as expressions ! *)	
      
let rec checkContra l =
  match l with
    [] -> ()
  | (n1,n2)::tl ->
      let bflag = not (isExp (findDag n1)) in
      if bflag then (
	let n1_node = findDag n1 in
	let n2_node = findDag n2 in
	if (n1_node.isAsserted & n2_node.mesg.goal) || (n2_node.isAsserted & n1_node.mesg.goal)  then (
	  let t = if n2_node.mesg.goal then n2_node else n1_node in
	  let gnode = if n1_node.isAsserted then n1_node else n2_node in
	  let info = t.mesg in
	  (* t is the guy we have the not pf of --- it was a goal *)
	  (* gnode is the guy we have pf of --- it was asserted *)
	  (* from here its identical to this section of addGoal ...*)
	  let pcons = "bIffMP" in
	let (pcons1 ,flip) = infRule t.op in
	match t.op with

	  (Geq| Gt| Leq| Lt)-> if (!debug) then (print_string "CC: Simplex should get this contra !!")

	|	_ ->
	    let p4 = P.P ((!sysProveIt) (proofInfoToMesg info)) in (* pf not t*)
	    let p1 = P.P ((!sysProveIt) (proofInfoToMesg gnode.mesg)) in (* pf gnode *)
	    let p2 = P.P (proveEq (gnode.id) (t.id)) in (* pf gnode <=> t *)
	    let p3 = P.P (P.G (pcons,[p1;p2])) in (* pf t *)
	    if flip then 
	      raise (S.Contradiction (P.G (pcons1,[(p4);(p3)])))
	    else
	      raise (S.Contradiction (P.G (pcons1,[(p3);(p4)])))
	  ));
	if Hashtbl.mem disEqTable (n1,n2) then (
	  let pf1 = proveEq n1 n2 in
	  let pf2 = (!sysProveIt) (proofInfoToMesg (Hashtbl.find disEqTable (n1,n2))) in
	  if (!debug) then print_string "gggg0";
	  raise (S.Contradiction (P.G ("eqNeq_FI",[(P.P pf1);(P.P pf2)])))
	    )
	else (
	  if Hashtbl.mem disEqTable (n2,n1) then (
	    let pf1 = proveEq n2 n1 in
	    let pf2 = (!sysProveIt) (proofInfoToMesg (Hashtbl.find disEqTable (n2,n1))) in
	    if (!debug) then print_string "gggg1";
	    raise (S.Contradiction (P.G ("eqNeq_FI",[(P.P pf1);(P.P pf2)])))
	      )
	  else
	    checkContra tl)
	    
	  
let print_discovery (n1,n2) = print_string "(";print_int n1;print_string ",";print_int n2;print_string ")"

    
let addAssume (t1,t2) info =
  let redundant (a,b) = not ((a,b)=(t1,t2) || (b,a)=(t1,t2) ) in
  discoveryList := [];
  if ((UF.find (Hashtbl.find ufTable t1) <> UF.find (Hashtbl.find ufTable t2))
      & not (info.owner = "CC"))
  then (
    (try 
      merge (t1,t2) (info)
    with Stack_overflow -> E.s (E.bug "CC: s.o. here"))
      ;
    checkContra (!(discoveryList));
    if (!debug) then (
      print_string "discovered ";
      List.iter print_discovery (List.filter redundant !discoveryList);
      print_string "\n"
	);
    let _ = if ((!rflag) && (!debug)) then showEqClass () else [] in
    List.filter redundant (!discoveryList)
      )
  else []
      
(*
let goalRule op =
  match op with
    BFun(s) -> ("bIff_NotBFun")
  | NotBFun(s) ->"bIff_BFun" 
  | Eq -> "bIff_Neq"
  | Neq -> "bIff_Eq"
  | Geq -> "bIff_Lt"
  | Gt -> "bIff_Leq" 
  | Leq -> "bIff_Gt"
  | Lt -> "bIff_Geq"
  | _ -> E.s (E.bug "CC: Invalid op in Node")
*)
let equivClass n =
  List.map findDag ((findDag (UF.find (Hashtbl.find ufTable n.id))).eqClass)
      
let addAssert t info =
  if (t.isAsserted) then ()
  else (
    if (t.mesg.goal = true) 
    then (
      let (pcons,flip) = infRule t.op in
      match t.op with
	(Geq| Gt| Leq| Lt)-> if (!debug) then (print_string "CC: Simplex should get this contra !!")

      |	_ ->
	  let p1 = P.P ((!sysProveIt) (proofInfoToMesg info)) in
	  let p2 = P.P ((!sysProveIt) (proofInfoToMesg t.mesg)) in
	  if flip then 
	    raise (S.Contradiction (P.G (pcons,[(p2);(p1)])))
	  else
	    raise (S.Contradiction (P.G (pcons,[(p1);(p2)])))
	      )
    else (
      try 
	let gnode = List.find (function n -> n.mesg.goal = true) (equivClass t) in  
	let pcons = "bIffMP" in 
	let (pcons1,flip) = infRule gnode.op in (*t is the node being asserted *)
	match t.op with

	  (Geq| Gt| Leq| Lt)-> if (!debug) then (print_string "CC: Simplex should get this contra !!")	    
	|	_ ->
	    let p1 = P.P ((!sysProveIt) (proofInfoToMesg info)) in (* pf t *)
	    let p4 = P.P ((!sysProveIt) (proofInfoToMesg gnode.mesg)) in (* pf not gnode *)
	    let p2 = P.P (proveEq (t.id) (gnode.id) ) in (* pf t <=> gnode *)
	    let p3 = P.P (P.G (pcons,[p1;p2])) in (* pf gnode *)
	    if flip then 
	      raise (S.Contradiction (P.G (pcons1,[(p4);(p3)])))
	    else
	      raise (S.Contradiction (P.G (pcons1,[(p3);(p4)])))
      with Not_found ->
	
	t.isAsserted <- true;
	t.mesg <- info;
	assertStack := (t.id,0)::(!assertStack)
				    )
	)
      
let addGoal t info =
  if (t.mesg.goal) then ()
  else (
    if (t.isAsserted = true)
    then (
      let (pcons,flip) = infRule t.op in
      match t.op with
	(Geq| Gt| Leq| Lt)->if (!debug) then (print_string "CC: Simplex should get this contra !!")
	  
      |	_ ->
	  let p2 = P.P ((!sysProveIt) (proofInfoToMesg info)) in
	  let p1 = P.P ((!sysProveIt) (proofInfoToMesg t.mesg)) in
	  if flip then 
	    raise (S.Contradiction (P.G (pcons,[p2;p1])))
	  else
	    raise (S.Contradiction (P.G (pcons,[p1;p2])))
	      )
    else (
      try
	let gnode = List.find (function n -> n.isAsserted) (equivClass t) in  
	let pcons = "bIffMP" in
	let (pcons1 ,flip) = infRule t.op in
	match t.op with

	  (Geq| Gt| Leq| Lt)-> if (!debug) then (print_string "CC: Simplex should get this contra !!")

	|	_ ->
	    let p4 = P.P ((!sysProveIt) (proofInfoToMesg info)) in (* pf not t*)
	    let p1 = P.P ((!sysProveIt) (proofInfoToMesg gnode.mesg)) in (* pf gnode *)
	    let p2 = P.P (proveEq (gnode.id) (t.id)) in (* pf gnode <=> t *)
	    let p3 = P.P (P.G (pcons,[p1;p2])) in (* pf t *)
	    if flip then 
	      raise (S.Contradiction (P.G (pcons1,[(p4);(p3)])))
	    else
	      raise (S.Contradiction (P.G (pcons1,[(p3);(p4)])))
      with Not_found ->
      (*t.isAsserted <- true;*)
      t.mesg <- info;
      assertStack := (t.id,1)::(!assertStack);
	)
	)
      
(* this should only be called a second time ... *)      
let recheckGoal t =
  let info = t.mesg in
  if (t.isAsserted = true)
    then (
      let (pcons,flip) = infRule t.op in
      match t.op with

	(Geq| Gt| Leq| Lt)->if (!debug) then (print_string "CC: Simplex should get this contra !!")
      |	_ ->
	  let p2 = P.P ((!sysProveIt) (proofInfoToMesg info)) in
	  let p1 = P.P ((!sysProveIt) (proofInfoToMesg t.mesg)) in
	  if flip then 
	    raise (S.Contradiction (P.G (pcons,[p2;p1])))
	  else
	    raise (S.Contradiction (P.G (pcons,[p1;p2])))
	      )
    else (
      try
	let gnode = List.find (function n -> n.isAsserted) (equivClass t) in  
	let pcons = "bIffMP" in
	let (pcons1,flip) = infRule t.op in
	match t.op with

	  (Geq| Gt| Leq| Lt)->if (!debug) then (print_string "CC: Simplex should get this contra !!")
	    
	|	_ ->
	    let p4 = P.P ((!sysProveIt) (proofInfoToMesg info)) in
	    let p1 = P.P ((!sysProveIt) (proofInfoToMesg gnode.mesg)) in
	    let p2 = P.P (proveEq (gnode.id) (t.id)) in
	    let p3 = P.P (P.G (pcons,[p1;p2])) in
	    if flip then 
	      raise (S.Contradiction (P.G (pcons1,[(p4);(p3)])))
	    else
	      raise (S.Contradiction (P.G (pcons1,[(p3);(p4)])))
      with Not_found ->()
      (*t.isAsserted <- true;*)
      	  )
      
let expensiveChecks () =
  let rec expCheck l =
    match l with
      (-1,-1)::tl -> expCheck tl
    | (i,0)::tl -> expCheck tl
    | (i,1)::tl -> (recheckGoal (findDag i));(expCheck tl)
    | _::tl -> E.s (E.bug "CC: Junk in assertStack !")
    | [] -> ()
  in
  expCheck (!assertStack)
    
    
let addDiseq (t1,t2) info  =
  if (!debug) then (print_string "CC: adding diseq";print_int t1;print_string " ";print_int t2;print_string "... \n");
  if not(UF.find (Hashtbl.find ufTable t1) = UF.find (Hashtbl.find ufTable t2))
  then (
    Hashtbl.add disEqTable (t1,t2) info;
    (disEqStack := (t1,t2)::(!disEqStack))
      )
  else (
    if (!debug) then (
      print_string "CC: got CONTRA! :";
      );
    let p1 = P.P (proveEq t1 t2) in
    let p2 = P.P ((!sysProveIt) (proofInfoToMesg info)) in
    raise (S.Contradiction (P.G ("eqNeq_FI",[p1;p2])))
      )
      
let transClos () =
  (* We basically just update roots ! *)
  let upd n p_n = 
    let ufn = Hashtbl.find ufTable n in
    (!p_n).root <- !(Hashtbl.find dagTable (UF.find ufn))
  in
  Hashtbl.iter upd dagTable
    

let snapshot () =
  try
    (if (!debug) then print_string "CC: snapshot called\n");
		D.snapshot ()
  with Not_found -> E.s (E.bug "CC: raises Not_found in Snapshot")
  (*E.s (E.unimp "Eq.snapshot")*)
let undo () =
  try
    (if (!debug) then print_string "CC: undo called\n");
		D.undo ()
  with Not_found -> E.s (E.bug "CC: raises Not_found in Undo")
  (* E.s (E.unimp "Eq.undo") *)

let makeMessage (n1,n2) =
  {S.body = S.Equality ((D.findDag n1),(D.findDag n2));
    S.owner = !ccid;
    S.id = 0;
    S.goal = false;}




let assume mlist msg =  
  if (!debug) then (
    
    ignore (printf "%s: Assuming... %a\n" !myId  Satproc.dm msg)
      );
  (
  try 
  match msg.S.body with
    S.Equality (n1,n2) ->
	let _ = if (!debug) then (print_string "CC: Equality "; print_int n1.id; print_string ", ";print_int n2.id;print_string " ...\n") in
		
	if msg.S.goal then (( addDiseq (n1.D.id,n2.D.id) (mesgToProofInfo msg) );mlist)
      else (
	let retList =  addAssume (n1.D.id,n2.D.id) (mesgToProofInfo msg) in
	(List.map makeMessage retList ) @ mlist)
  | S.Literal n  ->
      (* print_string "Literal ";let _ = (D.dn () n) in *)
      if msg.S.goal then (
	 addGoal n (mesgToProofInfo msg);
	match n.D.op with
	  D.Eq ->
	    (try 
	      let n1 = List.nth n.D.args 0 in
	      let n2 = List.nth n.D.args 1 in
	      
	    (( addDiseq (n1.D.id,n2.D.id) (mesgToProofInfo msg) );mlist)
	    with Failure _ -> E.s (E.bug "CC: Eq node in DAG doesnt have 2 args..."))
	| D.Neq -> (
(* not neq --> eq ... so I will generate new things... *)
	    try 
	      let n1 = List.nth n.D.args 0 in
	      let n2 = List.nth n.D.args 1 in
	      let retList =  addAssume (n1.D.id,n2.D.id) (mesgToProofInfo msg) in
	    (List.map makeMessage retList ) @ mlist
	    with Failure _ -> E.s (E.bug "CC: Neq node in DAG doesnt have 2 args...")	)			
	| _ -> mlist (* not relevant ...,unless further diseq cases are there... *)
	    )
      else (
	(* its an assertion ... *)
	addAssert n (mesgToProofInfo msg);
	match n.D.op with
	  D.Neq ->(
	    try 
	      let n1 = List.nth n.D.args 0 in
	      let n2 = List.nth n.D.args 1 in
	      let _ = if (!debug) then print_string "CC: now adding diseq" in
	      (( addDiseq (n1.D.id,n2.D.id) (mesgToProofInfo msg) );mlist)
	    with Failure _ -> E.s (E.bug "CC: Neq node in DAG doesnt have 2 args...")	)	
	| D.Eq -> (* eq ... so I will generate new things... *)
	    (
	    try 
	      let n1 = List.nth n.D.args 0 in
	      let n2 = List.nth n.D.args 1 in
	      
	      let retList =  addAssume (n1.D.id,n2.D.id) (mesgToProofInfo msg) in
	      (List.map makeMessage retList ) @ mlist
	    with Failure _ -> E.s (E.bug "CC: Eq node in DAG doesnt have 2 args...")       )			
	| _ -> mlist (* not relevant ...,unless further diseq cases are there... *)
	    )
      
  with (S.Contradiction p) -> (if (!debug) then (print_string "GOT CONTRA!!\n"; ignore (printf "%s: contra proof ... %a\n" !myId  P.dp p));      raise (S.Contradiction p))
  (*| _ -> E.s (E.bug "Random Bug in addAssume")*);)

let msgFilter m =
  match m.S.body with
    S.Literal n -> if m.S.owner = "CC" then E.s (E.bug "CC: Where did this come from ") else true
  | _ -> true
	
    
let assume1 m1 m2 =
  let ret = List.filter msgFilter (assume m1 m2) in
  let _ = if (!debug)  then print_string "CC: Leaving assume \n" in
  ret
    
	
(* E.s (E.unimp "Eq.assume") *)

let proveIt msg = if (!debug) then ( ignore (printf "%s: proveIt... %a\n" !myId  Satproc.dm msg);print_string "CC: ProveIt \n");
  match msg.S.body with
    S.Literal n -> if msg.S.owner = "CC" then E.s (E.bug "CC: How did CC claim a literal !!") (*else (!sysProveIt msg)*) else E.s (E.bug "CC: Why am I being asked to prove this !")
  | S.Equality (n1,n2) -> proveEq n1.id n2.id



(*
  match mesg.S.body with
    S.Equality (n1,n2) -> proveEq (n1.
*)

(*[Jhala] to do ... E.s (E.unimp "Eq.proveIt") *)
let reinitialize () =  D.dagReset ()
    
let initialize id = 
  ccid := id;
  Dag.debug := debug;
  { S.snapshot = snapshot;
    S.undo = undo;
    S.assume = assume1;
    S.proveIt = proveIt;
    S.debug = debug;
    S.ident  = id;
    S.reinitialize = reinitialize;
  } 

