(*
 * Description:
 *
 * Copyright (c) 2000 by
 *  David Blei			blei@cs.berkeley.edu
 *  Chris Harrelson		chrishtr@cs.berkeley.edu
 *  Ranjit Jhala		jhala@cs.berkeley.edu
 *  Rupak Majumdar		rupak@cs.berkeley.edu
 *  George C. Necula	necula@cs.berkeley.edu
 *  Shree P. Rahul		sprahul@cs.berkeley.edu
 * 	Wes Weimer			weimer@cs.berkeley.edu
 *  Dror Weitz			dror@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1.  Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products derived from 
 * this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

(* This module defines the language of proofs *)


(* In order to maintain a compact representation of proofs we represent a 
 * proof node as the name of the axiom along with a list of arguments. An 
 * exception is made for the hypothetical and parametric proof nodes. *)
type parg = 
    I of int
  | E of Input.exp
  | P of proof
  | F of Input.pred

and proof = 
    Hyp of Input.var                    (* A hypothesis, 
                                           introduced with Impi*) 
  | Alli of Input.var * proof
  | Impi of Input.var * Input.pred * proof (* bind var to a hypothetical 
                                            * proof of the pred. The last 
                                            * argument is a proof of the 
                                            * right hand side, possibly 
                                            * refering to var *)

                                        (* Bind var to a proof of the 
                                         * negation of pred. The second 
                                         * argument is a proof of false *)
  | Contra of Input.var * Input.atom * proof

  | G of string * parg list             (* This is a generic proof 
                                         * constructor *)
  | Unimplemented                       (* We'll expand the proof language as 
                                         * needed in the future. Use 
                                         * Unimplemented until then *)


exception ProofCheck of proof
   (* Infer the "type" of a proof. First argument is a list of variables and 
    * the second is a mapping of hypothesis variables to the predicates they 
    * prove. Raise ProofCheck exception if not a valid proof.  *)
val infer : Input.var list -> 
            (Input.var * Input.pred) list -> proof -> Input.pred

val infer_reset : unit -> unit

val inferred_atoms : unit -> Input.atom list

val dp : unit -> proof -> VampyrePretty.doc


(* begin circular dependency hack: *)
val simp_and_subtract_is_zero : (Input.exp -> Input.exp -> bool) ref
val simp_and_subtract_is_at_least_zero : (Input.exp -> Input.exp -> bool) ref
(* end circular dependency hack *)
